# coding=UTF8
# Copyright © 2011 Rob Leachman
# Please see the file COPYING in this distribution for license information
#
# TODO:
#DONE: OS-specific readme's or something
#DONE: Default music volume should be lower
#DONE: I should have received a copy of the GPL
#DONE: print # of lives on maze
#DONE: sounds need to be repacked into low-bit-rate MP3s for size
#DONE: The command-line recording replay feature doesn't seem to work as advertised
#
# Fix the format of the GUI traceback, wrap hammers it
#
# Speed up after level 15?
# Should enable risky spawn per the doc, at level 10?
# *** default to 1024x768 if the machine can handle it!
#
# 0.1 issues and feedback -
# - Is it intended that I be able to peg snakes by hitting them in the side
#   of the head?
# - It would be great to have tooltip-like help bubbles the first time each 
#   new thing appears on screen---white eggs, spotted eggs, frog, green snakes, etc.
#--------------------------------
# ROADMAP TO 1.0:
#
# - Debug issues found in beta
#
# - Rework animation logic to include movement = velocity/time
#    (then adjust enemy speed intelligently (FPS and level)
#
# - Rework segment movement logic to correctly position every frame
#
# - Paint collisions better, the segments should "pop"
#
# - Roll in new maze on winning level instead of full redraw
#
# - Provide features expected in 2011:
# * user profiles
#    (and set speeds based on experience)
# * "campaign" mode
# --> ramp graphics/sound from 8-bit '83 style to 2011
# --> challenge levels
#     pacman
# --> achievements
#     finished level with tail intact
#     killed all with bites to the neck
#     killed all with headshots (no tail nibbling)
#     killed all with nibbles
#     ate every frog (that got ate and didn't escape?)
#     successfully hatched an egg
#     finished with 1 segment
#     spawned N times with no success
#     extra life on points
#     killed N enemy spawns
# --> beat all 25 mazes
# --> powerups (speed slime, invincibility, what else?)
#
# - Multiplayer
# * Support 2 players on split screen sitting side by side
# * Competiton/cooperation modes?
# * Network play?
#
# - Missions
# * Locked maze areas with keys and such
#--------------------------------
#
#DEFERRED AT 0.1:
# - "attract" autoplay demo at title
# - fully implement gamepad/joystick
# - fake 3D (not-black background, black shadows)
# - rotation? scaling animation of segments?
# - get the glyph font pretty at lores
# - rework all files to use Cerealizer; never import pickle
#
# - (2011 idea: if frog eats egg, becomes super frog for a few seconds
#    and will hop faster and likely eat the other egg too or at least
#    last a bit longer)
# - Hatch our snake and have him help

import sys
import os
import pickle # for shipped game files only
import pygame
import pprint
import random
import datetime
import copy
import re # for pyconsole
import traceback
import textwrap
from pygame.locals import *

from lib import Global
from lib import Mainline
from lib import Snake
from lib import Segment
from lib import Frog
from lib import Esnake
from lib import Scores

from lib import Noise
from lib import MazeWallSprite
from lib import Maze
from lib import pyconsole
from lib import pyconsole_syntax
from lib import pyconsole_functions
from lib import cerealizer


def door_management(screen, 
                    drawnMaze, 
                    snake, 
                    lastEnemy, 
                    ourStartPause, 
                    enemyLaunchDoorPause, 
                    lifeIndicators, 
                    lives, 
                    ourSnakeIsSafe):
    """ The pen doors open at launch, and close once the snake has passed the
        door. Plus enemies. And the off-screen entrances. And the exit.
    """
    if snake.launchDoor < 1 and not snake.driveToStart:
        if snake.launchDoor < 0:
            if ourStartPause > 20:
                ourStartPause -= 1
            else:                    
                if snake.doorTimer > 0:
                    snake.doorTimer -= 1
                    if snake.doorTimer == 0:
                        snake.doorTimer = 3
                        snake.launchDoor += .2
                        if snake.launchDoor > -.2:
                            snake.launchDoor = 0
                        doorman(screen, drawnMaze, 1, snake.launchDoor * -1)
        else:
            # Get it closed once we pass into playfield
            if snake.body[snake.length-1].fancyDoor:
                if snake.doorTimer > 0:
                    snake.doorTimer -= 1
                    if snake.doorTimer == 0:
                        snake.doorTimer = 3
                        snake.launchDoor += .1
                        doorman(screen, drawnMaze, 1, snake.launchDoor)
            else:
                # Cleared when the last seg moves from node 8 to 7
                if not snake.body[snake.length-1].nextY > 7:
                    #print "close the door"
                    snake.body[snake.length-1].fancyDoor = True
    if lastEnemy.launchDoor < 1:
        if enemyLaunchDoorPause > 0:
            enemyLaunchDoorPause -= 1
        else:                    
            if lastEnemy.launchDoor < 0:
                if lastEnemy.doorTimer > 0:
                    lastEnemy.doorTimer -= 1
                    if lastEnemy.doorTimer == 0:
                        lastEnemy.doorTimer = 3
                        lastEnemy.launchDoor += .2
                        if lastEnemy.launchDoor > -.2:
                            lastEnemy.launchDoor = 0
                        doorman(screen, drawnMaze, 0, lastEnemy.launchDoor * -1)
            else:
                try:
                    if lastEnemy.body[lastEnemy.length-1].fancyDoor:
                        if lastEnemy.doorTimer > 0:
                            lastEnemy.doorTimer -= 1
                            if lastEnemy.doorTimer == 0:
                                lastEnemy.doorTimer = 3
                                lastEnemy.launchDoor += .1
                                doorman(screen, drawnMaze, 0, lastEnemy.launchDoor)
                    else:
                        if not lastEnemy.body[lastEnemy.length-1].nextY > 7:
                            lastEnemy.body[lastEnemy.length-1].fancyDoor = True
                except IndexError:
                    # just slam it shut if we lose our tracking snake
                    lastEnemy.launchDoor = 1
                    doorman(screen, drawnMaze, 0, lastEnemy.launchDoor)
        if lastEnemy.erased:
            lastEnemy.launchDoor = 1
            doorman(screen, drawnMaze, 0, lastEnemy.launchDoor)
            
    # Ditto for the off-screen entrance of the enemies:
    if lastEnemy.enterDoor < 1:
        if lastEnemy.enterDoor < 0:
            if lastEnemy.enterDoorTimer > 0:
                lastEnemy.enterDoorTimer -= 1
                if lastEnemy.enterDoorTimer == 0:
                    lastEnemy.enterDoorTimer = 3
                    lastEnemy.enterDoor += .3
                    if lastEnemy.enterDoor > -.3:
                        lastEnemy.enterDoor = 0
                    doorman(screen, drawnMaze, 3, lastEnemy.enterDoor * -1)
        else:
            try:
                if lastEnemy.body[lastEnemy.length-1].fancyEnterDoor:
                    if lastEnemy.enterDoorTimer > 0:
                        lastEnemy.enterDoorTimer -= 1
                        if lastEnemy.enterDoorTimer == 0:
                            lastEnemy.enterDoorTimer = 3
                            lastEnemy.enterDoor += .1
                            doorman(screen, drawnMaze, 3, lastEnemy.enterDoor)
                else:
                    if not lastEnemy.body[lastEnemy.length-1].nextY > 10:
                        lastEnemy.body[lastEnemy.length-1].fancyEnterDoor = True
            except IndexError:
                lastEnemy.enterDoor = 1
                doorman(screen, drawnMaze, 3, lastEnemy.enterDoor)
        if lastEnemy.erased:
            lastEnemy.enterDoor = 1
            doorman(screen, drawnMaze, 3, lastEnemy.enterDoor)
    # ...and also ditto the snake's off-screen entrance, need subroutine
    if snake.enterDoor < 1:
        if snake.homeGoal == [12, 8] or snake.driveToStart:
            if snake.enterDoor < 0:
                if snake.doorTimer > 0:
                    snake.doorTimer -= 1
                    if snake.doorTimer == 0:
                        snake.doorTimer = 3
                        snake.enterDoor += .2
                        if snake.enterDoor > -.2:
                            snake.enterDoor = 0
                        doorman(screen, drawnMaze, 4, snake.enterDoor * -1)
            else:
                # Get it closed once we pass into playfield
                if snake.body[snake.length-1].fancyEnterDoor:
                    if snake.doorTimer > 0:
                        snake.doorTimer -= 1
                        if snake.doorTimer == 0:
                            snake.doorTimer = 3
                            snake.enterDoor += .1
                            doorman(screen, drawnMaze, 4, snake.enterDoor)
                else:
                    # Cleared when the last seg of the last snake...
                    if (snake.body[snake.length-1].nextX > 10 and 
                        (lifeIndicators == lives - 1 or 
                         lifeIndicators == 2)):
                        snake.body[snake.length-1].fancyEnterDoor = True
        else: # baby leaving the screen
            snake.enterDoor += .2
            if snake.enterDoor > -.2:
                snake.enterDoor = 0
            doorman(screen, drawnMaze, 4, snake.enterDoor * -1)
            if snake.enterDoor == 0:
                snake.enterDoor = 1 # leave it open
    # Provide a sensible way out when we win
    if snake.outDoor < 1:
        if not snake.driveToStart and not snake.fancyDoor and snake.homeGoal <> [12, 8]:
            Mainline.clear_life_count(screen)
        if snake.outDoor < 0:
            # Just reuse the boolean... here it means "got to the door"
            if snake.fancyDoor:
                snake.outDoor += .2
                if snake.outDoor > -.2:
                    snake.outDoor = 0
                doorman(screen, drawnMaze, snake.outDoorNumber, snake.outDoor * -1)
            else:
                if fequal(snake.locationX, snake.nodeBeforeOut[0]) and fequal(snake.locationY, snake.nodeBeforeOut[1]):
                    #print "quick, open sesame!", snake.nodeBeforeOut[0], snake.nodeBeforeOut[1], snake.outDoorNumber
                    snake.fancyDoor = True
                    # If we're opening the door out for the baby,
                    # and it is going offscreen, open the enter
                    # door as well...
                    if snake.outDoorNumber == 2 and snake.homeGoal == [10, 13]:
                        snake.enterDoor = -1
        else:
            # Get it closed once we get home
            if ourSnakeIsSafe:
                #print "home now!"
                if snake.doorTimer > 0:
                    snake.doorTimer -= 1
                    if snake.doorTimer == 0:
                        snake.doorTimer = 3
                        snake.outDoor += .2
                        doorman(screen, drawnMaze, snake.outDoorNumber, snake.outDoor)
                        if snake.outDoor == 1:
                            # Got it closed, now prime for baby
                            #print "closed!"
                            snake.outDoorNumber = 2
                            snake.outDoor = -1
                            snake.fancyDoor = False
                            snake.doorTimer = 0    
    return (ourStartPause, enemyLaunchDoorPause)


def doorman(screen, drawnMaze, door, percentage):
    """ the doorman will open and close the "doors" to the pens as required
        door specified as int: 0=enemy pen, 1=our pen, 2=our out
    """
    # Set "door" coordinates based on position flag, and paint the door
    # at the given percentage (how much of the door isn't blanked out)
    if door == 0:
        doorLocation = (0, 8)
        orientation = "horizontal"
    elif door == 1:
        doorLocation = (12,8)
        orientation = "horizontal"
    elif door == 2:
        doorLocation = (10,10)
        orientation = "vertical"
    elif door == 3:
        doorLocation = (0,11)
        orientation = "horizontal"
    else:
        doorLocation = (10,11)
        orientation = "horizontal"
    # Get the image of what the door looks like before we paint over it;
    # a little hokey but I guess that's the nature of evolving code.
    paintedDoor = MazeWallSprite(doorLocation, orientation, (0,255,0), Global.wallWidth, None)
    image = paintedDoor.getDoorImage(drawnMaze, Global.wallWidth)
    # Now paint it closed, and over that in black the part that is open
    wall = MazeWallSprite(doorLocation, orientation, (0,255,0), Global.wallWidth, image)
    wall.paint(screen, Global.wallWidth)
    if percentage < 1:
        wall.paintPortion(screen, Global.wallWidth, percentage, (0,0,0))


def unleash_snakes(snake, enemies, screen, drawnMaze, lives):
    snake.coilCount = 0
    snake.driveToStart = False
    snake.countdown = 20
    snake.eggOdds = snake.origEggOdds
    Mainline.write_life_count(screen, lives)
    # If we have life indicators update the screen copy so the frog won't trash
    if lives > 1:
        maskX = int(11.5*Global.cellSize+Global.offsetX+Global.fudgeX)
        maskY = int(9*Global.cellSize+Global.offsetY+Global.fudgeY)
        remaskRect = pygame.Rect(maskX, maskY, int(Global.cellSize*1.05), int(Global.cellSize*1.55))
        drawnMaze.blit(screen, remaskRect, remaskRect)
        #pygame.draw.rect(screen, [0,255,0], remaskRect, 1)
        maskX = int(9.5*Global.cellSize+Global.offsetX+Global.fudgeX)
        maskY = int(9.5*Global.cellSize+Global.offsetY+Global.fudgeY)
        maskHeight = Global.mazeBottom - maskY - Global.wallWidth
        remaskRect = pygame.Rect(maskX, maskY, int(Global.cellSize*3.1), maskHeight)
        drawnMaze.blit(screen, remaskRect, remaskRect) 
        #pygame.draw.rect(screen, [0,255,0], remaskRect, 1)
    return drawnMaze


def init_snakes(screen,
                graph,
                limitFrames,
                targetFPS,
                ourStartPause,
                winningSnake,
                startedWinningSnakeLength,
                enemyLengths,
                snakeVelocity,
                enemyVelocity,
                spriteSheet, 
                dramaticEntry,
                lives 
                ):
    #print "INIT SNAKES", winningSnake
    if limitFrames > 0:
        ourLength = theirLength = 6
        if limitFrames > 200: # we must be longer in racetrack speed testing
            ourLength = 6
            theirLength = 5
        snake = make_snake(graph, snakeVelocity, False, countdown = 0, newSnakeLength=ourLength, spriteSheet = spriteSheet)
        snake.eggOdds = snake.origEggOdds = 0
        snake.driveToStart = False
        enemies = []
        enemy = make_snake(graph, enemyVelocity, True, 0, 0, 0, newSnakeLength=theirLength,  spriteSheet = spriteSheet)
        enemy.debuggerName = 0
        enemy.eggOdds = 0
        enemies.append([enemy, False, 0, 99])
        enemy = make_snake(graph, enemyVelocity, True, 12, 0, 0, newSnakeLength=theirLength, spriteSheet = spriteSheet)
        enemy.debuggerName = 1
        enemy.eggOdds = 0
        enemies.append([enemy, False, 1, 99])
        enemy = make_snake(graph, enemyVelocity, True, 0, 10, 0, newSnakeLength=theirLength, spriteSheet = spriteSheet)
        enemies.append([enemy, False, 2, 99])
        enemy.debuggerName = 2
        enemy.eggOdds = 0
        lastEnemy = 2
    else:
        startTime = ourStartPause
        if winningSnake is not None or startedWinningSnakeLength > 0:
            staticIndicators = lives - 1
            if staticIndicators > 2:
                staticIndicators = 2
            for i in range(staticIndicators):
                if i == 0:
                    snake = make_snake(graph,
                                   snakeVelocity,
                                   False,
                                   startX = 12.0,
                                   startY = 9.37,
                                   countdown = 0,
                                   newSnakeLength = 1,
                                   spriteSheet = spriteSheet)
                    snake.dir = pygame.K_UP
                else:
                    snake = make_snake(graph,
                                   snakeVelocity,
                                   False,
                                   startX = 11.33,
                                   startY = 10.0,
                                   countdown = 0,
                                   newSnakeLength = 1,
                                   spriteSheet = spriteSheet)
                    snake.dir = pygame.K_RIGHT
                snake.set_sprite(snake.dir)
                snake.render(screen)
                if i == 0:
                    snake.rect.top += Global.snakeSize
                else:
                    snake.rect.left -= Global.snakeSize
                screen.blit(snake.tailImage, [snake.rect.left, snake.rect.top])
            Mainline.write_life_count(screen, lives)
            if winningSnake is not None:
                snake = winningSnake
                snake.velocity = snakeVelocity
                snake.init_snake_per_level(startTime, False)
                snake.body[snake.length-1].fancyDoor = False
                snake.body[snake.length-1].isTail = True
            else:
                snake = make_snake(graph, 
                                   snakeVelocity, 
                                   False, 
                                   12, 
                                   8, 
                                   countdown = startTime,
                                   newSnakeLength=startedWinningSnakeLength,
                                   spriteSheet = spriteSheet)
                snake.driveToStart = False
        else:
            # Set location here so we don't have to always test with the
            # dramatic entry feature enabled...
            if dramaticEntry:
                startX = 10
                startY = 11
                startTime = ourStartPause
            else:
                startX = 12
                startY = 8
                startTime = 20
            snake = make_snake(graph, 
                               snakeVelocity, 
                               False, 
                               startX, 
                               startY, 
                               countdown = startTime, 
                               spriteSheet = spriteSheet)
            snake.driveToStart = dramaticEntry
        enemies = []
        eLen = snake.length + 2
        if eLen > 6:
            eLen = 6

        if eLen < 3:
            eLen = 3
        #eLen = 1
        # First enemy waits 10 frames (after our entry is complete) their door will open
        enemy = make_snake(graph, 
                           enemyVelocity, 
                           True, 
                           0, 
                           11, 
                           10, 
                           newSnakeLength = eLen, 
                           spriteSheet = spriteSheet, 
                           oldLen=enemyLengths[0])
        enemy.debuggerName = 0
        enemies.append([enemy, False, 0, 99])
        nextStart = startTime + int(120 * (targetFPS / 60.0))
        enemy = make_snake(graph, 
                           enemyVelocity, 
                           True, 
                           0, 
                           11, 
                           nextStart, 
                           newSnakeLength = eLen, 
                           spriteSheet = spriteSheet, 
                           oldLen=enemyLengths[1])
        enemy.debuggerName = 1
        enemies.append([enemy, False, 1, 99])
        nextStart = startTime + int(260 * (targetFPS / 60.0))
        # provide a bonus for starting at max length: the new third enemy 
        # will be green not orange
        if snake.length == 6:
            eLen = 5        
        enemy = make_snake(graph, 
                           enemyVelocity, 
                           True, 
                           0, 
                           11, 
                           nextStart, 
                           newSnakeLength = eLen, 
                           spriteSheet = spriteSheet, 
                           oldLen=enemyLengths[2])
        enemy.debuggerName = 2
        enemies.append([enemy, False, 2, 99])
        lastEnemy = 2
        if dramaticEntry:
            snake.eggOdds = 0
    return (snake, enemies, lastEnemy)


def make_snake(graph, 
               initVelocity, 
               makeEnemy, 
               startX=12, 
               startY=8, 
               countdown=0, 
               facing=None, 
               newSnakeLength=None, 
               spriteSheet=None,
               oldLen=None):
    if makeEnemy:
        #print "new enemy at", startX, startY
        snake = Esnake(graph, 
                       initVelocity, 
                       color=(255,0,0), 
                       x=startX, 
                       y=startY, 
                       countdown=countdown, 
                       spriteSheet = spriteSheet)
        if facing is not None:
            snake.dir = facing
    else:
        snake = Snake(graph, 
                      initVelocity, 
                      color=(0,0,255), 
                      x=startX, 
                      y=startY, 
                      countdown=countdown, 
                      spriteSheet = spriteSheet)
    if newSnakeLength is None:
        newSnakeLength = snake.length
    if oldLen is not None:
        newSnakeLength = oldLen
    snake.length = newSnakeLength
    for i in range(snake.length):
        if i == 0:
            segmentHead = snake
        else:
            segmentHead = snake.body[i-1]
        segment = Segment([startX,startY], segmentHead, spriteSheet=spriteSheet)
        segment.debuggerName = snake.debuggerName
        # Pretty sure this was deemed hopeless, but whatever... I had to fix
        # direction later anyway
        if facing is not None:
            segment.dir = facing
            segment.sprite = "wrong"
            segment.set_sprite(facing)
        snake.add_segment(segment)
    snake.body[snake.length - 1].isTail = True
    return snake


def grow_snake(snake, spriteSheet):
    # Insert new at the head, much better since we have a tail graphic
    segmentHead = snake.tail
    ripple = snake.tail.histLocation + 1 # why +1? tested out that way.
    sameTailAsBefore = segmentHead.tail
    isMoving = segmentHead.moving
    segment = Segment([0,0], segmentHead, segmentHead, spriteSheet)
    segment.debuggerName = snake.debuggerName    
    segment.histLocation = segmentHead.histLocation
    segment.moving = isMoving # the rare chance frog hops on unlaunched head
    segment.head = snake
    segment.tail = segmentHead
    segment.identity = segment.identity + 100 # so it's unique in debug   
    snake.tail = segmentHead.head = segment
    segmentHead.tail = sameTailAsBefore
    # Each subsequent history location needs to "ripple" and increase until
    # it is pointing to where the segment's tail was before the add.
    for i in range (snake.length):
        snake.body[i].histRipple += ripple
    snake.body.insert(0, segment)
    snake.length += 1


def fequal(a, b):
    return abs(a-b) < 1E-6


def determine_optimal_FPS(screen,
                          fullscreen, 
                          ourStartPause,
                          musicVolumeController,
                          effectsVolumeController):
    """ We hope to run at 60 FPS but let's see what this baby can do,
        run 200 frames and get the average FPS for the last 100 and call
        that the max performance for this machine at the res and FS """
    # Force-test slow computer calcs:
    #return 60

    if fullscreen:
        fsSwitch = "fullscreen"
    else:
        fsSwitch = "windowed"
    # Note used "method 0" in case something better comes later
    cfgKey = "%dx%d-%s-0" % (Global.screenResolution[0], 
                           Global.screenResolution[1], 
                           fsSwitch)
    config = get_config()
    if config.has_key(cfgKey):
        frameAverage = config[cfgKey]
        if not Global.productionMode:
            print "Using previous result: %d average FPS at %s" \
            % (frameAverage, cfgKey)
    else:
        targetFPS = 60
        calc_optimal_speeds(targetFPS)
        Global.startSpeed = Global.maxSpeed
        (frameAverage, 
         fullscreen,
         selectedReplay) = play_the_game(screen,
                               fullscreen,
                               ourStartPause,
                               None,
                               200,
                               targetFPS,
                               musicVolumeController,
                               effectsVolumeController)
        if frameAverage > 60:
            # sometimes I'd get 62 FPS from a test using 60, queer?
            frameAverage = 60
        # If they can't do 30 FPS it's not gonna work... but we'll let
        # them decide that. I'm not tuning anything for less.
        if frameAverage < 30:
            frameAverage = 30
        # Save our result, assume it's true every time
        config[cfgKey] = frameAverage
        put_config(config)
        if not Global.productionMode:
            print "Test result: %d average FPS at %s" \
            % (frameAverage, cfgKey)
    return frameAverage


def calc_optimal_speeds(targetFPS):
    """ Set start and max speeds per frame rate.
        Does not yield quite what I need but close enough for now """
    targetPixPerSecond = 4.5 * Global.snakeSize
    targetStartSpeed = 1 / (targetPixPerSecond / targetFPS / Global.cellSize)
    Global.startSpeed = int(round(targetStartSpeed))
    if not Global.productionMode:
        print "target FPS=", targetFPS
        print "start pps=", targetPixPerSecond, "speed=", targetStartSpeed, "int=", Global.startSpeed
    targetPixPerSecond = 5.5 * Global.snakeSize
    targetMaxSpeed = 1 / (targetPixPerSecond / targetFPS / Global.cellSize)
    Global.maxSpeed = int(round(targetMaxSpeed))
    if not Global.productionMode:
        print "max pps=", targetPixPerSecond, "speed=", targetMaxSpeed, "int=", Global.maxSpeed
    # Expect speeds like 27 down to 22 for 60FPS and only 13 down to 11
    # for 30FPS, not very many steps at low FPS as coded now.


def calc_handicapped_speeds(snakeVelocity, handicap, level):
    """ Set our speed based on the player's  handicap, and their speed
        based on our speed and level. They will be quite slow at level 1
        and full-tilt fun at level 6, just a bit slower than we are.
        (As written today speed isn't granular enough to really tune their
        speed very well, gotta work on that). """
    snakeVelocity -= handicap
    enemyHandicap = 9 - level
    if enemyHandicap < 3:
        # so enemyV=snakeV+3 at level 6
        enemyHandicap = 3 
    enemyVelocity = snakeVelocity + enemyHandicap
    #enemyVelocity = 99
    snakePPF = 1.0 / snakeVelocity * Global.cellSize
    enemyPPF = 1.0 / enemyVelocity * Global.cellSize        
    return (snakeVelocity, enemyVelocity, snakePPF, enemyPPF)


def get_data_dir(subpath = None):
    """ Get the folder for our data. On Mac OS X put everything in the
        documents directory. Unix, the way seems to be hide it in a dot
        folder in $HOME. On WIndows nobody cares, until it's time to
        uninstall the program, so put it in the same place as the program.
        
        An optional subfolder argument can be provided, for recordings.
    """
    # props Gummbum's Trolls Outta Luckland for pointing the way
    HOME_DIR = os.path.expanduser('~')
    if sys.platform == "darwin": # mac style
        USER_DIR = os.path.join(HOME_DIR, 'Documents')
        DATA_DIR = os.path.join(USER_DIR, 'pythenteneData')
    elif sys.platform != "win32": # presume Unix if not OS X or Windows
        DATA_DIR = os.path.join(HOME_DIR, '.pythentene')
    else: # for Windows we pile it all in pgms dir for easy uninstall
        DATA_DIR = os.path.join('.', 'data')
    if subpath is not None:
        DATA_DIR = os.path.join(DATA_DIR, subpath)
    # Init the user's application directory for scores and settings.
    os.umask(0022)        
    if not os.path.isdir(DATA_DIR):
        os.makedirs(DATA_DIR)
    return DATA_DIR


def put_config(config):
    CONFIG_DIR = get_data_dir()
    SETTINGS_FILE = os.path.join(CONFIG_DIR, 'settings.dat')    
    output = open(SETTINGS_FILE, 'w')
    cerealizer.dump(config, output)
    output.close()
    
    
def get_config():
    CONFIG_DIR = get_data_dir()
    SETTINGS_FILE = os.path.join(CONFIG_DIR, 'settings.dat')    
    try:
        pkl_file = open(SETTINGS_FILE, 'r')
        config = cerealizer.load(pkl_file)
        pkl_file.close()  
    except IOError:
        # Make a new config dictionary and store it. For now we can guess
        # just a few values: fullscreen = false and the starting volume
        # levels, and we'll assume there isn't a computer in the world
        # that can't do 640x480.
        #
        # Will add more at the point they're needed (and determined).
        config = {}
        config['resolutionFlag'] = 0
        config['fullscreen'] = False
        config['musicVolume'] = 20
        config['effectsVolume'] = 50
        config['joystickOn'] = False
        put_config(config)
    return config


def put_demo_recording(level, frameHistory):
    DEMO_DIR = get_data_dir('recordings')
    now = datetime.datetime.now()
    filename = "level_%d_%s.demo" % (level, now.strftime("%Y%m%d.%H%M%S"))
    DEMO_FILE = os.path.join(DEMO_DIR, filename)    
##    print "writing a record of", len(frameHistory), "frames to", DEMO_FILE
##    debugHist = list(frameHistory)
##    demoFrame = 0
##    while demoFrame != -1:
##        try:
##            nextRecording = debugHist.pop()
##            demoFrame = nextRecording[0]
##            demoMove = nextRecording[1]
##            print demoFrame,  demoMove
##        except IndexError:
##            demoFrame = -1
    # write demo to file
    output = open(DEMO_FILE, 'wb')
    cerealizer.dump(frameHistory, output)
    output.close()


def get_demo_recording(filename, fullPathProvided=False):
    """ Get the demo to playback. Could be from command-line, if so the
        full path to file will be provided. Normally we'd look up the
        program data directory (per OS) """
    if not fullPathProvided:
        DEMO_DIR = get_data_dir('recordings')
        DEMO_FILE = os.path.join(DEMO_DIR, filename)    
    else:
        DEMO_FILE = filename
    frameRecordFromFile = None
    try:
        pkl_file = open(DEMO_FILE, 'rb')
        frameRecordFromFile = cerealizer.load(pkl_file)
        pkl_file.close()  
    except IOError:                        
        #print "no recording to demo!"
        pass
    return frameRecordFromFile


def get_handicap(fullscreen):
    if fullscreen:
        fsSwitch = "fullscreen"
    else:
        fsSwitch = "windowed"
    # Note used "method 0" in case something better comes later
    handicapCfgKey = "handicap-%dx%d-%s-0" % (Global.screenResolution[0], 
                           Global.screenResolution[1], 
                           fsSwitch)
    config = get_config()
    if config.has_key(handicapCfgKey):
        handicap = config[handicapCfgKey]
        if not Global.productionMode:
            print "Using speed handicap from config=%d" % handicap
    else:
        handicap = 0
        config[handicapCfgKey] = handicap
        put_config(config)
    return handicap


def put_handicap(fullscreen, handicap):
    if fullscreen:
        fsSwitch = "fullscreen"
    else:
        fsSwitch = "windowed"
    # Note used "method 0" in case something better comes later
    handicapCfgKey = "handicap-%dx%d-%s-0" % (Global.screenResolution[0], 
                           Global.screenResolution[1], 
                           fsSwitch)
    config = get_config()
    config[handicapCfgKey] = handicap
    put_config(config)    


def render_fancy_pause_screen(screen, 
                              fullscreen, 
                              currentResolution, 
                              newResolution,
                              musicControl = None,
                              effectsControl = None,
                              doingDemo=None):
    cover = pygame.Surface(screen.get_size())
    cover.fill((0,0,0))
    cover_alpha = 200
    cover.set_alpha(cover_alpha)
    screen.blit(cover,(0,0))    

    if doingDemo:
        text = "PLAYBACK PAUSED"
    else:
        text = "PAUSED"
    fontFile = os.path.join("fonts", "cafe_nero_m54.ttf")
    if Global.res != 3:
        if Global.res == 2: # more of the same BS
            font = pygame.font.Font(fontFile, Global.topFontSize * 2)
        else:
            font = pygame.font.Font(fontFile, Global.topFontSize)
    else:
        font = pygame.font.Font(fontFile, 40) # now really rethinking this
    textOut = font.render(text, 1, (0,0,0))
    centerX = int((Global.screenResolution[0] - textOut.get_width()) / 2)
    centerY = int((Global.screenResolution[1] - textOut.get_height()) / 2)
    niceBox = int(Global.topFontSize / 2)
    pygame.draw.rect(screen, (255,255,255), [centerX - niceBox, centerY - niceBox, textOut.get_width() + niceBox*2, textOut.get_height() + niceBox*2], 0)
    screen.blit(textOut, (centerX, centerY))
    
    if doingDemo:
        text = "<down> ends demo and restarts game <space> continues playing demo <esc> quits"
    else:
        if fullscreen:
            text = "<up> graphics <i> instructions <r> replays <h> highs <c> credits <esc> quits <space> "
        else:
            text = "<up> graphics <i> instructions <r> replays <h> highs <c> credits <esc> quits <space> "
        if Global.resChange:
            text += "restarts"
        else:
            text += "resumes"
    fontFile = os.path.join("fonts", "white_rabbit.ttf")
    if Global.res  == 2:
        subFontSize = int(Global.bottomFontSize*1.25)        
    elif Global.res == 3:
        subFontSize = int(Global.bottomFontSize*1.5)
    else:
        subFontSize = int(Global.bottomFontSize)
    font = pygame.font.Font(fontFile, subFontSize)
    subTextOut = font.render(text, 1, (0,0,0))
    centerX = int((Global.screenResolution[0] - subTextOut.get_width()) / 2)
    Y = (centerY - niceBox) + (textOut.get_height() + niceBox*2) + Global.topFontSize
    niceBox = int(subFontSize / 2)
    pygame.draw.rect(screen, (255,255,255), [centerX - niceBox, Y - niceBox, subTextOut.get_width() + niceBox*2, subTextOut.get_height() + niceBox*2], 0)
    screen.blit(subTextOut, (centerX, Y))
    
    volumeText = "Set music and effect volume with the mouse:"
    volumeTextOut = font.render(volumeText, 1, (0,0,0))
    # For the controls box, use same good X centering as previous box, and 
    # start upper at where the previous box started, plus height, plus 10 px
    volumeBoxX = centerX - niceBox
    volumeBoxY = (Y - niceBox) + (subTextOut.get_height() + niceBox*2) + 10
    volumeBoxWidth = subTextOut.get_width() + niceBox*2
    pygame.draw.rect(screen, (255,255,255), [volumeBoxX, volumeBoxY, volumeBoxWidth, (subTextOut.get_height() + niceBox*2) * 3], 0)
    # The instruction text goes in our controls box, 10 px in from left and
    # down 3x the font size
    screen.blit(volumeTextOut, (volumeBoxX + 10, volumeBoxY + niceBox))
    if musicControl is not None:
        # So now it's obvious isn't it? Indent the first 10 from the left but
        # also add 50 for the icon. Down 4 x font size. The 2 controls should
        # each be 1/3rd the width of the box, with a nice 30 px seperator.
        # Oh and we want those centered horizontally, of course.
        # ... at which point it tested close enough to move on!
        controlWidth = (subTextOut.get_width() + niceBox*2) / 3
        controlCenter = (volumeBoxWidth - (controlWidth + 20) * 2 - 30 - 100) / 2
        musicControl.reposition((volumeBoxX + 10 + 50 + controlCenter, volumeBoxY + niceBox*4), (controlWidth, niceBox*4))
        effectsControl.reposition((volumeBoxX + 10 + 50 + controlWidth + 50 + 30 + controlCenter, volumeBoxY + niceBox*4), (controlWidth, niceBox*4))

    # Put the version number at the bottom of the controller box, that is:
    controllerBoxBottom = volumeBoxY + (subTextOut.get_height() + niceBox*2) * 3
    versionTextOut = font.render("Version " + Global.program_version, 1, (0,0,0))
    # And at the extreme right, so:
    rightJustified = volumeBoxX + volumeBoxWidth - versionTextOut.get_width() - 5
    screen.blit(versionTextOut, (rightJustified, controllerBoxBottom - versionTextOut.get_height()))
    return screen.copy()


def do_pause_screen(screen, 
                    fullscreen, 
                    musicVolumeController, 
                    effectsVolumeController, 
                    testEffect,
                    joystick,
                    demo):
    replay = None
    # In case they fiddle with resolution or FS and then cancel:
    currentResolution = newResolution = Global.res
    currentStartSpeed = Global.startSpeed
    currentMaxSpeed = Global.maxSpeed
    currentFullscreen = fullscreen
    screenCopy = screen.copy()
    pauseCopy = render_fancy_pause_screen(screen, 
                                          fullscreen, 
                                          currentResolution, 
                                          newResolution, 
                                          musicVolumeController, 
                                          effectsVolumeController, 
                                          demo)
    console = pyconsole.Console(
        screen, # my screen
        (10, 10, Global.screenResolution[0]-20, Global.screenResolution[1]-20),
        functions={"displayFPS":pyconsole_functions.displayFPS, 
                   "joystick":pyconsole_functions.joystick}, # functions for the console
        key_calls={"d":sys.exit}, # ctrl+d calls sys.exit()
        syntax={pyconsole_syntax.re_add:pyconsole_syntax.console_add, 
                pyconsole_syntax.re_function:pyconsole_syntax.console_func}
        )    
    paused = True
    musicVolumeControllerMouseHeldDown = False
    effectsVolumeControllerMouseHeldDown = False
    mousePos = [0, 0]
    config = get_config()
    Global.setJoystickEnabled = joystickEnabled = config['joystickOn']
    while paused:
        console.process_input()
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                if not Global.productionMode:
                    print "DEBUG: quit!"    
                sys.exit()
            elif event.type == pygame.KEYDOWN:
                if event.key == K_w and pygame.key.get_mods() & KMOD_CTRL:
                    console.set_active()
                elif event.key == K_1 and pygame.key.get_mods() & KMOD_CTRL: 
                    # "Just a little hack so you can play with both pyconsole and python"
                    # Hit ctrl w to hide pyconsole, then ctrl 1 to switch modes, then ctrl w again to show the console
                    console.setvar("python_mode", not console.getvar("python_mode"))
                    console.set_interpreter() 
                    if not Global.productionMode:
                        print "DEBUG: python mode=", console.getvar("python_mode")
                elif (pygame.key.get_mods() == 1024): # command key
                    if (event.key == 113):   # q
                        if not Global.productionMode:
                            print "DEBUG: apple-Q quit!"    
                        sys.exit()
                elif event.key == pygame.K_ESCAPE:
                    if not Global.productionMode:
                        print "DEBUG: <esc> quit!"    
                    sys.exit()
                elif event.key == pygame.K_7:
                    if not Global.productionMode:
                        failInTheMiddleOfMyGUI
                        
                elif event.key == pygame.K_i and not demo:
                    Mainline.do_instructions(screen)
                    if Global.resChange:
                        screen.fill((0,0,0))
                    else:
                        screen.blit(screenCopy, (0,0))                    
                    pauseCopy = render_fancy_pause_screen(screen, 
                                                          fullscreen, 
                                                          currentResolution, 
                                                          newResolution)
                elif event.key == pygame.K_c and not demo:
                    Mainline.do_credits(screen)
                    if Global.resChange:
                        screen.fill((0,0,0))
                    else:
                        screen.blit(screenCopy, (0,0))                    
                    pauseCopy = render_fancy_pause_screen(screen, 
                                                          fullscreen, 
                                                          currentResolution, 
                                                          newResolution)
                elif  event.key == pygame.K_r and not demo:
                    demoDirectory = get_data_dir('recordings')
                    replay = Mainline.select_replay(screen, demoDirectory)
                    if replay != "cancelled":
                        paused = False
                    else:
                        replay = None
                        if Global.resChange:
                            screen.fill((0,0,0))
                        else:
                            screen.blit(screenCopy, (0,0))                    
                        pauseCopy = render_fancy_pause_screen(screen, fullscreen, currentResolution, newResolution)
                elif event.key == pygame.K_DOWN and demo:
                    replay = "replayDone"
                    paused = False
                elif event.key == pygame.K_UP and not demo:
                    rezSelected = "notReady"
                    while rezSelected != "ready" and rezSelected != "cancel":
                        if currentResolution == newResolution and fullscreen == currentFullscreen:
                            Global.resChange = False
                        else:
                            downInMyHeart = Global.setJoystickEnabled
                            Global.init_resolution(newResolution)
                            Global.setJoystickEnabled = downInMyHeart
                            calc_optimal_speeds(60)
                            screen = Mainline.set_screen_mode(fullscreen) 
                            Global.resChange = True
                        (screen, 
                         fullscreen, 
                         rezSelected, 
                         newResolution) = Mainline.select_resolution(screen, fullscreen, newResolution)
                        #print "SELECTED NEW RES:", newResolution
                    if rezSelected == "cancel":
                        #print "cancelled, set res back and resume..."
                        newResolution = currentResolution
                        Global.init_resolution(currentResolution)
                        Global.startSpeed = currentStartSpeed
                        Global.maxSpeed = currentMaxSpeed
                        fullscreen = currentFullscreen
                        screen = Mainline.set_screen_mode(fullscreen)       
                        Global.resChange = False
                    config = get_config()
                    config['fullscreen'] = fullscreen
                    put_config(config)
                    screen.fill((0,0,0))
                    pauseCopy = render_fancy_pause_screen(screen, 
                                                          fullscreen, 
                                                          currentResolution, 
                                                          newResolution, 
                                                          musicVolumeController, 
                                                          effectsVolumeController)
                elif event.key == pygame.K_h and not demo:
                    if not Global.productionMode:
                        print "DEBUG: display high scores"
                    screen.fill((0,0,0))
                    Scores.do_scores(screen, 0, 0, get_data_dir(), True)
                    if Global.resChange:
                        screen.fill((0,0,0))
                    else:
                        screen.blit(screenCopy, (0,0))                    
                    pauseCopy = render_fancy_pause_screen(screen, 
                                                          fullscreen, 
                                                          currentResolution, 
                                                          newResolution)
                elif event.key == pygame.K_SPACE:
                    paused = False
            elif event.type == pygame.MOUSEMOTION:
                # There's so much code here... should hide it, somehow
                # Oh well I've done (much) worse
                mousePos = event.pos
                if musicVolumeControllerMouseHeldDown:
                    musicVolumeController.set_level(mousePos)
                    pygame.mixer.music.set_volume(musicVolumeController.setting/100.0) 
                if effectsVolumeControllerMouseHeldDown:
                    effectsVolumeController.set_level(mousePos)
            elif event.type == pygame.MOUSEBUTTONDOWN:
                musicVolumeControllerMouseHeldDown = \
                  musicVolumeController.mouse_is_on_control(mousePos)                    
                effectsVolumeControllerMouseHeldDown = \
                  effectsVolumeController.mouse_is_on_control(mousePos)
                if effectsVolumeControllerMouseHeldDown:
                    effectsVolumeController.set_level(mousePos)
            elif event.type == pygame.MOUSEBUTTONUP:
                if musicVolumeControllerMouseHeldDown == False:
                    if musicVolumeController.set_level_by_click(mousePos):
                        pygame.mixer.music.set_volume(musicVolumeController.setting/100.0)
                musicVolumeControllerMouseHeldDown = False
                if effectsVolumeControllerMouseHeldDown: # done sliding
                    testEffect.set_volume(effectsVolumeController.setting/100.0)
                    Noise.play_effect(testEffect)                
                else:# not sliding
                    if effectsVolumeController.set_level_by_click(mousePos):
                        testEffect.set_volume(effectsVolumeController.setting/100.0)
                        Noise.play_effect(testEffect)
                effectsVolumeControllerMouseHeldDown = False 
            elif joystick is not None:
                joystickButtons = []
                for button in range(joystick.get_numbuttons()):
                    #print button, joystick.get_button(button)
                    joystickButtons.append(joystick.get_button(button))
                for i in range(len(joystickButtons)):
                    if joystickButtons[i]:
                        paused = False
        if replay is None:        
            screen.blit(pauseCopy, (0,0))            
            musicVolumeController.paint_control()
            effectsVolumeController.paint_control()
        console.draw()    
        if Global.setJoystickEnabled != joystickEnabled:
            config = get_config()
            config['joystickOn'] = Global.setJoystickEnabled
            put_config(config)
            joystickEnabled = Global.setJoystickEnabled
        pygame.display.flip()  
    config = get_config()
    config['musicVolume'] = musicVolumeController.setting
    config['effectsVolume'] = effectsVolumeController.setting
    config['resolutionFlag'] = Global.res
    put_config(config)
    screen.blit(screenCopy, (0,0))
    return screen, fullscreen, replay



def get_next_bonus_life(score):
    """ Extra serpents are awarded for reaching 20,000 points, 50,000 points
        and every 50,000 points scored thereafter. """
    if score < 20000:
        nextBonus = 20000
    else:
        nextBonus = 50000
        while score >= nextBonus:
            nextBonus += 50000
    #print "next bonus=", nextBonus
    return nextBonus


def init_effects(effectsVolumeController):
    soundEffect = []
    soundEffect.append(Noise.load_sound('87565__SGAK__gulps-edited'))
    Noise.FROG_DINED = len(soundEffect) - 1
    soundEffect.append(Noise.load_sound('15689__reinsamba__tree_frog-edited'))
    Noise.DINED_ON_FROG = len(soundEffect) - 1
    soundEffect.append(Noise.load_sound('11871__medialint__frog_rusk_stroke-edited'))
    Noise.FROG_MOVED = len(soundEffect) - 1
    soundEffect.append(Noise.load_sound('27130__pryght-headshot'))
    Noise.BONUS_LIFE = len(soundEffect) - 1
    soundEffect.append(Noise.load_sound('21732__Anton-nibbledDead'))
    Noise.NIBBLED_DEAD = len(soundEffect) - 1
    soundEffect.append(Noise.load_sound('21732__Anton-nibbledOne'))
    Noise.NIBBLED_ONE = len(soundEffect) - 1
    soundEffect.append(Noise.load_sound('21732__Anton-multiSegNibble'))
    Noise.NIBBLED_MANY = len(soundEffect) - 1
    soundEffect.append(Noise.load_sound('71031__poots__snakey_woman-edited'))
    Noise.HEADSHOT = len(soundEffect) - 1
    soundEffect.append(Noise.load_sound('71031__poots__snakey_woman-gotMe'))
    Noise.GOTME = len(soundEffect) - 1
    soundEffect.append(Noise.load_sound('71031__poots__snakey_woman-nibbledMe'))
    Noise.NIBBLED_ME = len(soundEffect) - 1
    soundEffect.append(Noise.load_sound('82248__Robinhood76-serpentEatsEgg'))
    Noise.SNAKE_ATE_EGG = len(soundEffect) - 1
    soundEffect.append(Noise.load_sound('71031__poots__snakey_woman-newSnake'))
    Noise.SNAKE_HATCHES = len(soundEffect) - 1
    soundEffect.append(Noise.load_sound('W'))
    Noise.OBLIGATORY = len(soundEffect) - 1    
    for effect in soundEffect:
        effect.set_volume(effectsVolumeController.setting/100.0)
    return soundEffect


def play_the_level(screen, 
                    fullscreen, 
                    level, 
                    mazeLevel, 
                    lives, 
                    score, 
                    highestScore, 
                    randomSeed, 
                    frameRecord,
                    ourStartPause,
                    limitFrames,
                    targetFPS,
                    initSpeed,
                    handicap,
                    musicVolumeController,
                    effectsVolumeController,
                    winningSnake,
                    enemyLengths,
                    forceDramaticEntry,
                    startedWinningSnakeLength
                    ):
    dramaticEntry = False
    if winningSnake is None and startedWinningSnakeLength == 0:
        dramaticEntry = True
    # For debugging it makes it nice to be able to force this:
    #dramaticEntry = False        
    # ...so must get and reset when replaying a demo
    if forceDramaticEntry is not None:
        #print "force dramatic entry:", forceDramaticEntry
        dramaticEntry = forceDramaticEntry
    frameRate = targetFPS # try -1 sometime for grins
    myClock = pygame.time.Clock()
    wallSpriteSheet = pygame.image.load(os.path.join("graphics", "wallSprites.png")).convert_alpha()    
    mazeFile = os.path.join("mazes", "level" + str(mazeLevel) + ".maze")
    # I wouldn't do mazes again this way but since it's already coded...
    maze=Maze()
    intersections=[[0]*12 for x in range(14)]
    intersectionMap=[[[0]*4 for y in range(12)] for x in range(14)] 
    maze["intersection"] = intersections
    maze["intersectionMap"] = intersectionMap
    graph = maze.load(mazeFile)
    #Mainline.render_graph(screen, graph)
    if not limitFrames > 0:
        Mainline.render_maze(screen, maze, intersectionMap, level, wallSpriteSheet)
    # The graph needs invisible one-ways so the snakes can get out of the 
    # launching area and into the playfield.
    neighborList = {}
    neighborList[(12, 7)] = 1
    graph[(12, 8)] = neighborList
    neighborList = {}
    neighborList[(0, 7)] = 1
    graph[(0, 8)] = neighborList
    neighborList = {}
    neighborList[(0, 10)] = 1
    graph[(0, 11)] = neighborList
    neighborList = {}
    neighborList[(10, 10)] = 1
    graph[(10, 11)] = neighborList
    soundEffect = init_effects(effectsVolumeController)
    
    Mainline.write_footers(screen, highestScore, limitFrames, frameRecord)
    # Grab an early draft copy for fancy lead-in, alert the player when
    # testing machine speed, and get the level started in clever fashion.
    Mainline.write_headers(screen, level, False)
    Mainline.write_score(screen, score, False)
    drawnMaze = screen.copy()            
    if limitFrames > 0:
        Mainline.write_tuning_text(screen)       
    if Global.productionMode and not limitFrames > 0:
        screen.fill((0,0,0))
        for wiper in range(0, Global.screenResolution[1] + 1, 10):
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        sys.exit()
            position = pygame.Rect(0, wiper, Global.screenResolution[0], 10)                        
            screen.blit(drawnMaze, position, position)
            myClock.tick(60)
            pygame.display.flip()    
    # Set speeds based on handicap (even if zero so the enemy speed is set)
    #print "LEVEL: %d INIT SNAKE VELOCITY=%d PAUSE=%d" % (level, initSpeed, ourStartPause)
    (snakeVelocity, 
     enemyVelocity, 
     snakePixPerFrame, 
     enemyPixPerFrame) = calc_handicapped_speeds(initSpeed, handicap, level)    
    #print "PLAY THE LEVEL: Target FPS=", targetFPS, "speed=", snakeVelocity
    #print "size=", Global.cellSize, "fullscreen=", fullscreen
    #print "snake pixels per frame=", snakePixPerFrame
    spriteSheet = pygame.image.load(os.path.join("graphics", "snake" + str(Global.snakeSize) + "sheet.png")).convert_alpha()
    (snake, enemies, lastEnemy) = init_snakes(screen,
                                              graph,
                                              limitFrames,
                                              targetFPS,
                                              ourStartPause,
                                              winningSnake,
                                              startedWinningSnakeLength,
                                              enemyLengths,
                                              snakeVelocity,
                                              enemyVelocity,
                                              spriteSheet, 
                                              dramaticEntry,
                                              lives
                                             )
    nextMove = None
    weDead = False
    ourEggX = ourEggY = -1
    theirEggX = theirEggY = -1
    snakePreggers = None
    allEnemiesDead = False
    gameIsDone = False
    winner = False
    spawn = False
    ourSnakeIsSafe = False
    livesEarned = 0
    nobodyHome = 0
    selectedReplay = None
    frameReplayX = frameReplayY = 0
    aborted = 0
    needCopy = needCopyBack = didCopy = False
    lifeIndicators = 0
    motherSnake = None
    # Save a "demo" recording for review later. It needs to play at the
    # exact same speed... so will use the base speed +/- handicap.
    frameCount = 0
    frameHistory = []
    frameFPS = 0
    startWinningSnakeLength = 0
    if winningSnake is not None:
        startWinningSnakeLength = snake.length
    hist = (frameCount, (level, 
                         randomSeed, 
                         initSpeed, 
                         Global.program_version, 
                         ourStartPause, 
                         targetFPS, 
                         handicap, 
                         lives, 
                         dramaticEntry, 
                         startWinningSnakeLength, 
                         enemyLengths,
                         Global.screenResolution
                        )
           )
    frameHistory.insert(0, hist)
    # Seed the RNG, and reseed it if we're doing a demo.
    random.seed(randomSeed)
    demo =  False
    if frameRecord is not None:
        #print "DEMO!", len(frameRecord)
        assert len(frameRecord) > 2, "invalid demo file only has %d frames" % len(frameRecord)
        demo = True
        hist = frameRecord.pop()
        randomSeed = hist[1][1]
        #print "DEMO RANDOM SEED=", randomSeed
        random.seed(randomSeed)
        firstMove = frameRecord.pop()
        demoFrame = firstMove[0]
        demoMove = firstMove[1]
        nextFrameX = firstMove[3]
        nextFrameY = firstMove[4]
    nextBonusLife = get_next_bonus_life(score)
    Mainline.write_headers(screen, level, demo)
    Mainline.write_score(screen, score, demo)
    # Beautiful, save a copy (for frog erase)
    drawnMaze = screen.copy()
    # Create a frog and set the pen doors ready to open. The snake door is 
    # obvious; for the enemy side, use the last snake to launch as the 
    # controller. Our door out is closed, hope we survive to need it opened!
    # Leave the doors alone and disable the frog if doing speed calcs.
    frog = Frog([-1,-1], limitFrames)
    enemyLaunchDoorPause = 30 # needs node check like the other doors
    if not limitFrames > 0:
        snake.launchDoor = -1
        enemies[lastEnemy][0].launchDoor = -1
        enemies[lastEnemy][0].enterDoor = -1
        snake.outDoor = 1
        snake.enterDoor = -1
        if not dramaticEntry:
            snake.enterDoor = 1
    snake.direct(pygame.K_UP, graph, False)
    pygame.event.set_blocked(pygame.MOUSEMOTION)
    
    # Joysticks: not too hard, and seems to work, but I need more hardware
    # to test. Could try to fake a dev environment but what fun is that.
    # For now it's a hidden option (from console).
    config = get_config()
    joystickEnabled = config['joystickOn']
    joystick = None
    joystickHat = None
    joystickVertAxis = None
    joystickHorzAxis = None
    if joystickEnabled:
        if pygame.joystick.get_count() > 0:
            if not Global.productionMode:
                print "found a joystick,",
            joystick = pygame.joystick.Joystick(0)
            joystick.init()
            if joystick.get_numhats() > 0:
                if not Global.productionMode:
                    print "found a hat,",
                joystickHat = 0
            if joystick.get_numaxes() > 0:
                if not Global.productionMode:
                    print "found %d axes," % joystick.get_numaxes(),
                if joystick.get_numaxes() > 2:
                    # two sticks found, we'll use my right-hand joystick
                    joystickVertAxis = 2
                    joystickHorzAxis = 3
                    if not Global.productionMode:
                        print "will use right-hand axes,",
                else:
                    if not Global.productionMode:
                        print "will use only axes,",
                    joystickVertAxis = 1
                    joystickHorzAxis = 0
            if not Global.productionMode:
                print "joystick init complete"
    while not gameIsDone:
        #print "debugger frame=", frameCount
        frameCount += 1
        pausing = False
        pygame.event.pump()
        joyLeft = joyRight = joyUp = joyDown = joyPause = False
        if joystick is not None:
            joystickButtons = []
            for button in range(joystick.get_numbuttons()):
                joystickButtons.append(joystick.get_button(button))
            # This approach makes enough sense; pressing 2 buttons at once 
            # in this game does not.
            if joystickButtons[0]:
                joyDown = True
            elif joystickButtons[1]:
                joyRight = True
            elif joystickButtons[2]:
                joyLeft = True
            elif joystickButtons[3]:
                joyUp = True
            elif joystickButtons[10]:
                joyPause = True
        if joystickHat is not None:
            # My hat sucks, and this code doesn't help... it doesn't snap
            # back to zero but snaps to opposite sometimes. Anyway this seems
            # solid enough, if a little complicated:
            hatPos = joystick.get_hat(joystickHat)
            if snake.dir == pygame.K_UP or snake.dir == pygame.K_DOWN:
                if hatPos[0] == 1:
                    joyRight = True
                elif hatPos[0] == -1:
                    joyLeft = True
            else:
                if hatPos[1] == 1:
                    joyUp = True
                elif hatPos[1] == -1:
                    joyDown = True
            if snake.dir == pygame.K_UP and not joyRight and not joyLeft:
                if hatPos[1] == -1:
                    joyDown = True
            if snake.dir == pygame.K_DOWN and not joyRight and not joyLeft:
                if hatPos[1] == 1:
                    joyUp = True
            if snake.dir == pygame.K_LEFT and not joyUp and not joyDown:
                if hatPos[0] == 1:
                    joyRight = True
            if snake.dir == pygame.K_RIGHT and not joyUp and not joyDown:
                if hatPos[0] == -1:
                    joyLeft = True
        if joystickHorzAxis is not None:
            joyHorz = joystick.get_axis(joystickHorzAxis)
            joyVert = joystick.get_axis(joystickVertAxis)
            if joyHorz > .75:
                joyRight = True
            elif joyHorz < -.75:
                joyLeft = True
            if joyVert > .75:
                joyDown = True
            elif joyVert < -.75:
                joyUp = True
        if demo:
            if frameCount == demoFrame:
                # Make the same move as before, and try for another demo
                if demoMove == -1 or demoMove == 1:
                    (snake.velocity, 
                     enemyVelocity, 
                     snakePixPerFrame, 
                     enemyPixPerFrame) = calc_handicapped_speeds(snake.velocity, demoMove, level)
                    for e in enemies:
                        enemy = e[0]
                        enemy.velocity = enemyVelocity                     
                else:
                    # I had real trouble more than once with the demo replay
                    # not matching the actual level actions. Now if this
                    # happens we'll be sure to know... at the cost of bigger
                    # replay files. TODO: some day take this out and don't
                    # save nextFrame data in the file, when sure it is fixed.
                    ##print "replaying %d %s %f %f = %f %f" % (
                    ##   frameCount, pygame.key.name(demoMove), snake.locationX, 
                    ##   snake.locationY, nextFrameX, nextFrameY)
                    assert (nextFrameX == snake.locationX and 
                            nextFrameY == snake.locationY), \
                       "out of sync frame %d %f %f <> %f %f" % (frameCount, 
                        snake.locationX, snake.locationY, nextFrameX, nextFrameY)
                    nextMove = snake.direct(demoMove, graph, allEnemiesDead)
                try:
                    nextRecording = frameRecord.pop()
                    demoFrame = nextRecording[0]
                    demoMove = nextRecording[1]
                    nextFrameX = nextRecording[3]
                    nextFrameY = nextRecording[4]
                except IndexError:
                    demoFrame = -1
        # "Racetrack" debugger mode:        
        if limitFrames > 0:
            if allEnemiesDead:
                print "end on frame", frameCount - 1, "so about", (frameCount - 1)/targetFPS, "seconds"
                gameIsDone = True
            if abs(snake.speedX) + abs(snake.speedY) == 0:
                if snake.dir == pygame.K_UP:
                    keepRacing = pygame.K_LEFT
                elif snake.dir == pygame.K_LEFT:
                    keepRacing = pygame.K_DOWN
                elif snake.dir == pygame.K_DOWN:
                    keepRacing = pygame.K_RIGHT
                else:
                    keepRacing = pygame.K_UP
                nextMove = snake.direct(keepRacing, graph, allEnemiesDead, True)
        # Take direction from the player until all enemies are dead (skip if
        # doing speed calcs). Note each move for replay later. The arrows
        # serve double-duty, when demo we use keydown below
        key = pygame.key.get_pressed() # we need keys held down, not events
        frameMove = None
        if key[pygame.K_DOWN] or key[pygame.K_s] or joyDown:
            if demo:
                gameIsDone = True
                aborted = 5
            else:
                frameMove = pygame.K_DOWN
        elif key[pygame.K_UP] or key[pygame.K_w] or joyUp:
            frameMove = pygame.K_UP
        elif key[pygame.K_RIGHT] or key[pygame.K_d] or joyRight:
            frameMove = pygame.K_RIGHT
        elif key[pygame.K_LEFT] or key[pygame.K_a] or joyLeft:
            frameMove = pygame.K_LEFT
        if frameMove is not None:
            frameReplayX = snake.locationX
            frameReplayY = snake.locationY
            #print "save replay %d %s at %f %f" % (frameCount, pygame.key.name(frameMove), frameReplayX, frameReplayY)            
            if not demo:
                nextMove = snake.direct(frameMove, 
                                        graph, 
                                        (allEnemiesDead or 
                                         limitFrames>0 or 
                                         snake.driveToStart)
                                       )
        else:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif joyPause:
                    pausing = True
                elif event.type == pygame.KEYDOWN:
                    if (pygame.key.get_mods() == 1024): # command key
                        if (event.key == 113):   # q
                            sys.exit()
                    elif event.key == pygame.K_ESCAPE or event.key == pygame.K_p:
                        pausing = True
                    elif event.key == pygame.K_n:
                        gameIsDone = True
                        aborted = 1
                    elif event.key == pygame.K_RIGHT and demo:
                        frameRate += 5
                    elif event.key == pygame.K_LEFT and demo:
                        frameRate -= 5
                        if frameRate < 1:
                            frameRate = 1
                    elif event.key == pygame.K_BACKSLASH and demo:
                        frameRate = -1
                    elif event.key == pygame.K_MINUS and not demo:
                        (snake.velocity, 
                         enemyVelocity, 
                         snakePixPerFrame, 
                         enemyPixPerFrame) = calc_handicapped_speeds(snake.velocity, -1, level)
                        for e in enemies:
                            enemy = e[0]
                            enemy.velocity = enemyVelocity         
                        handicap -= 1
                        #print "handicap=", handicap, "new speed=", snake.velocity    
                        hist = (frameCount, -1, frameFPS)
                        frameHistory.insert(0, hist)                        
                        put_handicap(fullscreen, handicap)
                    elif event.key == pygame.K_EQUALS and not demo:
                        #print "current", snake.velocity
                        if snake.velocity < 4:
                            print "TOO FAST"
                            Noise.play_effect(soundEffect[Noise.FROG_DINED])
                        else:
                            (snake.velocity, 
                             enemyVelocity, 
                             snakePixPerFrame, 
                             enemyPixPerFrame) = calc_handicapped_speeds(snake.velocity, +1, level)
                            for e in enemies:
                                enemy = e[0]
                                enemy.velocity = enemyVelocity         
                            handicap += 1
                            #print "handicap=", handicap, "new speed=", snake.velocity    
                            hist = (frameCount, +1, frameFPS)
                            frameHistory.insert(0, hist)                        
                            put_handicap(fullscreen, handicap)
                    elif not Global.productionMode:
                        if event.key == pygame.K_COMMA:
                            frameRate -= 5
                            if frameRate < 1:
                                frameRate = 1
                        elif event.key == pygame.K_PERIOD:
                            frameRate += 5
                            # No cap if not production? try it for awhile...
                            #if frameRate > 60:
                            #    frameRate = 60
                        elif event.key == pygame.K_g:
                            grow_snake(snake, spriteSheet)
                        elif event.key == pygame.K_x:
                            # Death from above!
                            for e in enemies:
                                enemy = e[0]
                                enemyDead = e[1]
                                if not enemyDead:                    
                                    enemy.nibbled = True
                                    break
                        elif event.key == pygame.K_3:
                            now = datetime.datetime.now()
                            screenie = "level_%d_%s.screenshot" % (level, now.strftime("%Y%m%d.%H%M%S"))
                            pygame.image.save(screen, screenie + ".png")
                        elif event.key == pygame.K_t:
                            print "SCORE CHEATER"
                            score += 10000
                            Mainline.write_score(screen, score, False)
                        elif event.key == pygame.K_w:
                            gameIsDone = True
                            winner = True
                        elif event.key == pygame.K_p:
                            print "DEBUGGER SNAKE:"
                            print snake
                        elif event.key == pygame.K_i:
                            frog.controlFrog("north")
                        elif event.key == pygame.K_k:
                            frog.controlFrog("south")
                        elif event.key == pygame.K_j:
                            frog.controlFrog("west")
                        elif event.key == pygame.K_l:
                            frog.controlFrog("east")
                        elif event.key == pygame.K_m:
                            gameIsDone = True
                            aborted = 2      
        if pausing:
            if limitFrames > 0:
                sys.exit()
            else:
                pygame.event.set_allowed(pygame.MOUSEMOTION)
                pygame.mouse.set_visible(True)
                screen, fullscreen, selectedReplay = do_pause_screen(screen, 
                    fullscreen,
                    musicVolumeController,
                    effectsVolumeController,
                    soundEffect[Noise.DINED_ON_FROG],
                    joystick,
                    demo)
                pygame.mouse.set_visible(False)
                pygame.event.set_blocked(pygame.MOUSEMOTION)
                for effect in soundEffect:
                    effect.set_volume(effectsVolumeController.setting/100.0)
                if Global.resChange:
                    Global.resChange = False
                    gameIsDone = True
                    aborted = 3
                elif selectedReplay is not None:
                    gameIsDone = True
                    aborted = 4
        # If it's time for the frog to change locations, play a sound. Move
        # the frog (set next) as required, then erase him and see if we
        # killed him, or they killed him.
        if frog.hop((theirEggX, theirEggY), 
                    (ourEggX, ourEggY), 
                    (allEnemiesDead or snake.driveToStart)):
            Noise.play_effect(soundEffect[Noise.FROG_MOVED])
        frog.move()
        frog.erase(screen, drawnMaze)
        frogExpired = False
        if snake.check_frog(frog):
            frogExpired = True
            frog.died(screen, drawnMaze)
            # Eating a frog scores 500 points and adds one new segment.
            if snake.length < 6:
                grow_snake(snake, spriteSheet)
            score += 500
            Mainline.write_score(screen, score, demo)
        for e in enemies:
            enemy = e[0]
            enemyDead = e[1]
            if not enemyDead:                    
                if enemy.check_frog(frog):
                    frogExpired = True
                    frog.died(screen, drawnMaze)
                    if enemy.length < 6:
                        grow_snake(enemy, spriteSheet)
        if frogExpired:
            Noise.play_effect(soundEffect[Noise.DINED_ON_FROG])
        if nextMove is not None:
            nextMove = snake.direct(nextMove, graph, allEnemiesDead)
        for e in enemies:
            enemy = e[0]
            enemyDead = e[1]
            if not enemyDead:
                playerLengthAndIdleSwitch = snake.length
                if abs(snake.speedX) + abs(snake.speedY) == 0:
                    nobodyHome += 1
                else:
                    nobodyHome = 0
                # Don't let the enemies appear stupid
                if nobodyHome > targetFPS * 7: # b0n3b0y says seven seconds
                    playerLengthAndIdleSwitch = snake.length * -1
                enemy.move(graph, 
                           screen,
                           frog, 
                           enemies, 
                           (snake.goalX, snake.goalY, playerLengthAndIdleSwitch))
                alreadyRippling = False
                for i in range(enemy.length):
                    if enemy.body[i].histRipple > 0:
                        alreadyRippling = True                
                for i in range(enemy.length):
                    needRipple = enemy.body[i].locate(enemy, enemyPixPerFrame, alreadyRippling)
                    if needRipple < 0:
                        alreadyRippling = True
                        for j in range (i, enemy.length):
                            enemy.body[j].histRipple += 1
                    elif needRipple > int(round(enemyPixPerFrame)):
                        enemy.body[i].histLocation -= 1                            
        if not weDead:
            # Since we're not dead move to the next location. Use autopilot
            # to get home if all the enemies have been beaten. If we do move
            # on autopilot, upon arriving home if we have an egg on the
            # field, hatch it by making a new short snake who has a goal of
            # just barely getting into the safe pen; some coiling stuff I
            # abandoned; no chance of the baby making an egg; set the flag
            # so we add an extra life at the end of this; ring the bell and
            # get the door shut that was left open by the mother, it will
            # then prime for the baby to open it again. If no egg we're
            # just done.
            #
            # When we win, cut a path home:
            if allEnemiesDead and snake.homeGoal == [12, 8]:
                #print "get us home", snake.homeGoal
                neighborList = graph[(12, 7)]
                neighborList[(12, 8)] = 1
                graph[(12, 7)] = neighborList            
            gotHome = snake.move(nextMove, 
                                 graph, 
                                 screen, 
                                 autoPilot=(allEnemiesDead and limitFrames == 0)) # screen only for debugging
            if snake.homeGoal == [12, 9]:
                if ((lifeIndicators == 1 and fequal(snake.nextX, 12) and snake.nextY < 9.4) or 
                    (lifeIndicators == 2 and snake.nextX > 11.3 and fequal(snake.nextY, 10))):
                    #print "park snake", lifeIndicators
                    gotHome = True
            if gotHome and snake.driveToStart:
                #print "ready to start"
                gotHome = False
                if lives > 1: # remember we count extra lives
                    #print "need clever life indicators"
                    if (not needCopy and 
                        not needCopyBack and 
                        not didCopy and 
                        lifeIndicators < lives - 1 and 
                        lifeIndicators < 2):
                        needCopy = True
                    elif not needCopyBack:
                        needCopyBack = True
                else:
                    #print "one life to live"
                    # Ready to rock, since no indicators needed
                    drawnMaze = unleash_snakes(snake, enemies, screen, drawnMaze, lives)
                    ourStartPause = 20
            # on the way out to the playfield:
            if snake.driveToStart:
                if lives == 1:
                    snake.coilCount = 2
                if snake.length > 1 and lives > 1:
                    if snake.length == 4 and snake.coilCount == 0:
                        snake.coilCount = 1
                    if not snake.driveToStart or snake.length > 1:
                        if (fequal(snake.nextX, 12) and 
                            snake.nextY > 8.49 and 
                            snake.dir == pygame.K_DOWN and 
                            snake.coilCount < snake.length):
                                nextMove = snake.direct(pygame.K_UP, graph, False)
                                snake.coilCount += 1
                        elif (fequal(snake.nextX, 12) and 
                              snake.nextY < 8.1 and 
                              snake.dir == pygame.K_UP and 
                              snake.coilCount < snake.length - 1):
                            nextMove = snake.direct(pygame.K_DOWN, graph, False)
            # and on the way back:
            if ((allEnemiesDead and limitFrames == 0) and not gotHome):
                if (snake.length == 6) and snake.coilCount == 1:
                    snake.coilCount += 1
                if (fequal(snake.nextX, 12) and 
                    snake.nextY > 8.49 and 
                    snake.dir == pygame.K_DOWN and 
                    snake.coilCount < snake.length + 1):
                        nextMove = snake.direct(pygame.K_UP, graph, False)
                        snake.coilCount += 1
                        snake.offPlayfield = True
                elif (fequal(snake.nextX, 12) and 
                      snake.nextY < 8.1 and 
                      snake.dir == pygame.K_UP and 
                      snake.coilCount < snake.length 
                      and snake.offPlayfield):
                        nextMove = snake.direct(pygame.K_DOWN, graph, False)
            if allEnemiesDead and gotHome:
                if ourEggX > -1:
                    ourSnakeIsSafe = True
                    #print "winning snake got home", snake.locationX, snake.locationY
                    #print "NEW LIFE", ourEggX, ourEggY
                    motherSnake = snake
                    snake = make_snake(graph,
                                       10,
                                       False,
                                       startX = ourEggX,
                                       startY = ourEggY,
                                       countdown = 0,
                                       newSnakeLength = 1,
                                       spriteSheet = spriteSheet)
                    snakePixPerFrame = 1.0 / snake.velocity * Global.cellSize
                    neighborList = {}
                    neighborList[(10, 10)] = 1
                    graph[(9, 10)] = neighborList
                    
                    # If few lives remain just drive into the pen, else off
                    if lives < 3:
                        snake.homeGoal = [11,10]
                    else:
                        # Cut a path well off the screen (in case of unlucky frog)
                        snake.homeGoal = [10, 13]
                        neighborList = {}
                        neighborList[(10, 11)] = 1
                        graph[(10, 10)] = neighborList
                        neighborList = {}
                        neighborList[(10, 12)] = 1
                        graph[(10, 11)] = neighborList
                        neighborList = {}
                        neighborList[(10, 13)] = 1
                        graph[(10, 12)] = neighborList
                        neighborList = {}
                        neighborList[(10, 12)] = 1
                        graph[(10, 13)] = neighborList
                        # Remove the path to the launch area:
                        neighborList = graph[(12, 7)]
                        del neighborList[(12, 8)] 
                        graph[(12, 7)] = neighborList                       
                    snake.nodeBeforeOut = [9, 10]
                    # The baby never coils, so these are unused?
                    snake.coilGoal = 10
                    snake.coilCount = 0
                    snake.eggOdds = 0
                    spawn = True
                    snake.doorTimer = 3
                    #snake.outDoorNumber = 2
                    snake.launchDoor = 1
                    snake.outDoor = 0
                    snake.driveToStart = False
                    Noise.play_effect(soundEffect[Noise.SNAKE_HATCHES])
                    ourEggX = -1
                else:
                    #print "DONE"
                    gameIsDone = True
                    winner = True

        # Segments paint last to first but must calculate their screen
        # coords first to last (duh!). Go see where the segments are 
        # calculated to be painted. We might discover a big gap, or worse,
        # overlap... try to fix it up.
        if not weDead:
            # See if we're already in the middle of a ripple... if so, no
            # tweaking on this frame.
            alreadyRippling = False
            for i in range(snake.length):
                if snake.body[i].histRipple > 0:
                    alreadyRippling = True
                    break
                # Also must check for a "tangle" where the player reverses, 
                # and then reverse again... the segments will be pointed in
                # the same direction but it's a mess so don't mess with it. 
                # Only the player reverses so don't waste this time on enemy.                
                if not alreadyRippling:
                    for j in range (snake.body[snake.length-1].histLocation):
                        if snake.history[j][3]: # did backflip?
                            alreadyRippling = True
                            break
            # Calc the new segment position and determine how far off it is
            # from the perfect spot right on the head's butt. If it overlaps
            # "ripple" back all the segments one history location. Conversely
            # if it is too far from the head push it up a notch -- but only
            # when the snake is moving (if he stops we'll push into never-
            # never land, was lucky to see this in testing, very rare).
            for i in range(snake.length):
                needRipple = snake.body[i].locate(snake, snakePixPerFrame, alreadyRippling)
                if needRipple < 0:
                    alreadyRippling = True
                    #print "NEED RIPPLE"
                    for j in range (i, snake.length):
                        snake.body[j].histRipple += 1
                elif ((needRipple > int(round(snakePixPerFrame))) and 
                      (snake.speedX <> 0 or snake.speedY <> 0) and 
                      snake.countdown == 0):
                    #print "need push", i, needRipple
                    snake.body[i].histLocation -= 1
        for e in enemies:
            enemy = e[0]
            enemyDead = e[1]
            if not enemyDead and not weDead:
                # Flip colors on length change, but not for life indicators
                if not (snake.driveToStart and snake.homeGoal <> [12, 8]):
                    if enemy.color == (255,0,0) and enemy.length < snake.length:
                        enemy.flip_color()
                    elif enemy.color == (0,255,0) and enemy.length >= snake.length:
                        enemy.flip_color()
                enemyLength = enemy.length + 1 # score the head too
                enemyDead = snake.check_enemy_kill(enemy, False)
                if enemyDead:
                    # Note this very same code is needed below if we kill
                    # a snake by eating the segments... might make a routine
                    # of it but whatever.
                    if snakePreggers is not None:
                        if enemy.debuggerName == snakePreggers.debuggerName:
                            #print "head two for the price of one!"
                            # Killed the snake with the egg... if it
                            # wasn't already spawned to the playfield,
                            # zap the preggers pointer -- else leave
                            # it be until the egg hatches, when it will
                            # clear
                            if not theirEggX > -1:
                                #print "head killed the mother snake"
                                snakePreggers.eggSize = 0
                                snakePreggers = None                    
                    enemies[e[2]][1] = True
                    enemies[e[2]][3] = 99
                    
                    # Eating serpents head on scores 200 points per segment 
                    # on levels 1 and 2 and adds a new segment to your 
                    # serpent. The score per segment increases by 200 
                    # points every other level.
                    if level < 3:
                        score += enemyLength * 200
                    else:
                        score += (level - 1) * 200 * enemyLength
                    Mainline.write_score(screen, score, demo)
                    Noise.play_effect(soundEffect[Noise.HEADSHOT])
                    if snake.length < 6:
                        grow_snake(snake, spriteSheet)
                    # /Note
                        
            if not enemyDead and not weDead:
                weDead = enemy.check_enemy_kill(snake, True)
                if weDead:
                    gameIsDone = True
                    Noise.play_effect(soundEffect[Noise.GOTME])
            if snake.length == 0 or snake.nibbled:
                # Why this test? weDead should do it? shrug
                gameIsDone = True
        if limitFrames > 0 and frameCount > limitFrames:
            gameIsDone = True
                
        if not gameIsDone:
            screen.lock()
            # Erase the snakes
            for i in range (snake.length-1, -1, -1):
                snake.body[i].erase(screen)            
            snake.erase(screen)

            # If we're not dead yet, see if they ate our last segment. In
            # any case, if our length changes here forget all growing eggs.
            if not weDead:
                oldSnakeLength = snake.length
                snake.embarrassingDeath = False
                consequences = snake.check_death()
                if consequences < 0:
                    weDead = True
                    if snake.embarrassingDeath:
                        Noise.play_effect(soundEffect[Noise.OBLIGATORY])
                    else:
                        Noise.play_effect(soundEffect[Noise.GOTME])
                if snake.length != oldSnakeLength:
                    snake.eggSize = 0
                    if consequences > 0:
                        Noise.play_effect(soundEffect[Noise.NIBBLED_ME])
            for e in enemies:
                enemy = e[0]
                enemyDead = e[1]
                # Erase the live enemy snake, or if dead, erase it just once
                if not enemyDead or not enemy.erased:
                    for i in range (enemy.length-1, -1, -1):
                        enemy.body[i].erase(screen)            
                    enemy.erase(screen)
                    if enemyDead and not enemy.needDeathPainted:
                        enemy.erased = True
            screen.unlock()
            for e in enemies:
                enemy = e[0]
                enemyDead = e[1]
                if not enemyDead:
                    # Shoot... it's all good except one tricky condition:
                    # So we go see if any segments got nibbled, and if so
                    # pop them off the body, and then test to see if the
                    # result is a dead snake (no body segments or whatever).
                    # If it died and there's a pregnant snake out there, see
                    # if it was this one, if so and the egg was on the tail
                    # (and not on the field) then there's no pregnant snake
                    # and we're good, kill the snake and grow ours unless
                    # too long. FINE! but if we eat the tail of a pregnant
                    # snake we need to kill the egg too. So note the length 
                    # of the snake and if it changes and we're the pregnant
                    # snake, clear the egg.
                    oldLength = enemy.length
                    wages = enemy.check_death()
                    if wages < 0:
                        enemyDead = True
                    if enemyDead:
                        #print "DEAD SNAKE"
                        # Dupicate code from above... bad practice :(
                        if snakePreggers is not None:
                            if enemy.debuggerName == snakePreggers.debuggerName:
                                #print "two for the price of one!"
                                # Killed the snake with the egg... if it
                                # wasn't already spawned to the playfield,
                                # zap the preggers pointer -- else leave
                                # it be until the egg hatches, when it will
                                # clear
                                if not theirEggX > -1:
                                    #print "killed the mother snake"
                                    snakePreggers.eggSize = 0
                                    snakePreggers = None
                        enemies[e[2]][1] = True
                        enemies[e[2]][3] = 99
                        # Eating serpents scores 100 points per segment on 
                        # levels 1 and 2. The score per segment increases 
                        # by 100 points every other level.
                        # NOTE: here we've eaten the last segment and
                        # killed the snake, from behind... so it's a per-
                        # segment score only.
                        if level < 3:
                            score += 100 * (oldLength + 1)
                        else:
                            score += ((level - 1) * 100) * (oldLength + 1)
                        Mainline.write_score(screen, score, demo)
                        Noise.play_effect(soundEffect[Noise.NIBBLED_DEAD])
                        # No more growth on tail niblles!
                        #if snake.length < 6:
                        #    grow_snake(snake, spriteSheet)
                        # /DupeCode
                    elif oldLength <> enemy.length:
                        # Eating serpents scores 100 points per segment on 
                        # levels 1 and 2. The score per segment increases 
                        # by 100 points every other level.
                        segsEaten = oldLength - enemy.length
                        if wages > 0:
                            if level < 3:
                                score += 100 * segsEaten 
                            else:
                                score += ((level - 1) * 100) * segsEaten 
                            Mainline.write_score(screen, score, demo)
                            if segsEaten > 1:
                                Noise.play_effect(soundEffect[Noise.NIBBLED_MANY])
                            else:
                                Noise.play_effect(soundEffect[Noise.NIBBLED_ONE])
                        
                        if snakePreggers is not None and not theirEggX > -1:
                            if enemy.debuggerName == snakePreggers.debuggerName:
                                #print "nibbled the mother snake"
                                snakePreggers.eggSize = 0
                                snakePreggers = None
            foundOneWasDead = False
            foundTwoWereDead = False
            if not allEnemiesDead and not theirEggX > -1:
                foundAllWereDead = True
                for e in enemies:
                    if not e[1]:
                        foundAllWereDead = False
                    else:
                        if foundOneWasDead:
                            foundTwoWereDead = True
                        foundOneWasDead = True
                # If all the enemies are dead, set the switch so we stop
                # looking for more of them, suppress new eggs from starting,
                # and prime the door out since we'll be there soon.
                if foundAllWereDead:
                    allEnemiesDead = True
                    snake.eggOdds = 0
                    snake.outDoor = -1
                    snake.fancyDoor = False
                    snake.doorTimer = 0
            # No farming, once we're down to one snake stop the eggs.
            if foundTwoWereDead:
                snake.eggOdds = 0
                # Hmm, only one snake left. Hunting without remorse results
                # in a boring death spiral... and my previous idea to throw
                # in a rare random chance is no good. Try this: the last
                # snake is likely a goner anyway, so he always flees?
                for e in enemies:
                    enemy = e[0]
                    enemyDead = e[1]
                    if not enemyDead:
                        enemy.funFactor = 0 # was .9, now 0 -- flee or random?
            # If we haven't killed all the enemies and there's at least one
            # dead and none are marked as the spawning pregnant mother enemy
            # snake and no egg is already on the field... find one long
            # enough to spawn and mark that the spawner.
            if not allEnemiesDead:
                if foundOneWasDead and snakePreggers is None and not theirEggX > -1:
                    for e in enemies:
                        # If not dead and big enough to pop an egg and
                        # there is no egg on the field
                        if not e[1] and e[0].length > 1:
                            snakePreggers = e[0]
                            break
            # If they have a pregnant snake who hasn't dropped the egg yet,
            # mature it until it's ready to drop. Do same for ours, then
            # paint any dropped eggs (here and not later so we're over not
            # under them, looked funky, ha).            
            if snakePreggers is not None and not theirEggX > -1:
                if snakePreggers.eggSize < .9:
                    snakePreggers.grow_egg()
            if not weDead and snake.eggSize < .9 and not ourEggX > -1:
                snake.grow_egg()
            if ourEggX > -1:
                snake.paint_egg(screen, 1, ourEggX, ourEggY)            
            if theirEggX > -1:
                snakePreggers.paint_egg(screen, 1, theirEggX, theirEggY)

            if not weDead:
                for i in range (snake.length-1, -1, -1):
                    snake.body[i].render(screen)
# This plotter image was helpful when I was trying to adjust segment pixels
# but that was time wasted, it's hopeless. Need sub-pixel rendering!                    
##                    holdImage = snake.body[i].image
##                    snake.body[i].image = snake.body[i].plotter
##                    paintX = int(snake.body[i].locationX*Global.cellSize+Global.offsetX+Global.fudgeX)
##                    paintY = int(snake.body[i].locationY*Global.cellSize+Global.offsetY+Global.fudgeY)        
##                    snake.body[i].rect.center = [paintX, paintY]
##                    screen.blit(snake.body[i].image, [snake.body[i].rect.left, snake.body[i].rect.top])
##                    snake.body[i].image = holdImage
                    if i == snake.length-1:
                        lastSegX = snake.body[snake.length-1].nextX
                        lastSegY = snake.body[snake.length-1].nextY
                        if snake.eggSize > 0:
                            snake.paint_egg(screen, 
                                   snake.eggSize, 
                                   lastSegX, 
                                   lastSegY)                        
                        if snake.eggSize > .8:                        
                            if fequal(lastSegX, int(lastSegX+.005)) and fequal(lastSegY, int(lastSegY+.005)):
                                snake.eggSize = 0
                                ourEggX = int(lastSegX+.005)
                                ourEggY = int(lastSegY+.005)
                                #print "our egg=", ourEggX, ourEggY
                                snake.body[snake.length-1].nibbled = True
                                # so we don't make the "nibbled me" sound
                                snake.body[snake.length-1].spawning = True
                                # We don't get a tail segment back if it got
                                # nibbled (but enemies do, more interesting)
                                snake.body[snake.length-2].isTail = \
                                   snake.body[snake.length-1].isTail
            if not weDead:
                snake.render(screen)
                #print snake
            for e in enemies:
                enemy = e[0]
                enemyDead = e[1]
                if (not enemyDead) or (enemy.needDeathPainted):
                    if enemyDead:
                        enemy.needDeathPainted = False
                    for i in range (enemy.length-1, -1, -1):
                        enemy.body[i].render(screen)
                        if snakePreggers is not None:
                            if enemy.debuggerName == snakePreggers.debuggerName:
                                if i == enemy.length - 1:
                                    lastESegX = enemy.body[enemy.length-1].nextX
                                    lastESegY = enemy.body[enemy.length-1].nextY
                                    if snakePreggers.eggSize > 0:
                                        enemy.paint_egg(screen, 
                                                       snakePreggers.eggSize, 
                                                       lastESegX, 
                                                       lastESegY)
                                if snakePreggers.eggSize > .8:
                                    if fequal(lastESegX, int(lastESegX+.005)) and fequal(lastESegY, int(lastESegY+.005)):
                                        snakePreggers.eggSize = 0
                                        snakePreggers.hatchTimer = 0
                                        theirEggX = int(lastESegX+.005)
                                        theirEggY = int(lastESegY+.005)
                                        theirEggDir = enemy.body[i].dir
                                        enemy.body[enemy.length-1].nibbled = True
                                        # So we don't score it:
                                        enemy.body[enemy.length-1].spawning = True
                                        enemy.body[enemy.length-2].isTail = True
                    enemy.render(screen)
            if frog.state <> "waiting":
                heFeasted = False
                frog.render(screen) 
                if ourEggX > -1:
                    if snake.check_egg((ourEggX, ourEggY), (frog.atX, frog.atY)):
                        ourEggX = -1 
                        heFeasted = True
                if theirEggX > -1:
                    if snake.check_egg((theirEggX, theirEggY), (frog.atX, frog.atY)):
                        theirEggX = -1
                        heFeasted = True
                        snakePreggers.eggSize = 0
                        snakePreggers = None
                if heFeasted:
                    Noise.play_effect(soundEffect[Noise.FROG_DINED])
            if ourEggX > -1:
                for e in enemies:
                    enemy = e[0]
                    enemyDead = e[1]
                    if not enemyDead:   
                        if snake.check_egg((ourEggX, ourEggY), (enemy.locationX, enemy.locationY)):
                            ourEggX = -1
                            if enemy.length < 6:
                                grow_snake(enemy, spriteSheet)
                            Noise.play_effect(soundEffect[Noise.SNAKE_ATE_EGG])
            if theirEggX > -1:
                if snake.check_egg((theirEggX, theirEggY), (snake.locationX, snake.locationY)):
                    snakePreggers.eggSize = 0
                    snakePreggers = None
                    # Eating a spotted egg scores 150 points times the level 
                    # you are on and adds one new segment.
                    if snake.length < 6:
                        grow_snake(snake, spriteSheet)
                    score += 150 * level
                    Mainline.write_score(screen, score, demo)
                    Noise.play_effect(soundEffect[Noise.SNAKE_ATE_EGG])
                    frog.eraseAnEgg(screen, drawnMaze, theirEggX, theirEggY)
                    theirEggX = -1
            # If it is time to hatch their spawn then set the timers back
            # (for the next egg), find the snake that the new one is 
            # replacing, make a new short snake and set it moving in a valid
            # direction, then send the old snake to the bit bucket and insert
            # the new snake in its place.
            if theirEggX > -1:
                snakePreggers.hatchTimer += 1
                if snakePreggers.hatchTimer > snakePreggers.hatchTime:
                    # Reset the snake's hatch timer in case it gets
                    # pregnant again, and then clear the preggers pointer
                    # since we're hatching the spawn now
                    snakePreggers.hatchTimer = 0
                    snakePreggers = None
                    for e in enemies:
                        enemy = e[0]
                        enemyDead = e[1]
                        if enemyDead:
                            #print "replacing enemy", enemy.debuggerName
                            newEnemy = make_snake(graph,
                                                  enemyVelocity,
                                                  True, 
                                                  theirEggX, 
                                                  theirEggY, 
                                                  newSnakeLength=2, 
                                                  facing=theirEggDir, 
                                                  spriteSheet = spriteSheet)
                            newEnemy.debuggerName = enemy.debuggerName
                            # Don't fool with doors when new enemy hatches
                            newEnemy.launchDoor = newEnemy.enterDoor = 1
                            newEnemy.set_hatchling_goal(graph, 
                                                        theirEggX, 
                                                        theirEggY, 
                                                        theirEggDir)
                            # Throw out the old snake at this spot, and
                            # insert the new with the same name, not at all
                            # close to the frog (yet)
                            del enemies[newEnemy.debuggerName]
                            enemies.insert(newEnemy.debuggerName, 
                                           [newEnemy, False, 
                                           newEnemy.debuggerName, 
                                           99])
                            break
                    theirEggX = -1
                    Noise.play_effect(soundEffect[Noise.SNAKE_HATCHES])
            if snake.launching and not snake.driveToStart:
                # easy: after we set a goal to the left, can go down,
                # otherwise after we've climbed a bit it's safe.
                if snake.goalX == 11 or snake.locationY < 7: # so simple
                    #print "NOT LAUNCHING"
                    snake.launching = False
            # Need to know when all the segments get started, to cover a
            # rare bug where a new segment is growing and the snake is just
            # completing a reverse... we test for overlap to start a new
            # segment and in this one case it doesn't work. Only an issue for
            # the player since the enemies never backtrack.
            if not snake.allSegsOnPlayfield:
                snake.allSegsOnPlayfield = True
                for i in range(snake.length):
                    if not snake.body[i].moving:
                        snake.allSegsOnPlayfield = False
                        break
                if snake.allSegsOnPlayfield:
                    #print "snake fully moving"
                    pass
            # Shoot, there's another way that a new segment is growing and
            # will be mispositioned: if its head it about to turn, we'll
            # overlap for quite awhile... so we need this flag and logic
            # for enemies as well!
            for e in enemies:
                enemy = e[0]
                enemyDead = e[1]
                if not enemyDead and not enemy.allSegsOnPlayfield:
                    enemy.allSegsOnPlayfield = True
                    for i in range(enemy.length):
                        if not enemy.body[i].moving:
                            enemy.allSegsOnPlayfield = False
                            break
                    if enemy.allSegsOnPlayfield:
                        #print enemy.debuggerName, "fully moving"
                        pass
            # If we switched from our snake to a life indicator, switch it
            # back... this is surely too "fancy" and "clever" -- needs rework
            if needCopy:
                #print "so clever"
                needCopy = False
                if lifeIndicators == lives:
                    didCopy = True
                if lifeIndicators == 0:
                    snakeCopy = copy.copy(snake)
                snake = make_snake(graph,
                                   snakeVelocity,
                                   False,
                                   startX = 10,
                                   startY = 11,
                                   countdown = 0,
                                   newSnakeLength = 1,
                                   spriteSheet = spriteSheet)
                snakePixPerFrame = 1.0 / snake.velocity * Global.cellSize
                snake.direct(pygame.K_UP, graph, False)
                snake.driveToStart = True
                snake.homeGoal = [12,9]
                snake.nodeBeforeOut = [9, 10]
                snake.coilGoal = 10
                snake.coilCount = 0
                snake.eggOdds = 0
                snake.doorTimer = 3
                snake.launchDoor = 1
                snake.outDoor = 0
                snake.enterDoor = 0
                lifeIndicators += 1
            if needCopyBack:
                needCopyBack = False
                enterDoor = snake.enterDoor
                snake = snakeCopy
                snake.body[0].head = snake
                snake.enterDoor = enterDoor
                # Ready to rock, restored control to the snake we are playing
                drawnMaze = unleash_snakes(snake, enemies, screen, drawnMaze, lives)
                ourStartPause = 20
            if limitFrames == 0:
                (ourStartPause, 
                 enemyLaunchDoorPause) = door_management(screen, 
                                                         drawnMaze, 
                                                         snake, 
                                                         enemies[lastEnemy][0], 
                                                         ourStartPause, 
                                                         enemyLaunchDoorPause, 
                                                         lifeIndicators, 
                                                         lives, 
                                                         ourSnakeIsSafe)
            if score > nextBonusLife and not demo:
                livesEarned += 1
                Noise.play_effect(soundEffect[Noise.BONUS_LIFE]) 
                Mainline.write_life_count(screen, lives + livesEarned)
                nextBonusLife = get_next_bonus_life(score)
            # If a snake is driving off the screen update to mask it, then
            # save the movement on this frame (if any) for replay later.
            if aborted==0 and ((enemies[lastEnemy][0].launchDoor < 1) or 
                              (snake.homeGoal == [10, 13])):
                Mainline.mask_exit_paths(screen, drawnMaze)
            if frameMove is not None or limitFrames > 0:
                hist = (frameCount, frameMove, frameFPS, frameReplayX, frameReplayY)
                frameHistory.insert(0, hist)
        else:
            # If we died it'd be nice to see it (and animate it? someday!)
            if weDead:
                for e in enemies:
                    enemy = e[0]
                    for i in range (enemy.length-1, -1, -1):
                        enemy.body[i].erase(screen)            
                    enemy.erase(screen)
                for i in range(snake.length-1, -1, -1):
                    snake.body[i].erase(screen)            
                snake.erase(screen)
                for i in range(snake.length-1, -1, -1):
                    snake.body[i].render(screen)
                snake.render(screen)
                for e in enemies:
                    enemy = e[0]
                    enemyDead = e[1]
                    if not enemyDead:
                        for i in range (enemy.length-1, -1, -1):
                            enemy.body[i].render(screen)
                        enemy.render(screen)                    
                for deathCount in range(20):
                    for event in pygame.event.get():
                        if event.type == pygame.QUIT:
                            sys.exit()                        
                    myClock.tick(60)
                    pygame.display.flip()
            # That's all for this one... if we're out of lives then leave
            # the score on the screen
            wipeEnd = -1
            wipeLeft = 0
            wipeRight = Global.screenResolution[0]
            if (lives + livesEarned) == 1 and not spawn:
                wipeEnd = Global.clearScoreHeight
                if Global.res == 3:
                    wipeLeft = 122
            if limitFrames > 0 or aborted > 0:
                screen.fill((0,0,0))
            else:
                for wiper in range(Global.screenResolution[1] - 1, wipeEnd, -20):
                    for event in pygame.event.get():
                        if event.type == pygame.QUIT:
                            sys.exit()                 
                    pygame.draw.lines(screen, (0,0,0), False,
                     ((wipeLeft, wiper), (wipeRight, wiper)), 20)
                    myClock.tick(60)
                    pygame.display.flip()
        #print myClock.tick(frameRate)
        myClock.tick(frameRate)
        frameFPS = myClock.get_fps()
        if not Global.productionMode or Global.alwaysShowFPS:
            Mainline.write_fps(screen, frameFPS)
        pygame.display.flip()
    # We want to reuse our snake on a win... the mother snake if we spawned
    if not snake.homeGoal == [12, 8]:
        snake = motherSnake
    # If we left a shorty let's revive at that length
    enemyLengths = []
    for e in enemies:
        enemy = e[0]
        enemyDead = e[1]
        if enemyDead:
            enemyLengths.append(1)
        else:
            enemyLengths.append(enemy.length)
    enemyLengths = sorted(enemyLengths, reverse=True)
    return (screen, 
           fullscreen, 
           winner, 
           spawn, 
           score, 
           frameHistory, 
           aborted,
           selectedReplay,
           handicap,
           livesEarned,
           snake,
           enemyLengths)


def play_the_game(screen,
                  fullscreen,
                  ourStartPause,
                  frameRecordFromFile,
                  limitFrames,
                  targetFPS,
                  musicVolumeController,
                  effectsVolumeController):
    frameAverage = 0
    dataDir = get_data_dir()
    highestScore = Scores.get_highest_score_evar(dataDir)
    levelSpeed = currentSpeed = Global.startSpeed
    winningSnake = None
    enemyLengths = [None, None, None]
    didDramaticEntry = None
    startedWinningSnakeLength = 0
    if frameRecordFromFile is not None:
        getGame = list(frameRecordFromFile)
        hist = getGame.pop()
        level = mazeLevel = hist[1][0]
        levelSpeed = hist[1][2]
        ourStartPause = hist[1][4]            
        targetFPS = hist[1][5]
        handicap = hist[1][6]
        lives = hist[1][7]
        didDramaticEntry = hist[1][8]
        startedWinningSnakeLength = hist[1][9]
        enemyLengths = hist[1][10]
    else:
        lives = 3
        if limitFrames > 0:
            level = mazeLevel = 0
            handicap = 0 # never handicap the calcs
        else:
            level = mazeLevel = 1
            handicap = get_handicap(fullscreen)
    score = 0
    while lives > 0:
        screen.fill((0,0,0))
        randomSeed = random.getrandbits(32)
        # Play the replays over and over until <esc>
        if frameRecordFromFile is not None:
            frameRecord = list(frameRecordFromFile) #what was type from file?
            doDemo = True
        else:
            frameRecord = None
            doDemo = False
        (screen, 
         fullscreen, 
         win, 
         spawn, 
         score, 
         frameHistory, 
         aborted,
         selectedReplay,
         handicap,
         livesEarned,
         winningSnake,
         enemyLengths
        ) = play_the_level (
            screen, 
            fullscreen, 
            level, 
            mazeLevel, 
            lives, 
            score, 
            highestScore, 
            randomSeed,
            frameRecord,
            ourStartPause,
            limitFrames,
            targetFPS,
            levelSpeed,
            handicap,
            musicVolumeController,
            effectsVolumeController,
            winningSnake,
            enemyLengths,
            didDramaticEntry,
            startedWinningSnakeLength
        )
        lives += livesEarned
        # If we just played a demo then play it again, unless they aborted it
        if doDemo and aborted == 0:
            aborted = 4
        if aborted == 1 or limitFrames > 0:
            # aborted to start a new game
            lives = 0
        elif aborted == 2:
            lives = 0
            Global.startSpeed -= 1
            print "TEST SPEED HACK, force speed=", Global.startSpeed
        elif aborted == 3:
            # aborted because we need to recalc and restart
            lives = 0
            frameAverage = determine_optimal_FPS(screen,
                              fullscreen, 
                              ourStartPause,
                              musicVolumeController,
                              effectsVolumeController)
            calc_optimal_speeds(frameAverage)
            limitFrames = -1 # so we don't write demo file
        elif aborted == 4:
            # aborted because we want to see a replay, or just saw one
            limitFrames = -1 # so we don't write demo file
            lives = 0 # to get out
        elif aborted == 5:
            # aborted because we're bored with replay
            limitFrames = -1
            lives = 0
            selectedReplay = "replayDone"
        # If we didn't just play a recording or do a test, and if we want
        # demo recordings kept...
        if not doDemo and Global.recordLevelDemos and limitFrames == 0:
            put_demo_recording(level, frameHistory)
        elif limitFrames > 0:
            frameSum = 0
            for i in range(100):
                frameSum += frameHistory[i][2]
            frameAverage = int(round(frameSum / 100))
        if not doDemo:
            if win:
                # We want to gradually increase the speed from start to max,
                # over 5 levels, so at 60FPS from 27 at level 1 down to 22
                # at level 5. It's almost arbitrary anyway...
                speedStep = (Global.startSpeed - Global.maxSpeed) / 5.0
                currentSpeed -= speedStep
                levelSpeed = int(round(currentSpeed))
                if levelSpeed < Global.maxSpeed:
                    levelSpeed = Global.maxSpeed
                level += 1
                #print "play next level", level, "at base speed=", levelSpeed
                mazeLevel += 1
                if mazeLevel > 25:
                    mazeLevel = 1
                enemyLengths = [None, None, None]
            else:
                winningSnake = None
                lives -= 1
            if spawn:
                lives += 1
    if (not aborted > 0) and limitFrames == 0:
        # go see if they got a hi score, or just show them the heroes if not
        if Scores.do_scores(screen, score, level, get_data_dir()):
            Noise.play_music(Global.gameMusic)
    return (frameAverage, fullscreen, selectedReplay)


def do_main():
    pygame.display.set_icon(pygame.image.load(os.path.join("graphics", "gameIcon.png"))) 
    TEST_playbackRecording = ""

    config = get_config()
    fullscreen = config['fullscreen']
    
    Global.init_resolution(config['resolutionFlag'])
    #print "Python " + sys.version.split(' ')[0]
    #print "PyGame " + pygame.version.ver.replace('release', '')

# Python getopt is cool and so is DiveIntoPython where I learned about
# it: http://www.faqs.org/docs/diveintopython/kgp_commandline.html
#
# but for now I'm taking him up on "For simple programs that only take a 
# single argument and have no flags, you can simply use sys.argv[1] to 
# access the argument. There’s no shame in this; I do it all the time.
    commandLineRecordingArg = None
    try:
        commandLineRecordingArg = sys.argv[1]
    except IndexError:
        pass
    if commandLineRecordingArg is not None:
        playbackRecording = commandLineRecordingArg
    else:
        playbackRecording = TEST_playbackRecording
    pygame.mixer.pre_init(44100, 0, 2, 1024)
    pygame.init()
    Noise.init_mixer()
    screen = Mainline.set_screen_mode(fullscreen) 
    pygame.display.set_caption('Serpentine 2011')
    pygame.mouse.set_visible(False)
        
    INIT_MUSIC = config['musicVolume']
    INIT_EFFECTS = config['effectsVolume']
    # Bleh I've muddled the need to init the mixer with the need to display
    # controls to set the volumes. Oh well it's better than it was in the
    # first game!
    musicIcon = pygame.image.load(os.path.join("graphics", "musicIcon.png")) 
    effectsIcon = pygame.image.load(os.path.join("graphics", "effectsIcon.png"))            
    musicVolumeController = Noise.Volume(screen, 
                                         location=[200,570], 
                                         size=[240,40], 
                                         init=INIT_MUSIC, 
                                         iconImage=musicIcon)
    effectsVolumeController = Noise.Volume(screen, 
                                           location=[42,570], 
                                           size=[240,40], 
                                           init=INIT_EFFECTS, 
                                           iconImage=effectsIcon)    
    pygame.mixer.music.set_volume(musicVolumeController.setting/100.0) 
    Noise.play_music(Global.gameMusic)
    
    if Global.productionMode and commandLineRecordingArg is None:
        Mainline.title_page(screen)
        ourStartPause = 50 # need maybe some sort of countdown effect? shrug
    else:
        ourStartPause = 20 # just a few frames to get the door open nicely
# Will want this to play our demo recordings, when demo/attract mode is ready
##    frameRecordFromFile = None
##    filename = os.path.join("recordings", playbackRecording)
##    try:
##        pkl_file = open(filename, 'r')
##        frameRecordFromFile = cerealizer.load(pkl_file)
##        pkl_file.close()  
##    except IOError:                        
##        #print "no recording to demo!"
##        pass
    frameRecordFromFile = get_demo_recording(playbackRecording, True)
    if frameRecordFromFile is None:
        if commandLineRecordingArg is not None:
            print "could not load", commandLineRecordingArg
            sys.exit() 
        frameAverage = determine_optimal_FPS(screen,
                          fullscreen, 
                          ourStartPause,
                          musicVolumeController,
                          effectsVolumeController)
        calc_optimal_speeds(frameAverage)
    else:
        frameAverage = -1 # will set when we work the first hist entry
    # Racetrack mode - for debugging and speed testing
    if 1 == 0:
        pygame.mixer.music.set_volume(0/100.0) 
        targetFPS = 30
        calc_optimal_speeds(targetFPS)
        Global.startSpeed = Global.maxSpeed
        (frameAverage, 
         fullscreen,
         selectedReplay) = play_the_game(screen,
                               fullscreen,
                               ourStartPause,
                               None,
                               2000,
                               targetFPS,
                               musicVolumeController,
                               effectsVolumeController)
        Mainline.title_page(screen)
    # A little cheesy but as Dr Bart says, time to push this out the door
    while commandLineRecordingArg != "quit":
        (frameAvg, 
        fullscreen,
        selectedReplay) = play_the_game(screen,
                                   fullscreen,
                                   ourStartPause,
                                   frameRecordFromFile,
                                   0,
                                   frameAverage,
                                   musicVolumeController,
                                   effectsVolumeController)
        if selectedReplay is not None:
            if selectedReplay == "replayDone":
                if commandLineRecordingArg is not None or TEST_playbackRecording != "":
                    commandLineRecordingArg = "quit"
                frameRecordFromFile = None
            else:
                frameRecordFromFile = get_demo_recording(selectedReplay)
if __name__ == '__main__':
    GUI_DUMP_NO_CONSOLE = True
    
    if GUI_DUMP_NO_CONSOLE:
        try:
            do_main()
        except SystemExit:
            pass
        except:
            pygame.init()
            screen = pygame.display.set_mode([640,480])
            pygame.display.set_caption('Serpentine 2011 FAIL-ABORT')
            screen.fill((0,64,0))
            font = pygame.font.Font(os.path.join("fonts", "white_rabbit.ttf"), 12)
            shooot = "PROGRAM FAILURE - ABORTING - BUG IN VERSION " + Global.program_version
            textOut = font.render(shooot, 1, (255,255,255))
            screen.blit(textOut, [0, 0])
            nextY = textOut.get_height() + 1
            theWorstHappened = textwrap.wrap( traceback.format_exc(), 80)
            theWorstHappened.insert(0, '')
            theWorstHappened.insert(0, "Please send a screenshot of this error to callmerob@gmail.com\n\n")
            theWorstHappened.insert(0, '')
            #excLines = theWorstHappened.split('\n')
            for line in theWorstHappened:
                textOut = font.render(line, 1, (255,255,255))
                screen.blit(textOut, [0, nextY])
                nextY = nextY + textOut.get_height() + 1
            beg = "PRESS ESC TO CLOSE"
            textOut = font.render(beg, 1, (255,255,255))
            screen.blit(textOut, [0, nextY + 20])
            pygame.display.flip()          
            stillGrabbingScreenshot = True
            while stillGrabbingScreenshot:
                for e in pygame.event.get():
                    if e.type == QUIT:
                        sys.exit() 
                    elif e.type == KEYDOWN:
                        if e.key == K_ESCAPE:
                            sys.exit()                
                        else:
                            stillWaiting = False        
    else:
        do_main()
