# coding=UTF8
# Copyright © 2011 Rob Leachman
# Please see the file COPYING in this distribution for license information
import os
import random
import pygame

import Global
from Snake import Snake

class Esnake(Snake):
    def __init__(self,
                 graph,
                 velocity, 
                 x, 
                 y, 
                 countdown=0, 
                 color=(255,0,0), 
                 artPathHack="graphics", 
                 spriteSheet=None
                ):
        Snake.__init__(self, 
                       graph,
                       velocity, 
                       x, 
                       y, 
                       color, 
                       artPathHack, 
                       spriteSheet=spriteSheet
                    )
        self.velocity = velocity
        self.length = 4
        self.set_goal_towards(pygame.K_UP, graph)
        self.countdown = countdown
        self.hatchTimer = 0
        self.hatchTime = 425 # frog has nearly no chance; should be random and/or based on level
        self.launching = False # Enemy snakes launch without concerns
        self.funFactor = .2 # it's no fun to have them hunt well all the time
        self.debuggerName = "enemy snake!"
        self.needDeathPainted = True
        self.driveToStart = False


    def set_hatchling_goal(self, graph, x, y, initDir):
        self.locationX = x; self.locationY = y
        self.nextX = float(x); self.nextY = float(y)            
        neighborList = graph[(x, y)] 
        randomNeighborNumber = random.randrange(0, len(neighborList)) 
        randomNeighbor = neighborList.keys()[randomNeighborNumber]
        self.goalX = randomNeighbor[0]
        self.goalY = randomNeighbor[1]
        self.stepsToGoal = self.velocity
        calcSteps = float(self.stepsToGoal)
        self.speedX = (self.goalX - self.nextX) / calcSteps
        self.speedY = (self.goalY - self.nextY) / calcSteps
        self.set_sprite_from_speed()
        #print "hatchling at", self.nextX, self.nextY
        for i in range (self.length-1, -1, -1):
            self.body[i].sprite = "wrong"
            self.body[i].dir = self.dir
            self.body[i].set_sprite(self.dir)


    def move(self, graph, screen, frog, peers, playerGoal):
        if self.countdown > 0:
            self.countdown -= 1
            return
        elif self.countdown < 0:
            #print "waiting snake's dramatic entry"
            return
        Snake.move(self, None, graph, screen, frog, peers, False, playerGoal)
        # The enemy snakes never stop, so if we're at a goal (hit the wall
        # at the end of a cooridor, likely) set another with no backtrack
        if self.stepsToGoal < 1:
            assert self.stepsToGoal == 0, "neg steps! %d" % self.stepsToGoal
            neighborList = graph[(self.nextX, self.nextY)]
            #print "pick next from", neighborList
            backX = -9
            if len(neighborList) > 1:
                backX = self.nextX; backY = self.nextY
                if self.dir == pygame.K_RIGHT:
                    backX -= 1
                elif self.dir == pygame.K_LEFT:
                    backX += 1
                elif self.dir == pygame.K_DOWN:
                    backY -= 1
                else:
                    backY += 1
                del neighborList[(backX, backY)] # actually updates the graph!
                #print "without backtrack:", neighborList
            # Turns out (there's that phrase again) it seems we always make 
            # decisions as part of set_goal_onward... so length here 
            # is always 1, no real choice to make.
            randomNeighborNumber = random.randrange(0, len(neighborList)) 
            randomNeighbor = neighborList.keys()[randomNeighborNumber]
            self.goalX = randomNeighbor[0]
            self.goalY = randomNeighbor[1]
            if backX <> -9:
                # Now put it back!
                neighborList[(backX, backY)] = 1
                graph[(self.nextX, self.nextY)] = neighborList
            self.stepsToGoal = self.velocity
            calcSteps = float(self.stepsToGoal)
            self.speedX = (self.goalX - self.nextX) / calcSteps
            self.speedY = (self.goalY - self.nextY) / calcSteps
            self.sprite = "wrong"
            self.set_sprite_from_speed()


    def set_sprite_from_speed(self):
        if self.speedX > 0:
            dir = pygame.K_RIGHT
        elif self.speedX < 0:
            dir = pygame.K_LEFT
        elif self.speedY > 0:
            dir = pygame.K_DOWN
        else:
            dir = pygame.K_UP
        self.dir = dir
        self.set_sprite(self.dir)


    def set_goal_onward(self, graph, frog, peersList, playerGoal):
        playerIsIdleOrJustTerrible = False
        playerX = playerGoal[0]
        playerY = playerGoal[1]
        playerLen = playerGoal[2]
        if playerLen < 0:
            playerLen = playerLen * -1
            playerIsIdleOrJustTerrible = True
        #print "hunt player:", playerX, playerY, playerIsIdleOrJustTerrible
        atX = self.goalX
        atY = self.goalY
        #print "Esnake onward! steps=", self.stepsToGoal, "at:", atX, atY
        assert self.stepsToGoal == 0, "negative steps! %d" % self.stepsToGoal
        neighborList = graph[(atX, atY)]
        if len(neighborList) > 2:
            #print "Set goal onward: set a goal"
            # No backtracks
            backX = -9
            if len(neighborList) > 1:
                backX = atX; backY = atY
                if self.dir == pygame.K_RIGHT:
                    backX -= 1
                elif self.dir == pygame.K_LEFT:
                    backX += 1
                elif self.dir == pygame.K_DOWN:
                    backY -= 1
                else:
                    backY += 1
                del neighborList[(backX, backY)]
            needRandomGoal = True
            peersList[self.debuggerName][3] = 99
            if frog.isEligible():
                #print " frog at:", int(frog.atX), int(frog.atY)
                path = self.shortest_path(graph, 
                                         (atX, atY), 
                                         (int(frog.nextX), int(frog.nextY))
                                         )
                if path is not None:
                    # Don't everybody hunt the frog: note how close I am to
                    # it and assume I'm the hunter, but then check my peers
                    # (screen out dead snakes and myself) and if somebody
                    # else has already calculated they are as close, I will
                    # not hunt the frog this time.
                    peersList[self.debuggerName][3] = len(path)
                    letUsHaveMeHunt = True
                    for p in peersList:
                        if p[2] <> self.debuggerName and not p[1]:
                            if len(path) >= p[3]:
                                letUsHaveMeHunt = False
                    if letUsHaveMeHunt:
                        if self.length >= playerLen:
                            #print "we don't need the stinkin' frog!"
                            peersList[self.debuggerName][3] = 99
                        else:
                            #print "enemy", self.debuggerName, "on the hunt!"
                            if len(path) < 2:
                                #print "PUNT"
                                pass
                            else:
                                self.goalX = path[1][0]
                                self.goalY = path[1][1]
                                needRandomGoal = False
            # Either I'm not hunting the frog or am on top of it... let's
            # see about our opponent. If I'm at least as long as him I want
            # to hunt, otherwise I should flee!           
            backX2 = -9
            if needRandomGoal:
                path = self.shortest_path(graph, 
                                           (atX, atY), 
                                           (playerX, playerY)
                                         )
                if path is not None:
                    if len(path) > 1:
                        needFlee = True
                        if self.length >= playerLen:
                            # If we tried to kill every time it'd be no fun
                            if (random.random() < self.funFactor) or playerIsIdleOrJustTerrible:
                                #print "HUNTING"
                                self.goalX = path[1][0]
                                self.goalY = path[1][1]
                                needRandomGoal = False
                                needFlee = False
                            else:
                                #print "MERCY"
                                needFlee = False
                        if needFlee:
                            #print " flee!"
                            backX2 = path[1][0]
                            backY2 = path[1][1]
                            del neighborList[(backX2, backY2)]
            if needRandomGoal:
                randomNeighborNumber = random.randrange(0, len(neighborList)) 
                randomNeighbor = neighborList.keys()[randomNeighborNumber]
                self.goalX = randomNeighbor[0]
                self.goalY = randomNeighbor[1]
            # Put back any nodes removed
            if backX <> -9:
                neighborList[(backX, backY)] = 1
                graph[(atX, atY)] = neighborList
            if backX2 <> -9:
                neighborList[(backX2, backY2)] = 1
                graph[(atX, atY)] = neighborList
            self.stepsToGoal = self.velocity
            calcSteps = float(self.stepsToGoal)
            self.speedX = (self.goalX - atX) / calcSteps
            self.speedY = (self.goalY - atY) / calcSteps
            if self.speedX > 0:
                dir = pygame.K_RIGHT
            elif self.speedX < 0:
                dir = pygame.K_LEFT
            elif self.speedY > 0:
                dir = pygame.K_DOWN
            else:
                dir = pygame.K_UP
            self.dir = dir
            self.set_sprite(self.dir) 
        else:
            Snake.set_goal_onward(self, graph)
