# coding=UTF8
# Copyright © 2011 Rob Leachman
# Please see the file COPYING in this distribution for license information
# (This thing is super cheesy but go ahead and take it ha ha)
import os
import sys
import random
import pygame

import Global


class Frog(pygame.sprite.Sprite):
    def __init__(self, location, limitFrames, artPathHack="graphics"):
        frogSize = Global.snakeSize
        pygame.sprite.Sprite.__init__(self)
        spriteSheet = pygame.image.load(os.path.join(artPathHack, 
          "frog" + str(frogSize) + "sheet.png")).convert_alpha()
        self.imageNorth = pygame.Surface(
          (frogSize, frogSize), pygame.SRCALPHA, 32)
        self.imageSouth = pygame.Surface(
          (frogSize, frogSize), pygame.SRCALPHA, 32)
        self.imageEast = pygame.Surface(
          (frogSize, frogSize), pygame.SRCALPHA, 32)
        self.imageWest = pygame.Surface((
          frogSize, frogSize), pygame.SRCALPHA, 32)
        self.imageNorth.blit(spriteSheet, 
          (0, 0), (0, 0, frogSize, frogSize))
        self.imageEast.blit(spriteSheet, 
          (0, 0), (frogSize+1, 0, frogSize, frogSize))
        self.imageSouth.blit(spriteSheet, 
          (0, 0), (0, frogSize+1, frogSize, frogSize))
        self.imageWest.blit(spriteSheet, 
          (0, 0), (frogSize+1, frogSize+1, frogSize, frogSize))        
        self.image = self.imageNorth
        self.rect = self.image.get_rect()
        self.atX = self.nextX = location[0]
        self.atY = self.nextY = location[1]
        self.state = "waiting"
        self.waitCount = self.random_off_timer() # stay off the screen
        self.velocity = 1.0
        self.steps = 40
        self.speedX = 0
        self.speedY = 0
        self.hopSteps = 0
        self.eraserOffset = Global.snakeSize/2
        if limitFrames > 0:
            self.waitCount = limitFrames + 1 # no frogs for speed calcs


    def __str__(self):
        msg = "Frog!\n"
        msg += " dir: %s oldX:%d Y:%d newX:%d Y:%d\n" % (self.dir, self.rect.left, self.rect.top, self.nextX, self.nextY)
        return msg


    def random_off_timer(self):
        ''' stay off the screen for some random interval '''
        # funny story, this:
        #return int(math.floor(random.random()*650)) + 250
        return random.randint(250,800)


    def random_hop_timer(self):
        ''' wait random between hops '''
        if random.random() < .3:
            return random.randint(20,40)
        else:
            return random.randint(130,160)


    def paint(self, screen):
        # Huh he paints every frame he's present, as if we erase the entire
        # screen each cycle. Interesting but the least of my concerns now.
        if self.speedX > 0:
            self.image = self.imageEast
        elif self.speedX < 0:
            self.image = self.imageWest
        elif self.speedY > 0:
            self.image = self.imageSouth
        elif self.speedY < 0:
            self.image = self.imageNorth
        paintX = int(self.atX*Global.cellSize+Global.offsetX+1)
        paintY = int(self.atY*Global.cellSize+Global.offsetY+1)
        self.rect.center = [paintX, paintY]
        screen.blit(self.image, [self.rect.left, self.rect.top])


    def render(self, screen):
        self.atX = self.nextX
        self.atY = self.nextY
        return self.paint(screen)


    def erase(self,screen, drawnMaze):
        """ Erase the frog by copying the original background; could be
            outside the drawn maze and that's fine, just do it. """
        paintX = self.atX*Global.cellSize+Global.offsetX+1 
        paintY = self.atY*Global.cellSize+Global.offsetY+1
        position = pygame.Rect(paintX-self.eraserOffset, 
                               paintY-self.eraserOffset, 
                               Global.snakeSize, 
                               Global.snakeSize)
        screen.blit(drawnMaze, position, position)
        # Force his location off the screen if he exited
        if self.state == "waiting":
            self.atX = self.atY = -1


    def eraseAnEgg(self, screen, drawnMaze, locationX, locationY):
        """ Erase an egg. Only needed for when the player gobbles and
            immediately backtracks, this hack seemed the best way to do it.
            Or at least the handiest way.
        """
        paintX = locationX*Global.cellSize+Global.offsetX+1 
        paintY = locationY*Global.cellSize+Global.offsetY+1
        position = pygame.Rect(paintX-self.eraserOffset, 
                               paintY-self.eraserOffset, 
                               Global.snakeSize, 
                               Global.snakeSize)
        screen.blit(drawnMaze, position, position)


    def setSprite(self, dir):
        if self.sprite != dir:
            self.sprite = dir
            if self.sprite == "north":
                self.image = self.imageNorth
            elif self.sprite == "south":
                self.image = self.imageSouth
            elif self.sprite == "east":
                self.image = self.imageEast
            #elif self.sprite = "west":
            else:
                self.image = self.imageWest   


    def hop(self, theirEgg, ourEgg, penHijinks):
        newHop = False
        # Leave him out if the level is almost over. He could still trash
        # our pen if he's on and hops in after the mother gets home...
        # Also need to keep him away while we get launched.
        if self.state == "waiting" and not penHijinks:
            assert self.waitCount > -1, "frog wait invalid %d" % self.waitCount
            self.waitCount -= 1
            if self.waitCount < 0:
                self.state = "enter"
        if self.state == "enter":
            newHop = True
            self.speedX = self.speedY = 0
            enterPlane = random.choice(("vertical", "horizontal")) 
            headsOrTails = random.choice(("heads","tails"))
            if enterPlane == "horizontal":
                if headsOrTails == "heads":
                    self.nextY = -.5
                    self.speedY = self.velocity / self.steps                    
                else:
                    self.nextY = 10.5
                    self.speedY = self.velocity / -self.steps                    
                self.nextX = random.randint(0, 12)
            else:
                if headsOrTails == "heads":
                    self.nextX = -.5
                    self.speedX = self.velocity / self.steps
                else:
                    self.nextX = 12.5
                    self.speedX = self.velocity / -self.steps
                self.nextY = random.randint(0, 10)
            # Hack so we don't trash score at 1024x768:
            if self.nextX == -.5 and self.nextY == 1 and Global.res == 3:
                self.nextY = 2
            self.state = "entering"
            self.hopSteps = 0
            self.waitCount = self.random_hop_timer()
        if self.state == "hop" or self.state == "safe":
            self.waitCount -= 1
            if self.waitCount < 0:
                self.waitCount = self.random_hop_timer()
                movePlane = random.choice(("vertical", "horizontal"))
                newHop = True
                # Target their egg (if it exists) unless ours exists
                # and is closer.
                targetEggX = theirEgg[0]
                targetEggY = theirEgg[1]
                if ourEgg[0] > -1:
                    if not targetEggX > -1:
                        targetEggX = ourEgg[0]
                        targetEggY = ourEgg[1]
                    else:
                        distToTheirs = abs(self.nextX - targetEggX) + abs(self.nextY - targetEggY)
                        distToOurs = abs(self.nextX - ourEgg[0]) + abs(self.nextY - ourEgg[1])
                        if distToOurs < distToTheirs:
                            targetEggX = ourEgg[0]
                            targetEggY = ourEgg[1]
                needRandom = True
                if targetEggX > -1:
                    if movePlane == "horizontal":
                        needRandom = False
                        if self.nextX > targetEggX:
                            self.speedX = self.velocity / -self.steps
                        elif self.nextX < targetEggX:
                            self.speedX = self.velocity / self.steps
                        elif self.nextY > targetEggY:
                            self.speedY = self.velocity / -self.steps
                        elif self.nextY < targetEggY:
                            self.speedY = self.velocity / self.steps
                        else:
                            needRandom = True
                    else:
                        needRandom = False
                        if self.nextY > targetEggY:
                            self.speedY = self.velocity / -self.steps
                        elif self.nextY < targetEggY:
                            self.speedY = self.velocity / self.steps
                        elif self.nextX > targetEggX:
                            self.speedX = self.velocity / -self.steps
                        elif self.nextX < targetEggX:
                            self.speedX = self.velocity / self.steps
                        else:
                            needRandom = True                        
                if needRandom:
                    headsOrTails = random.choice(("heads","tails")) 
                    if movePlane == "horizontal":
                        if headsOrTails == "heads":
                            self.speedX = self.velocity / self.steps
                        else:
                            self.speedX = self.velocity / -self.steps
                    else:
                        if headsOrTails == "heads":
                            self.speedY = self.velocity / self.steps
                        else:
                            self.speedY = self.velocity / -self.steps
                self.hopSteps = self.steps
                self.state="hopping"
        if not self.state == "waiting" and not self.state == "entering":
            if self.nextX < 0 or self.nextY < 0 or self.nextX > 12 or self.nextY > 10:
                self.state = "waiting"
                self.waitCount = self.random_off_timer()
                self.speedX = 0
                self.speedY = 0
            else:
                if self.state == "hop" or self.state == "safe":
                    if ((self.nextX == 0 and self.nextY > 7) or
                        (self.nextX == 12 and self.nextY > 7) or
                        (self.nextY == 10 and self.nextX > 9) or
                        (self.nextX < 0 or self.nextY < 0)):
                        self.state = "safe"
        return newHop


    def died(self, screen, background):
        self.erase(screen, background)
        self.atY = -1
        self.atX = -1
        self.rect.top = 0
        self.rect.left = 0
        self.state = "waiting"
        self.waitCount = self.random_off_timer()


    def fequal(self, a, b):
        #http://www.velocityreviews.com/forums/t351983-precision-for-equality-of-two-floats.html    
        return abs(a-b) < 1E-6


    def move(self):
        if self.state == "hopping" or self.state == "entering":
            if self.hopSteps > self.steps - 5:
                self.nextX += self.speedX
                self.nextY += self.speedY
                self.hopSteps -= 1            
            if self.hopSteps > self.steps - 20:
                self.nextX += self.speedX
                self.nextY += self.speedY
                self.hopSteps -= 1
            self.nextX += self.speedX
            self.nextY += self.speedY
            self.hopSteps -= 1
            if self.fequal(self.nextX, int(self.nextX+.005)) and self.fequal(self.nextY, int(self.nextY+.005) ):
                self.nextX = int(self.nextX+.005)
                self.nextY = int(self.nextY+.005)
                self.speedX = 0
                self.speedY = 0
                self.state = "hop"


    def controlFrog(self, dir):
        if self.state == "hopping":
            return
        self.speedX = self.speedY = 0
        if dir == "north":
            self.speedY = self.velocity / -self.steps
        elif dir == "south":
            self.speedY = self.velocity / self.steps
        elif dir == "east":
            self.speedX = self.velocity / self.steps
        else:
            self.speedX = self.velocity / -self.steps
        self.hopSteps = self.steps
        self.state = "hopping"


    def isEligible(self):
        if self.state == "hop" or self.state == "hopping":
            if ((self.nextX == 0 and self.nextY > 7) or
                (self.nextX == 12 and self.nextY > 7) or
                (self.nextY == 10 and self.nextX > 9) or
                (self.nextX < 0 or self.nextY < 0)):
                return False            
            return True
        return False


if __name__ == '__main__':
    Global.init_resolution(2)
    pygame.init()
    screen = pygame.display.set_mode(Global.screenResolution)
    
    #Global.snakeSize = 34

    f = Frog([-1,-1], 0, "../graphics")

    #eraser = pygame.surface.Surface([Global.snakeSize,Global.snakeSize])
    eraser = pygame.surface.Surface(Global.screenResolution)
    eraser.fill((0,0,0))
    
    screen.blit(f.imageNorth, (100,100))
    screen.blit(f.imageSouth, (100,150))
    screen.blit(f.imageEast, (150,125))
    screen.blit(f.imageWest, (50,125))
    
    
    while True:    
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()
            if event.type == pygame.KEYDOWN:
                if (pygame.key.get_mods() == 1024): # command key
                    if (event.key == pygame.K_q):
                        sys.exit()
                if event.key == pygame.K_h:
                    f.waitCount = -1
#                    f.hop()
                elif event.key == pygame.K_i:
                    f.controlFrog("north")
                elif event.key == pygame.K_k:
                    f.controlFrog("south")
                elif event.key == pygame.K_j:
                    f.controlFrog("west")
                elif event.key == pygame.K_l:
                    f.controlFrog("east")                    
                    
        f.hop((0,0), (0,0))
        f.move()
        f.erase(screen, eraser)
        if f.state <> "waiting":
            f.render(screen)
                        
        pygame.display.flip()          
