# coding=UTF8
# Copyright © 2011 Rob Leachman
# Please see the file COPYING in this distribution for license information

import os

program_version = '0.1a'

# Set debugger flag at BOJ, all else is now dynamic (from graphics menu)
try:
    pkl_file = open("test", 'r')
    pkl_file.close() 
    productionMode = False
except IOError:                        
    productionMode = True
    
# This can be set from the console:                       
alwaysShowFPS = False
    
# Another couple values we want somewhat handy: switch to save demo 
# recordings, my guess is yes always but...
# Also the music, starts at BOJ and also after we change tracks on hi score
recordLevelDemos = True
gameMusic = 'EricScotPorter_Minimalisticon_Remaster'


def init_resolution(resolutionCode):
    global res, screenResolution, topFontSize, clearScoreLeft, \
        clearScoreHeight, bottomFontSize, hiScoreColumn1, hiScoreColumn2, \
        hiScoreColumn3, hiScoreTop, snakeSize, mouthWidth, mouthHeight, \
        shimWidth, shimHeight, wallWidth, wallSize, offsetX, offsetY, \
        mapOffsetX, mapOffsetY, fudgeX, fudgeY, glyphFontSize, cellSize, \
        shimOffset, startSpeed, maxSpeed, resChange, mazeBottom, \
        postSize, setJoystickEnabled
    
    res = resolutionCode

    if res == 0:
        # Default resolution, everybody can do 640x480. I started here
        # with Spoonbill's 640x480 = snakesize 17 to get ratios and stuff.
        # (speed 27 is the worst for the "need sub-pixel painting" issue)
        screenResolution = [640,480]
        topFontSize = 20
        clearScoreLeft = 410
        clearScoreHeight = 21
        bottomFontSize = 11
        hiScoreColumn1 = 150
        hiScoreColumn2 = 300
        hiScoreColumn3 = 450
        hiScoreTop = 120
        
        snakeSize = 17
        mouthWidth = 7
        mouthHeight = 3
        shimWidth = 9
        shimHeight = 9

        wallWidth = 11 # base 7 now 11
        postSize = 15
        
        wallSize = snakeSize + 3
        
        offsetX = 115
        offsetY = 64
        mapOffsetX = 93
        mapOffsetY = 8
        mazeBottom = 427
        fudgeX = fudgeY = .005
        
        glyphFontSize = 10

    elif res == 1:
        # Awesome on my screen and yours provided you too have an Extended 
        # 80-column card. Ah the good old days.
        screenResolution = [1280,960]
        topFontSize = 40
        clearScoreLeft = 830
        clearScoreHeight = 43
        bottomFontSize = 20
        hiScoreColumn1 = 300
        hiScoreColumn2 = 600
        hiScoreColumn3 = 900    
        hiScoreTop = 300
        
        snakeSize = 34
        mouthWidth = 14
        mouthHeight = 6
        shimWidth = 18
        shimHeight = 18

        wallWidth = 21 # base 14 now 21
        postSize = 35
        
        wallSize = snakeSize + 6
        
        offsetX = 230
        offsetY = 125
        mapOffsetX = 187
        mapOffsetY = 14
        # For map editor, should we ever need it again:
        #mapOffsetX = 0
        #mapOffsetY = 0
        mazeBottom = 851
        fudgeX = fudgeY = 1.005
        
        glyphFontSize = 21

    elif res == 2:
        # Not too shabby, this is the max resolution on po8's notebook :)
        screenResolution = [800,600]
        topFontSize = 15
        clearScoreLeft = 620
        clearScoreHeight = 21
        bottomFontSize = 10
        hiScoreColumn1 = 150
        hiScoreColumn2 = 350
        hiScoreColumn3 = 550
        hiScoreTop = 150    
        
        snakeSize = 25
        mouthWidth = 9
        mouthHeight = 5
        shimWidth = 13
        shimHeight = 13   

        wallWidth = 15 # base 9, now 15 with posts
        postSize = 25
        
        wallSize = snakeSize + 3
        
        offsetX = 99
        offsetY = 50
        mapOffsetX = 67
        mapOffsetY = -32
        mazeBottom = 583
        fudgeX = fudgeY = .005
        
        glyphFontSize = 13  
        
    elif res == 3:
        # Same big snakes as as res 1 but doesn't require an 80-column card, 
        # just bog standard graphics, I shoe-horned it all in!
        screenResolution = [1024,768]
        topFontSize = 15
        clearScoreLeft = 620
        clearScoreHeight = 0
        bottomFontSize = 10
        hiScoreColumn1 = 150
        hiScoreColumn2 = 500
        hiScoreColumn3 = 800    
        hiScoreTop = 200
        
        snakeSize = 34
        mouthWidth = 14
        mouthHeight = 6
        shimWidth = 18
        shimHeight = 18

        wallWidth = 21
        postSize = 35
        
        wallSize = snakeSize + 6
        
        offsetX = 166
        offsetY = 40
        mapOffsetX = 122
        mapOffsetY = -71
        mazeBottom = 766
        fudgeX = .005
        fudgeY = 1.005
        
        glyphFontSize = 17

    else:
        print "invalid res", res
        print INVALID_RES # just crash

    # Computed
    cellSize = snakeSize * 2
    shimOffset = (snakeSize - shimWidth) / 2
    
    # Quick hack to offer joystick support, if desired; incomplete
    setJoystickEnabled = False

    # other junk, these are now calculated:
    resChange = False
    startSpeed = 0
    maxSpeed = 0
    # Now this is obsolete! Was hardcoded, then calculated, now it's not
    # even Global.
    #speed = 0 # how many steps to traverse a node, more is slower

