# coding=UTF8
# Copyright © 2011 Rob Leachman
# Please see the file COPYING in this distribution for license information

import pickle
import pprint

class Maze():
    def __init__(self):
        self.intersection=[[0]*12 for x in range(14)]
        self.wall_horizontal=[[0]*12 for x in range(13)]
        self.wall_vertical=[[0]*11 for x in range(14)]
        self.intersectionMap = None


    def __setitem__(self, key, item, required=False):
        #print "setitem key=", key, "item=", item, "required=", required
        if key == "intersection":
            self.intersection = item
        elif key == "intersectionMap":
            self.intersectionMap = item
        else:
            if required:
                new = 2
            else:
                new = 1
            newInt = 1
            if key == "vertical" or key=="verticalOFF":
                if key=="verticalOFF":
                    new = 0
                    newInt = -1
                self.wall_vertical[ item[0] ][ item[1] ] = new
                self.intersection[item[0]][item[1]] += newInt
                if item[1] < 11:
                    self.intersection[item[0]][item[1]+1] += newInt
                if self.intersectionMap is not None:
                    self.intersectionMap[item[0]][item[1]][0] = 1
                    if item[1] < 11:
                        self.intersectionMap[item[0]][item[1]+1][1] = 1
            elif key == "horizontal" or key=="horizontalOFF":
                if key=="horizontalOFF":
                    new = 0
                    newInt = -1            
                self.wall_horizontal[ item[0] ][ item[1] ] = new
                self.intersection[item[0]][item[1]] += newInt
                if item[0] < 13:
                    self.intersection[item[0]+1][item[1]] += newInt
                if self.intersectionMap is not None:
                    self.intersectionMap[item[0]][item[1]][2] = 1
                    if item[0] < 13:
                        self.intersectionMap[item[0]+1][item[1]][3] = 1


    def getHorizontal(self,x,y):
        return self.wall_horizontal[x][y]


    def getVertical(self,x,y):
        return self.wall_vertical[x][y]    


    def __str__(self):
        msg = " Horizontal: " + str(self.wall_horizontal) + "\n"
        msg += " Intersection: " + str(self.intersection) + "\n"
        return msg


    def clear(self):
        for x in range (13):
            for y in range (12):
                self.wall_horizontal[x][y] = 0
        for y in range (11):                
            for x in range (14):
                self.wall_vertical[x][y] = 0
        for x in range (14):
            for y in range (12):
                self.intersection[x][y] = 0
        # The outside walls:        
        for x in range (13):
            self.__setitem__("horizontal", (x,0))
            self.__setitem__("horizontal", (x,11))
        for y in range (11):
            self.__setitem__("vertical", (0,y))
            self.__setitem__("vertical", (13,y))
        # Homes for player and opponent:
        homesVertical = ( (1,8), (1,9), (1,10), (12,8), (12,9), (10,10) )
        homesHorizontal = ( (0,8), (12,8), (10,10), (11,10) )
        for i in homesVertical:
            self.__setitem__("vertical", i, True)
        for i in homesHorizontal:
            self.__setitem__("horizontal", i, True)            


    def save(self, graph, filename="pythentene.maze"):
        #pprint.pprint(self.wall_vertical)
        #pprint.pprint(self.wall_horizontal)
        #pprint.pprint(graph)        
        output = open(filename, 'w')
        pickle.dump(self.wall_vertical, output)
        pickle.dump(self.wall_horizontal, output)
        pickle.dump(graph, output)
        output.close()


    def load(self, filename="pythentene.maze"):
        self.clear()

        pkl_file = open(filename, 'r')
        loadV = pickle.load(pkl_file)
        loadH = pickle.load(pkl_file)
        graph = pickle.load(pkl_file)        
        #graph = None
        #pprint.pprint(loadV)
        #pprint.pprint(loadH)
        pkl_file.close()
        for y in range (11):                
            for x in range (14):
                if loadV[x][y] == 1:
                    self.__setitem__("vertical", (x,y), True)
                elif loadV[x][y] == 2:
                    self.__setitem__("vertical", (x,y), False)
        for x in range (13):
            for y in range (12):
                if loadH[x][y] == 1:
                    self.__setitem__("horizontal", (x,y), True)
                elif loadH[x][y] == 2:
                    self.__setitem__("horizontal", (x,y), False)
        return graph
