# coding=UTF8
# Copyright © 2011 Rob Leachman
# Please see the file COPYING in this distribution for license information
import sys
import os
import pickle
import pygame
from pygame.color import THECOLORS
from pygame.locals import *

import Global
from MapSprite import MapSprite
from Maze import Maze
from SpotSprite import SpotSprite # for testing


class MazeWallSprite(MapSprite):
    def __init__(self, location, orientation, color, wallWidth, wallImage = None):
        shim = int((Global.wallWidth - wallWidth) / 2)
        
        if orientation == "horizontal":
            MapSprite.__init__(self, location, [Global.cellSize-wallWidth-shim, wallWidth], color, wallImage)
            self.rect.top += shim
        else:
            MapSprite.__init__(self, location, [wallWidth, Global.cellSize-wallWidth-shim], color, wallImage) 
            self.rect.left += shim
            
        self.orientation = orientation
        self.location = location


    def paint(self, screen, wallWidth):
        if self.orientation == "horizontal":
           return MapSprite.paint(self, screen, [wallWidth, 0])
        else:
           return MapSprite.paint(self,  screen, [0, wallWidth])


    def paintPortion(self, screen, wallWidth, percentage, color):
        if self.orientation == "horizontal":
            left = self.rect.left + Global.mapOffsetX + wallWidth
            top = self.rect.top + Global.mapOffsetY
            pygame.draw.rect(screen, color, (left, top, self.rect.width - (int(self.rect.width * percentage)), wallWidth), 0)            
        else:
            left = self.rect.left + Global.mapOffsetX
            top = self.rect.top + Global.mapOffsetY + wallWidth
            pygame.draw.rect(screen, color, (left, top, wallWidth, self.rect.height - (int(self.rect.height * percentage))), 0)


    def getDoorImage(self, screenImage, wallWidth):
        if self.orientation == "horizontal":
            image = pygame.Surface((self.rect.width, wallWidth), pygame.SRCALPHA, 32)            
            left = self.rect.left + Global.mapOffsetX + wallWidth
            top = self.rect.top + Global.mapOffsetY
        else:
            image = pygame.Surface((wallWidth, self.rect.height), pygame.SRCALPHA, 32)
            left = self.rect.left + Global.mapOffsetX
            top = self.rect.top + Global.mapOffsetY + wallWidth
        image.blit(screenImage, (0, 0), (left, top, image.get_width(), image.get_height()))
        return image
        
    
if __name__ == '__main__':
#base 7/9/13
#walls and corners with widths 11/15/21 with posts 15/25/35    
        
    def new_render_maze(screen, maze, intersectionMap, level, spriteSheet):
        vertImage = pygame.Surface((Global.postSize, Global.snakeSize * 2 - Global.wallWidth), pygame.SRCALPHA, 32)
        horzImage = pygame.Surface((Global.snakeSize * 2 - Global.wallWidth, Global.postSize), pygame.SRCALPHA, 32)
        print "vert", Global.postSize, Global.snakeSize * 2 - Global.wallWidth
        print "horz", Global.snakeSize * 2 - Global.wallWidth, Global.postSize
        postImage = []
        for x in range (4):
            for y in range(4):
                p = pygame.Surface((Global.postSize, Global.postSize), pygame.SRCALPHA, 32)
                postImage.append(p)
        wallSet = (level - 1) % 6
        wallSetX = wallSet * 150
        if Global.res == 0:
            wallSetY = 0
        elif Global.res == 2:
            wallSetY = 100
        else:
            wallSetY = 250
        # Argh this is so wrong! But easier than reworking the sprites right now.
        # Next time the posts will be as wide as the walls. What you're
        # seeing is the product of 1) not knowing in advance, and 2) four
        # tries to get this as close as it is now... sigh.
        horzImage.blit(spriteSheet, (0, 0), (wallSetX + Global.postSize + 1, wallSetY, Global.postSize, Global.postSize))
        horzImage.blit(spriteSheet, (Global.postSize, 0), (wallSetX + Global.postSize + 1, wallSetY, horzImage.get_width() - Global.postSize, horzImage.get_height()))
        vertImage.blit(spriteSheet, (0, 0), (wallSetX + Global.postSize * 3 + 3, wallSetY + Global.postSize + 1, Global.postSize, Global.postSize))
        vertImage.blit(spriteSheet, (0, Global.postSize), (wallSetX + Global.postSize * 3 + 3, wallSetY + Global.postSize + 1, vertImage.get_width(), vertImage.get_height() - Global.postSize))
        if Global.res == 0:
            demunge = 2
        elif Global.res == 2:
            demunge = 5
        else:
            demunge = 7
        demunger = pygame.surface.Surface([demunge, Global.postSize])
        demunger.fill((0,0,0))
        horzImage.blit(demunger, (0, 0), (0, 0, demunger.get_width(), demunger.get_height()))
        horzImage.blit(demunger, (horzImage.get_width() - demunger.get_width(), 0), (0, 0, demunger.get_width(), demunger.get_height()))
        demunger = pygame.surface.Surface([Global.postSize, demunge])
        demunger.fill((0,0,0))
        vertImage.blit(demunger, (0, 0), (0, 0, demunger.get_width(), demunger.get_height()))
        vertImage.blit(demunger, (0, vertImage.get_height() - demunger.get_height()), (0, 0, demunger.get_width(), demunger.get_height()))
        
        p = 0
        for x in range (4):
            for y in range(4):
                postImage[p].blit(spriteSheet, (0, 0), (wallSetX + Global.postSize * x + x, wallSetY + Global.postSize * y + y, Global.postSize, Global.postSize))
                p += 1
        for x in range (13):
            for y in range (12):
                if maze.getHorizontal(x,y) > 0:
                    wall = MazeWallSprite([x,y], "horizontal", (255,0,0), Global.postSize, wallImage = horzImage)
                    wall.paint(screen, Global.wallWidth)
        for y in range (11):                
            for x in range (14):
                if maze.getVertical(x,y) > 0:
                    wall = MazeWallSprite([x,y], "vertical", (0,255,0), Global.postSize, wallImage = vertImage)
                    wall.paint(screen, Global.wallWidth)        
        for x in range (14):
            for y in range (12):
                spotNumber = -1
                if intersectionMap[x][y][0] > 0: # if top
                    if intersectionMap[x][y][1] > 0: # if bottom
                        if intersectionMap[x][y][2] > 0: # if right                       
                            if intersectionMap[x][y][3] > 0: # if left
                                spotNumber = 6 # X
                            else:
                                spotNumber = 2 # LT
                        else:
                            if intersectionMap[x][y][3] > 0:
                                spotNumber = 10 # RT
                            else:
                                spotNumber = 13 # H
                    else:
                        if intersectionMap[x][y][2] > 0:
                            if intersectionMap[x][y][3] > 0:
                                spotNumber = 5 # UT
                            else:
                                spotNumber = 1 # UL
                        else:
                            if intersectionMap[x][y][3] > 0:
                                spotNumber = 9 # UR
                            else:
                                spotNumber = 12 # U
                else:
                    if intersectionMap[x][y][1] > 0:
                        if intersectionMap[x][y][2] > 0:
                            if intersectionMap[x][y][3] > 0:
                                spotNumber = 7 # BT
                            else:
                                spotNumber = 3 # BL
                        else:
                            if intersectionMap[x][y][3] > 0:
                                spotNumber = 11 # BR
                            else:
                                spotNumber = 14 # D
                    else:
                        if intersectionMap[x][y][2] > 0:
                            if intersectionMap[x][y][3] > 0:
                                spotNumber = 4 # H
                            else:
                                spotNumber = 0 # L
                        else:
                            if intersectionMap[x][y][3] > 0:
                                spotNumber = 8 # R
                if spotNumber > -1:
                    spot = SpotSprite([x, y], Global.postSize, (0,0,255), postImage[spotNumber])
                    spot.paint(screen)


    def test_render_maze(screen, maze, intersections, level, spriteSheet):
        
        spotImage = pygame.Surface((Global.wallWidth, Global.wallWidth), pygame.SRCALPHA, 32)
        vertImage = pygame.Surface((Global.wallWidth, Global.snakeSize * 2 - Global.wallWidth), pygame.SRCALPHA, 32)
        horzImage = pygame.Surface((Global.snakeSize * 2 - Global.wallWidth, Global.wallWidth), pygame.SRCALPHA, 32)
        
        if Global.res == 3:
            wallSetX = 1 * 100
        else:
            wallSetX = Global.res * 100
            
        wallLevel = level - 1
        if wallLevel > 3:
            wallLevel = 1
        # DEBUG: force a wall set
        #wallLevel = 3
        wallSetY = wallLevel * 100

        spotImage.blit(spriteSheet, (0, 0), (wallSetX, wallSetY, spotImage.get_width(), spotImage.get_height()))
        vertImage.blit(spriteSheet, (0, 0), (wallSetX, Global.wallWidth + wallSetY, vertImage.get_width(), vertImage.get_height()))
        horzImage.blit(spriteSheet, (0, 0), (wallSetX + Global.wallWidth, wallSetY, horzImage.get_width(), horzImage.get_height()))
        
        for x in range (13):
            for y in range (12):
                if maze.getHorizontal(x,y) > 0:
                    wall = MazeWallSprite([x,y], "horizontal", (255,0,0), Global.wallWidth, wallImage = horzImage)
                    wall.paint(screen, Global.wallWidth)
        for y in range (11):                
            for x in range (14):
                if maze.getVertical(x,y) > 0:
                    wall = MazeWallSprite([x,y], "vertical", (0,255,0), Global.wallWidth, wallImage = vertImage)
                    wall.paint(screen, Global.wallWidth)
        for x in range (14):
            for y in range (12):
                if intersections[x][y] > 0:
                    spot = SpotSprite([x, y], Global.wallWidth, (0,0,255), spotImage)
                    spot.paint(screen)
    ##            if x == 0 and y == 0:
    ##                print "upper left:", spot.rect.left + Global.mapOffsetX, spot.rect.top + Global.mapOffsetY
    ##            if x == 13 and y == 11:
    ##                print "lower right:", spot.rect.left + spot.rect.width + Global.mapOffsetX, spot.rect.top + spot.rect.height + Global.mapOffsetY
                    
                    
    def render_graph(screen, graph):
        for x in range (13):
            for y in range(11):
                nodeX = x * Global.cellSize + Global.offsetX 
                nodeY = y * Global.cellSize + Global.offsetY
                if graph.has_key((x, y)):
                    node = graph[x, y]
                    #print "Node:", x, y, node
                    if not node.has_key((x-1, y)):
                        #print "need west"
                        pygame.draw.lines(screen, (255,255,255), False, [[nodeX-Global.snakeSize,nodeY-Global.wallSize],[nodeX-Global.snakeSize,nodeY+Global.wallSize]],Global.wallWidth)
                    if not node.has_key((x+1, y)):
                        #print "need east"
                        pygame.draw.lines(screen, (255,255,255), False, [[nodeX+Global.snakeSize,nodeY-Global.wallSize],[nodeX+Global.snakeSize,nodeY+Global.wallSize]],Global.wallWidth)
                    if not node.has_key((x, y-1)):
                        #print "need north"
                        pygame.draw.lines(screen, (255,255,255), False, [[nodeX-Global.wallSize,nodeY-Global.snakeSize],[nodeX+Global.wallSize,nodeY-Global.snakeSize]],Global.wallWidth)
                    if not node.has_key((x, y+1)):
                        #print "need south"
                        pygame.draw.lines(screen, (255,255,255), False, [[nodeX-Global.wallSize,nodeY+Global.snakeSize],[nodeX+Global.wallSize,nodeY+Global.snakeSize]],Global.wallWidth)
                else:
                    pygame.draw.lines(screen, (255,255,255), False, [[nodeX+Global.snakeSize,nodeY-Global.wallSize],[nodeX+Global.snakeSize,nodeY+Global.wallSize]],Global.wallWidth)
                    pygame.draw.lines(screen, (255,255,255), False, [[nodeX-Global.snakeSize,nodeY-Global.wallSize],[nodeX-Global.snakeSize,nodeY+Global.wallSize]],Global.wallWidth)
                    pygame.draw.lines(screen, (255,255,255), False, [[nodeX-Global.wallSize,nodeY-Global.snakeSize],[nodeX+Global.wallSize,nodeY-Global.snakeSize]],Global.wallWidth)
                    pygame.draw.lines(screen, (255,255,255), False, [[nodeX-Global.wallSize,nodeY+Global.snakeSize],[nodeX+Global.wallSize,nodeY+Global.snakeSize]],Global.wallWidth)
                        
    
    
    Global.init_resolution(2)
    pygame.init()
    screen = pygame.display.set_mode(Global.screenResolution) 

    #screen.fill((0,255,255))

    spriteSheet = pygame.image.load(os.path.join("../graphics", "wallSprites.png")).convert_alpha() 
    
    
    maze=Maze()
    intersections=[[0]*12 for x in range(14)]
    intersectionMap=[[[0]*4 for y in range(12)] for x in range(14)]    
    
    maze["intersection"] = intersections 
    maze["intersectionMap"] = intersectionMap    
    graph = maze.load("../mazes/test.maze")    
    
    
    #test_render_maze(screen, maze, intersections, 1, spriteSheet)
    new_render_maze(screen, maze, intersectionMap, 1, spriteSheet)
    #render_graph(screen, graph)               

    while True:    
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()
            if event.type == pygame.KEYDOWN:
                if (pygame.key.get_mods() == 1024): # command key
                    if (event.key == 113):   # q
                        sys.exit()
                if (event.key == pygame.K_ESCAPE):                        
                    sys.exit()
                        
        pygame.display.flip()                     
