# coding=UTF8
# Copyright © 2011 Rob Leachman
# Please see the file COPYING in this distribution for license information
import sys
import pygame

import Global
from SpotSprite import SpotSprite

class NodeSprite(SpotSprite):
    
    def __init__(self, location, spotSize=5, color=(0,164,205)):
        SpotSprite.__init__(self, location, spotSize, color) 
        self.loc = location
        nodeShim = Global.cellSize / 2
        self.rect.left += nodeShim
        self.rect.top += nodeShim
        self.spotSize = spotSize
        
    def paint_arc(self, screen, dir):
        #nodeX, nodeY = (self.location[0]+1 * Global.cellSize, (self.location[1]+1) * Global.cellSize)
        nodeX, nodeY = (self.rect.left, self.rect.top)

        # Shim for display offset, same as MapSprite
        nodeX += Global.offsetX
        nodeY += Global.offsetY
        
        if dir == pygame.K_UP:
            pygame.draw.lines(screen, [0, 164, 205], False, [[nodeX, nodeY],[nodeX, nodeY - Global.cellSize]], 1)
        elif dir == pygame.K_DOWN:
            nodeX += self.spotSize - 1
            pygame.draw.lines(screen, [0, 164, 205], False, [[nodeX, nodeY],[nodeX, nodeY + Global.cellSize]], 1) 
        elif dir == pygame.K_RIGHT:
            pygame.draw.lines(screen, [0, 164, 205], False, [[nodeX, nodeY],[nodeX + Global.cellSize, nodeY]], 1)
        else:
            nodeY += self.spotSize - 1
            pygame.draw.lines(screen, [0, 164, 205], False, [[nodeX, nodeY],[nodeX - Global.cellSize, nodeY]], 1)
            
    def paint(self, screen, graph={}):
        SpotSprite.paint(self, screen)
        thisNode = (self.loc[0], self.loc[1])
        if not thisNode in graph: # better be true! except when testing :)
            print "No graph for node!", thisNode
            return
        #print graph[thisNode]
        for arc in graph[thisNode]:
            arcLoc = arc
            #print "arc X", arcLoc[0], "Y", arcLoc[1]
            if self.loc[0] + 1 == arcLoc[0] and self.loc[1] == arcLoc[1]:
                self.paint_arc(screen, pygame.K_RIGHT)
            elif self.loc[0] - 1 == arcLoc[0] and self.loc[1] == arcLoc[1]:
                self.paint_arc(screen, pygame.K_LEFT) 
            elif self.loc[0] == arcLoc[0] and self.loc[1] - 1 == arcLoc[1]:                               
                self.paint_arc(screen, pygame.K_UP)                 
            elif self.loc[0] == arcLoc[0] and self.loc[1] + 1 == arcLoc[1]:                                               
                self.paint_arc(screen, pygame.K_DOWN)              
            else:
                print "invalid node,", arcLoc, " not connected to", self.loc
                
    def path_debugger(self, screen, pathList, graph, goal):
        #print "Path:", pathList
        prev = None
        for i in pathList:
            #print "PATH ITEM", i
            #print graph[str(i)]
            #print "  len", len(graph[str(i)])
            if prev is not None:
                #print "from",prev,"to",i
                #fromNode = NodeSprite(eval(prev), spotSize=35, color=(0,255,0))                                        
                if len(graph[str(i)]) > 2:
                    toNode = NodeSprite(eval(i), spotSize=35, color=(255,0,0))
                else:
                    toNode = NodeSprite(eval(i), spotSize=35, color=(0,255,0))                    
                #fromNode.paint(screen)
                toNode.paint(screen)
                #break # we only want to plot where we're at, and where next
                
            prev = i
            
        # paint the goal special so we can see it
        if goal is not None:
            goalNode = NodeSprite(goal, spotSize=35, color=(255,0,255))
            goalNode.paint(screen)

if __name__ == '__main__':
    pygame.init()
    screen = pygame.display.set_mode([1280,960])
    
    s = SpotSprite([1,1], 11)           
    s.paint(screen)
    s = SpotSprite([1,2], 11)           
    s.paint(screen)
    s = SpotSprite([2,1], 11)           
    s.paint(screen)
    s = SpotSprite([2,2], 11)           
    s.paint(screen)

    n = NodeSprite([0,0])           

    # The point of this thing was only to figure out graph dictionaries.
    # Started with the need for a path finder (and I sourced a great one
    # at http://code.activestate.com/recipes/119466/ (using the Connelly 
    # Barnes solution, rocking). But then determined I must rethink my
    # whole approach... the snake lives on the graph first, and only then
    # as graphic pixels! Or something like that.
    
    # The graph is a dictionary and the keys are tuples... so much better
    # than using locations as lists and casting as strings, doh!
    graph = {}
    testNode = (1,1)

    arc1 = {(1,2):1}
    print "ARC1:" , arc1
    arc2 = {(1,0):1}
    arc3 = {(0,1):1}
    nodeNeighbors = {}
    nodeNeighbors[(1,2)] = 1
    nodeNeighbors[(1,0)] = 1
    print "NEIGHBORS:", nodeNeighbors
    graph[ testNode ] = nodeNeighbors
    print "GRAPH:", graph
    
    # Append:
    old = graph[testNode]
    print "old", old
    #old.append(arc3)
    old[(0,1)] = 1
    print old
    graph[ testNode ] = old
    print "RESULT:", graph
    
    # This will work... for any node and arc, if node already graphed then
    # add the arc, else create a new graph node with the arc
    nodeKey2 = (1,1)
    newArc = (2,1)

    if nodeKey2 in graph:
        neighborList = graph[nodeKey2]
    else:
        neighborList = {}
    neighborList[newArc] = 1
    graph[nodeKey2] = neighborList

        
    print "GRAPH:", graph
        
    n = NodeSprite([1,1])           
    n.paint(screen, graph)        
    
    while True:    
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()
            if event.type == pygame.KEYDOWN:
                if (pygame.key.get_mods() == 1024): # command key
                    if (event.key == 113):   # q
                        sys.exit()
                        
        pygame.display.flip()                                     