# coding=UTF8
# Copyright © 2011 Rob Leachman
# Please see the file COPYING in this distribution for license information
# Gotta love noise!
# I used music and sound code from http://www.pygame.org/project-Arcade+Tonk+Tanks-1078-2412.html
import sys #for testing

import pygame
import os.path

class dummysound:
    def play(self): pass
    def set_volume(self, volume): pass


def load_sound(file, soundPathHack="sounds"):
    if not pygame.mixer: return dummysound()
    #file = os.path.join('..','Sound', file)
    hires_file = os.path.join(soundPathHack, file + ".wav")
    try:
        testFile = open(hires_file, "r")
        sound = pygame.mixer.Sound(hires_file)
        #print 'Sound loaded'
        return sound
    except pygame.error:
        print 'Warning, unable to load sound', file
    except IOError:
        try:
            lores_file = os.path.join(soundPathHack, "lowres", file + ".ogg")
            testFile = open(lores_file, "r")
            sound = pygame.mixer.Sound(lores_file)
            return sound
        except pygame.error:
            print 'Warning, unable to load even lowres sound', lores_file
        except IOError:
            print 'Warning, IO error loading even lowres sound', lores_file
    return dummysound()


def init_mixer():
    if pygame.mixer and not pygame.mixer.get_init():
        print 'WARNING: sound init failed!'
        pygame.mixer = None


def play_music(file, soundPathHack="sounds"):
    if pygame.mixer:
        hires_music = os.path.join(soundPathHack, file + ".mp3")
        try:
            pygame.mixer.music.load(hires_music)
            pygame.mixer.music.play(-1)
        except pygame.error:
            try:
                lores_music = os.path.join(soundPathHack, "lowres", file + ".ogg")
                pygame.mixer.music.load(lores_music)
                pygame.mixer.music.play(-1)
            except pygame.error:
                print "missing even lores music file", lores_music


def play_effect(effect):
    """ Play the effect (unless it is missing, then fail silently) """
    if type(effect).__name__=='instance':
        return
    channel = pygame.mixer.find_channel(True)
    if channel:
        channel.play(effect)
    else:
        print "no channel for effect, will try to use main"
        effect.play()     


class Volume(pygame.sprite.Sprite):
    def __init__(self, theScreen, location=[0,0], size=[100,20], init=50, iconImage=None):
        self.screenSurface = theScreen
        pygame.sprite.Sprite.__init__(self)
        image_surface = pygame.surface.Surface([10,int(size[1]*.8)])
        image_surface.fill([255,255,255])
        self.image = image_surface.convert()
        self.rect = self.image.get_rect()
        self.controlLocation = location
        self.controlSize = size
        self.setting = init
        self.mousePosition = [0,0]
        self.icon = iconImage


    def reposition(self, location, size):
        self.controlLocation = location
        self.controlSize = size
        image_surface = pygame.surface.Surface([10,int(size[1]*.8)])
        image_surface.fill([255,255,255])
        self.image = image_surface.convert()
        self.rect = self.image.get_rect()


    def paint_control(self):
        left = self.controlLocation[0]
        top = self.controlLocation[1]
        width = self.controlSize[0]
        height = self.controlSize[1]
        pygame.draw.rect(self.screenSurface, [100,100,100], [left-5,top,width+20,height])
        if self.icon is not None:
            pygame.draw.rect(self.screenSurface, [100,100,100], [left-50,top,50,height])
            iconVertCenter = (height - self.icon.get_height()) / 2
            self.screenSurface.blit(self.icon, [left - 40, top + iconVertCenter])
        lineInMiddle = top+height/2
        pygame.draw.lines(self.screenSurface, [0,0,0], False, [[left,lineInMiddle],[left+width,lineInMiddle]], 1)
        for i in range(0,6):
            markAt = left+i*(width/5)
            if i == 5: # HMM
                markAt = left + width
            if i == 0 or i == 5:
                markTop = top+2
                markBot = top + height -4
            else:
                markTop = lineInMiddle - self.controlSize[1]*.2
                markBot = lineInMiddle + self.controlSize[1]*.2
            pygame.draw.lines(self.screenSurface, [0,0,0], False, [[markAt, markTop],[markAt ,markBot]], 1)
        offset = (self.setting/100.0)*width
        self.rect.centery = lineInMiddle 
        self.rect.centerx = left+offset
        self.screenSurface.blit(self.image, self.rect)


    def mouse_is_on_control(self, mousePos):
        left = self.rect.left
        top = self.rect.top
        width = 10
        height = self.controlSize[1]*.8
        if (mousePos[0] > left and 
           mousePos[0] < left+width and 
           mousePos[1] > top and 
           mousePos[1] < top+height):
            self.mousePosition = mousePos
            return True
        return False


    def set_level(self, mousePos):
        self.setting += (mousePos[0] - self.mousePosition[0])/2
        if self.setting < 0:
            self.setting = 0
        if self.setting > 100:
            self.setting = 100
        self.mousePosition = mousePos


    def set_level_by_click(self, mousePos):
        left = self.controlLocation[0]
        top = self.controlLocation[1]
        width = self.controlSize[0]
        height = self.controlSize[1]
        if (mousePos[0] > left and 
           mousePos[0] < left+width and 
           mousePos[1] > top and 
           mousePos[1] < top+height):
            if mousePos[0] < self.rect.centerx:
                self.setting -= 20
            else:
                self.setting += 20
            self.setting = int(round(self.setting / 20.0) * 20)
            if self.setting < 0:
                self.setting = 0
            if self.setting > 100:
                self.setting = 100
            return True
        return False


if __name__ == '__main__':
    pygame.init()
    screen = pygame.display.set_mode([800,600])
    myClock = pygame.time.Clock()
    artPathHack = ".."
    musicIcon = pygame.image.load(os.path.join(artPathHack, "graphics", "musicIcon.png")) 
    effectsIcon = pygame.image.load(os.path.join(artPathHack, "graphics", "effectsIcon.png"))    
    v = Volume(screen, location=[100,250], size=[250,40], init=100, iconImage=musicIcon)
    v.reposition(location = [200,0], size = [100,50])
    v.paint_control()
    v2 = Volume(screen, location=[450,250], size=[250,30], init=100, iconImage=effectsIcon)
    v2.paint_control()
# stuff needed for init:
    heldDown = False
    play_music('EricScotPorter_Minimalisticon_Remaster.mp3', '../sounds')
    pygame.mixer.music.set_volume(1) 
#    screen.blit(pygame.image.load(os.path.join(artPathHack, "graphics", "effectsIcon.png")), [10,570])
#    screen.blit(pygame.image.load(os.path.join(artPathHack, "graphics", "musicIcon.png")), [170,570])     
    while True:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()
            elif event.type == pygame.MOUSEMOTION:
                mousePos = event.pos
                if heldDown:
                    v.set_level(mousePos)
                    print v.setting
                    pygame.mixer.music.set_volume(v.setting/100.0)
            elif event.type == pygame.MOUSEBUTTONDOWN:
                heldDown = v.mouse_is_on_control(mousePos)
            elif event.type == pygame.MOUSEBUTTONUP:
                if heldDown == False:
                    v.set_level_by_click(mousePos)
                    print v.setting
                    pygame.mixer.music.set_volume(v.setting/100.0)
                heldDown = False
            elif (event.type == pygame.KEYDOWN):
                if (pygame.key.get_mods() == 1024): # command key
                    if (event.key == 113):   # q
                        sys.exit()
                if event.key == pygame.K_SPACE:
                    print "Generate!"
        v.paint_control()        
        myClock.tick(30)
        pygame.display.flip()
