# coding=UTF8
# Copyright © 2011 Rob Leachman
# Please see the file COPYING in this distribution for license information
import sys
import os
import math
from pygame import display,init,font,key,Rect,event,draw
from pygame.locals import *
import cerealizer

import Global
import Noise

# http://code.activestate.com/recipes/498181-add-thousands-separator-commas-to-formatted-number/
# Luciano Ramalho
def splitthousands(s, sep=','):  
    if len(s) <= 3: return s  
    return splitthousands(s[:-3], sep) + sep + s[-3:]


def keynat(string):
    string = string.split(':')[0] # sort by score only
    '''A natural sort helper function for sort() and sorted()
    without using regular expressions or exceptions.

    >>> items = ('Z', 'a', '10th', '1st', '9')
    >>> sorted(items)
    ['10th', '1st', '9', 'Z', 'a']
    >>> sorted(items, key=keynat)
    ['1st', '9', '10th', 'a', 'Z'] 
    
    http://code.activestate.com/recipes/285264/
    paul clinch   
    '''
    it = type(1)
    r = []
    for c in string:
        if c.isdigit():
            d = int(c)
            if r and type( r[-1] ) == it: 
                r[-1] = r[-1] * 10 + d
            else: 
                r.append(d)
        else:
            r.append(c.lower())
    return r


#http://www.pygame.org/project-ask-1571-.html
def ask(POS,LENGHT,FONT,MAXLEN=0,COLOR=(255,255,255)):
    SCREEN = display.get_surface()
    BG = SCREEN.copy()
    # I don't know why this didn't work in testing, but it didn't
    # and we don't need it for this game so... just leave key repeat alone
    #repeat = key.get_repeat()
    #key.set_repeat(100,50)
    RECTANGLE = Rect(POS,(LENGHT,FONT.get_height()))
    INDEX = 0
    INPUT = ['_']
    def echo(POS):
        while True:
            X = 0
            C = [Rect(0,0,0,0).move(POS,RECTANGLE.top)]
            for i in INPUT:
                size = FONT.size(i)
                C.append(Rect((X,0),size).move(POS,RECTANGLE.top))
                X += size[0]
            C.append(Rect(X,0,0,0).move(POS,RECTANGLE.top))
            if C[INDEX+1].left <= RECTANGLE.left:
                POS += RECTANGLE.left-C[INDEX].left
            elif C[INDEX+1].right >= RECTANGLE.right:
                POS += RECTANGLE.right-C[INDEX+2].right
            SCREEN.blit(BG,RECTANGLE,RECTANGLE)
            X = POS
            for i in INPUT:
                SCREEN.blit(FONT.render(i,1,COLOR),(X,RECTANGLE.top))
                X += FONT.size(i)[0]
            display.update(RECTANGLE)
            yield C

    gen = echo(RECTANGLE.left)
    C = gen.next()    
    while True:
        ev = event.wait()
        if ev.type == KEYDOWN:
            if ev.key == K_RETURN or ev.key == K_KP_ENTER: break
            elif ev.key == K_BACKSPACE:
                if INDEX:
                    INPUT.pop(INDEX-1)
                    INDEX -= 1
                    C = gen.next()
            elif ev.key in (K_END,K_HOME):
                INPUT.pop(INDEX)
                INDEX = [len(INPUT),0][ev.key==K_HOME]
                INPUT.insert(INDEX,'|')
                C = gen.next()    
            elif ev.key == K_LEFT:
                if INDEX:
                    INPUT[INDEX],INPUT[INDEX-1] = INPUT[INDEX-1],INPUT[INDEX]
                    INDEX -= 1
                    C = gen.next()
            elif ev.key == K_RIGHT:
                if INDEX<len(INPUT)-1:
                    INPUT[INDEX],INPUT[INDEX+1] = INPUT[INDEX+1],INPUT[INDEX]
                    INDEX += 1
                    C = gen.next()
            elif ev.key == K_DELETE:
                if INDEX<len(INPUT)-1:
                    INPUT.pop(INDEX+1)
                    C = gen.next()
            elif ev.key == K_ESCAPE: return None
            elif ev.key == K_TAB: pass
            elif (len(INPUT)-1<MAXLEN if MAXLEN else 1):
                # Letters and numbers and space, is it...
                # Well those plus shifted numbers so some punctuation too
                if (ev.key > 96 and ev.key < 123) or (ev.key > 47 and ev.key < 58) or (ev.key == 32):
                    INPUT.insert(INDEX,ev.unicode)
                    INDEX += 1
                    C = gen.next()
        elif ev.type == MOUSEBUTTONDOWN and ev.button == 1:
            CARACT = Rect(ev.pos,(0,0)).collidelist(C)
            if CARACT > -1:
                CARACT = CARACT+(0 if ev.pos[0]<C[CARACT].centerx else 1)-1
                S = 0 if INDEX>CARACT else -1
                INPUT.pop(INDEX)
                INPUT.insert(CARACT+S,'|')
                INDEX = CARACT+S
                C = gen.next()
            elif Rect(ev.pos,(0,0)).colliderect(RECTANGLE):
                event.post(event.Event(KEYDOWN,{'scancode': 115, 'key': 279, 'unicode': u'', 'mod': 4096}))
        elif ev.type == QUIT:
                sys.exit()                
    SCREEN.blit(BG,POS,(POS,RECTANGLE.size))
    display.update((POS,RECTANGLE.size))
    #key.set_repeat(*repeat)
    INPUT.pop(INDEX)
    return ''.join(INPUT)


def print_score_table(screen, scoreTable, header, theFont):
    #screen.fill((0,0,0))
    if Global.res == 2:
        fontSize = Global.topFontSize * 2
    elif Global.res == 3:
        fontSize = 40
    else:
        fontSize = Global.topFontSize 
    if Global.res != 3:
        draw.rect(screen, (0,0,0), [0,Global.clearScoreHeight,Global.screenResolution[0], Global.screenResolution[1]], 0)
    else:
        draw.rect(screen, (0,0,0), [125, 0, Global.screenResolution[0], Global.screenResolution[1]], 0)
        draw.rect(screen, (0,0,0), [0, Global.screenResolution[1]-20, 125, 20], 0)
        
    line = 0
    screen.blit(theFont.render(header, 1, (255,255,255)),(Global.hiScoreColumn1,Global.hiScoreTop + (line-3)*fontSize))
    maxScoreBox = theFont.render('999,999',1,(255,255,255))
    maxScoreBoxWidth = maxScoreBox.get_width()
    screen.blit(theFont.render('Name',1,(255,255,255)),(Global.hiScoreColumn1,Global.hiScoreTop + (line-1)*fontSize))
    scoreBox = theFont.render('Score',1,(255,255,255))
    screen.blit(scoreBox,(Global.hiScoreColumn2 + (maxScoreBoxWidth-scoreBox.get_width()),Global.hiScoreTop + (line-1)*fontSize))
    levelBox = theFont.render('Level',1,(255,255,255))
    screen.blit(levelBox,(Global.hiScoreColumn3 + (59-levelBox.get_width()),Global.hiScoreTop + line-1*fontSize))
    
    for i in scoreTable:
        line += 1
        name = i.split(':')[1].split('.')[0]
        if name == '/':
            name = ''
        score = i.split(':')[0]
        if int(score) > 999999:
            score = '999999'
        
        for j in range(5):
            if int(score) < math.pow(10,j+1):
                score = " " + score
        if int(score) > 999:
            score = splitthousands(score)
        else:
            score = " " + score
        
        level = i.split(':')[1].split('.')[1]
        if int(level) < 10:
            level = " " + level
        
        screen.blit(theFont.render(name,1,(255,255,255)),(Global.hiScoreColumn1, Global.hiScoreTop + line*fontSize))
        scoreBox = theFont.render(score,1,(255,255,255))
        screen.blit(scoreBox,(Global.hiScoreColumn2 + (maxScoreBoxWidth-scoreBox.get_width()),Global.hiScoreTop + line*fontSize))
        levelBox = theFont.render(level,1,(255,255,255))
        screen.blit(levelBox,(Global.hiScoreColumn3 + (59-levelBox.get_width()),Global.hiScoreTop + line*fontSize))

        
def get_scores(dataDir):
    SCORES_FILE = os.path.join(dataDir, 'scores.dat')    
    try:
        pkl_file = open(SCORES_FILE, 'r')
        scores = cerealizer.load(pkl_file)
        pkl_file.close()  
    except IOError:
        # I had to think a bit but came up with 10 key people:
        scores = (
                '5000:snider.1', # David Snider invented this game
                '4500:spoonbill.2', # Spoonbill (Ian Humphreys) remade it
                '4000:guido.3', # Guido van Rossum invented Python
                '3500:shinners.4', # Pete Shinners created (started?) Pygame
                '3000:connelly.5', # This game would be nothing without the fast path finder by Connelly Barnes
                '2500:po8.6', # Mad props to Dr. Bart for setting me straight
                '2000:neal.7', # Byron continues to inspire
                '1500:b0n3b0y.8', # My man provides an excuse
                '1000:carter.9', # We do it for our kids right?
                '500:warren.10' # Warren and Carter's book got me going
                )
        put_scores(scores, dataDir)
    return scores


def put_scores(scores, dataDir):
    SCORES_FILE = os.path.join(dataDir, 'scores.dat')    
    output = open(SCORES_FILE, 'w')
    cerealizer.dump(scores, output)
    output.close() 


def display_the_table(screen, scores, fontPathHack, header="HIGH SCORES", footer="Hit any key to continue"):
    if  fontPathHack is None:
        fontPathHack = ""
    fontFile = os.path.join(fontPathHack, "fonts", "white_rabbit.ttf")
    if Global.res == 2:
        fontSize = Global.topFontSize * 2
    elif Global.res == 3:
        fontSize = 40
    else:
        fontSize = Global.topFontSize
    theFont = font.Font(fontFile, fontSize)         
    print_score_table(screen, scores, header, theFont) 
    levelBox = theFont.render(footer, 1, (255,255,255))
    screen.blit(levelBox, (Global.hiScoreColumn1 ,Global.hiScoreTop + 12*fontSize))
    display.flip()
    return theFont, fontSize

        
def wait_quit_or_continue():
    """ Public method - wait for quit or esc to abort, or other to continue.
    """
    # Weird... this seemed right:
    #
    # while event.wait().type not in (QUIT,KEYDOWN): pass
    # if event.wait().type == QUIT:
    #    sys.exit()    
    #
    # but it would only exit on 2 clicks of QUIT which is weird,
    # also though I want esc=quit so:
    stillWaiting = True
    while stillWaiting:
        for e in event.get():
            if e.type == QUIT:
                sys.exit() 
            elif e.type == KEYDOWN:
                if e.key == K_ESCAPE:
                    sys.exit()                
                else:
                    stillWaiting = False
            
    
def do_scores(screen, yourScore, yourLevel, dataDir, displayOnly=False, fontPathHack=None):
    """ Public method - display the score table
    """
    # All good except in the event of a tie, the result is alpha sort
    # on name. Fixing this bug left as an exercise for the reader.
    gotHighScore = False
    scores = get_scores(dataDir)
    scores = sorted(scores, key=keynat, reverse=True)[0:10]
    if displayOnly:
        display_the_table(screen, scores, fontPathHack)
        wait_quit_or_continue()
        return
    newScores = list(scores)
    newScore = str(yourScore) + ':' + '/' + '.' + str(yourLevel)
    newScores.append(newScore)
    newScores = sorted(newScores, key=keynat, reverse=True)[0:10]
    (theFont, theFontSize) = display_the_table(screen, 
            newScores,
            fontPathHack, 
            "You are a Serpentine hero!", 
            "Type your name then hit <enter>")
    if newScore in newScores:
        gotHighScore = True
        Noise.play_music('EricScotPorter_TheStateofBeing_Remaster')
        # Maybe the last silly hack for shoehorn hires mode!
        if Global.res == 3:
            draw.rect(screen, (0,0,0), [0, 0, 125, Global.screenResolution[1]], 0)
            display.flip()
        line = newScores.index(newScore)
        name = ask((Global.hiScoreColumn1, Global.hiScoreTop + (line + 1) * theFontSize), 300, theFont, 10)
        if name is None:
            name=""
        newScore = str(yourScore) + ':' + name + '.' + str(yourLevel)
        newScores = list(scores)
        newScores.append(newScore)    
        scores = sorted(newScores, key=keynat, reverse=True)[0:10]
    print_score_table(screen, scores, "HIGH SCORES", theFont)
    put_scores(scores, dataDir)    
    levelBox = theFont.render("Hit any key to continue",1,(255,255,255))
    screen.blit(levelBox, (Global.hiScoreColumn1, Global.hiScoreTop + 12*theFontSize))
    display.flip()
    wait_quit_or_continue()
    return gotHighScore


def get_highest_score_evar(dataDir):
    """ Public method: find the best score yet
    """
    scores = get_scores(dataDir)
    for i in scores:
        return i.split(':')[0]


if __name__ == '__main__':
    font.init()
    Global.init_resolution(1)
    screen = display.set_mode((Global.screenResolution[0],Global.screenResolution[1]))

    fontFile = os.path.join("../fonts", "white_rabbit.ttf")
    f = font.Font(fontFile, Global.topFontSize) 

    yourScore = 3333
    yourLevel = 8
    do_scores(screen, yourScore, yourLevel, '.', fontPathHack="..")
