# coding=UTF8
# Copyright © 2011 Rob Leachman
# Please see the file COPYING in this distribution for license information
import pygame

#import Global
from Snake import *

class Segment(Snake):
    def __init__(self, location, head, copySegment=None, spriteSheet=None):
        Snake.__init__(self, head.graph, 0, location[0], location[1], head.color, spriteSheet=spriteSheet)
        #print "NEW SEG", copySegment, head.locationX, head.locationY
        
        self.locationX = head.locationX
        self.locationY = head.locationY
        self.nextX = float(self.locationX); self.nextY = float(self.locationY)
        if copySegment is not None:
            if copySegment.color == (0,255,0):
                self.imageWest = self.GimageWest
                self.imageEast = self.GimageEast
                self.imageSouth = self.GimageSouth
                self.imageNorth = self.GimageNorth
                self.tailImageNorth = self.GtailImageNorth
                self.tailImageSouth = self.GtailImageSouth
                self.tailImageEast = self.GtailImageEast
                self.tailImageWest = self.GtailImageWest   
                self.shimWest = self.GshimWest
                self.shimEast = self.GshimEast
                self.shimSouth = self.GshimSouth
                self.shimNorth = self.GshimNorth                
            self.color = copySegment.color
            self.dir = copySegment.dir
            self.sprite = "wrong"
            self.set_sprite(self.dir) 
        paintX = int(self.locationX*Global.cellSize+Global.offsetX+1)
        paintY = int(self.locationY*Global.cellSize+Global.offsetY+1)
        self.rect.center = [paintX, paintY]

        self.head = head
        self.head.tail = self
        self.identity = self.head.identity + 1
        self.histLocation = 0
        self.histRipple = 0 # ripple back N hist locations on add
        self.moving = False
        self.waitCount = 0
        self.isTail = False
        self.prevDir = None


    def move(self, snake, launching, screen=None):
        if self.histRipple > 0:
            self.histRipple -= 1
            self.histLocation += 1
            if self.tail is not None:
                self.tail.move(snake, launching)  
            return

        hist = snake.history[self.histLocation]
        histX = hist[0]
        histY = hist[1]
        histDir = hist[2]

        # Much cleaner and easier to understand than before: if we're not
        # moving, wait until we're sure to not overlap, then start moving.
        # Doesn't always put this segment right on the butt of the previous
        # but we'll fix that later (is my plan right now, at least there's
        # always room for the tail, and no flicker when we don't move in
        # integer pixels!)
        if self.moving:
            self.prevDir = self.dir # not sure I'll need this but...
            self.nextY = histY
            self.nextX = histX
            self.dir = histDir
        else:
            # If we used this history location, would it overlap with our
            # head? If so need to wait else this is the one to use.
            paintHead = pygame.Rect(0, 0, Global.snakeSize, Global.snakeSize)
            # Note: erase refers to the previous frame, we're interested in 
            # this frame, but it's all a hack anyway... there's no one good 
            # answer for any given FPS, will either sometimes overlap or
            # sometimes lag very far behind. No worries, the segments must be
            # repositioned for speed handicaps and so this position is fixed.
            paintHead.left = self.head.eraseX
            paintHead.top = self.head.eraseY
            testHist = pygame.Rect(0, 0, Global.snakeSize, Global.snakeSize)
            paintX = int(histX*Global.cellSize+Global.offsetX+Global.fudgeX)
            paintY = int(histY*Global.cellSize+Global.offsetY+Global.fudgeY)
            testHist.center = paintX, paintY
            if paintHead.colliderect(testHist):
                # Colliding so must wait... except in the event we're doing
                # a "grow" segment after the snake gets going, in which
                # case there's a rare bug that needs to be trapped: don't
                # ever wait longer than the first segment (the head's tail)
                # had to wait.
                self.histLocation += 1
                self.waitCount += 1
                if self.identity > 1:
                    if self.waitCount > snake.tail.waitCount + 1:
                        if snake.allSegsOnPlayfield:
                            #print "JUST GO!"
                            self.moving = True
                            self.nextX = histX
                            self.nextY = histY
                            self.dir = histDir
            else:
                # See the rant above. Given a choice of a very large gap,
                # or an overlap... force an overlap here and we'll fix it
                # once everyone's moving.
                self.histLocation -= 1
                hist = snake.history[self.histLocation]
                histX = hist[0]
                histY = hist[1]
                histDir = hist[2]
                
                self.moving = True
                self.nextX = histX
                self.nextY = histY
                self.dir = histDir
        if self.moving:        
            self.set_sprite(self.dir)
        if self.tail is not None:
            self.tail.move(snake, launching)        
