# coding=UTF8
# Copyright © 2011 Rob Leachman
# Please see the file COPYING in this distribution for license information

import os
import sys, pickle # for testing
import random
import heapq
import pygame
import Global

class Snake(pygame.sprite.Sprite):
    def __init__(self, graph, velocity, x, y, color=[0,0,255], artPathHack="graphics", countdown=0, spriteSheet=None):
        pygame.sprite.Sprite.__init__(self)
        
        self.length = 3 #3 if just a head, we're dead
        
        # Simple test sprite
##        image_surface = pygame.surface.Surface([Global.snakeSize, Global.snakeSize])
##        image_surface.fill(color)
##        self.image = image_surface.convert()
        
        eggGraphic = "egg" + str(Global.snakeSize)        
        if color != (0,0,255):
            eggGraphic += "e"

        self.mouthOffset = (Global.snakeSize - Global.mouthWidth) / 2
        shimWidth = Global.shimWidth
        shimHeight = Global.shimHeight            
        shimOffset = (Global.snakeSize - shimWidth) / 2
        
        self.imageNorth = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.imageSouth = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.imageEast = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.imageWest = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.headImageNorth = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.headImageSouth = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.headImageEast = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.headImageWest = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.bonusheadImageNorth = pygame.Surface((Global.mouthWidth, Global.mouthHeight), pygame.SRCALPHA, 32)
        self.bonusheadImageSouth = pygame.Surface((Global.mouthWidth, Global.mouthHeight), pygame.SRCALPHA, 32)
        self.bonusheadImageEast = pygame.Surface((Global.mouthHeight, Global.mouthWidth), pygame.SRCALPHA, 32)
        self.bonusheadImageWest = pygame.Surface((Global.mouthHeight, Global.mouthWidth), pygame.SRCALPHA, 32)
        self.tailImageNorth = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.tailImageSouth = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.tailImageEast = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.tailImageWest = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)         
        self.shimNorth = pygame.Surface((shimWidth, shimHeight), pygame.SRCALPHA, 32)
        self.shimSouth = pygame.Surface((shimWidth, shimHeight), pygame.SRCALPHA, 32)
        self.shimEast = pygame.Surface((shimHeight, shimWidth), pygame.SRCALPHA, 32)
        self.shimWest = pygame.Surface((shimHeight, shimWidth), pygame.SRCALPHA, 32)
        self.GimageNorth = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.GimageSouth = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.GimageEast = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.GimageWest = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.GheadImageNorth = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.GheadImageSouth = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.GheadImageEast = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.GheadImageWest = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.GbonusheadImageNorth = pygame.Surface((Global.mouthWidth, Global.mouthHeight), pygame.SRCALPHA, 32)
        self.GbonusheadImageSouth = pygame.Surface((Global.mouthWidth, Global.mouthHeight), pygame.SRCALPHA, 32)
        self.GbonusheadImageEast = pygame.Surface((Global.mouthHeight, Global.mouthWidth), pygame.SRCALPHA, 32)
        self.GbonusheadImageWest = pygame.Surface((Global.mouthHeight, Global.mouthWidth), pygame.SRCALPHA, 32)
        self.GtailImageNorth = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.GtailImageSouth = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.GtailImageEast = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)
        self.GtailImageWest = pygame.Surface((Global.snakeSize, Global.snakeSize), pygame.SRCALPHA, 32)            
        self.GshimNorth = pygame.Surface((shimWidth, shimHeight), pygame.SRCALPHA, 32)
        self.GshimSouth = pygame.Surface((shimWidth, shimHeight), pygame.SRCALPHA, 32)
        self.GshimEast = pygame.Surface((shimHeight, shimWidth), pygame.SRCALPHA, 32)
        self.GshimWest = pygame.Surface((shimHeight, shimWidth), pygame.SRCALPHA, 32)            

        snakeSet = 0
        if color != (0,0,255):
            snakeSet = Global.snakeSize * 2 + 2

        self.imageNorth.blit(spriteSheet, (0, 0), (0, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.imageSouth.blit(spriteSheet, (0, 0), (Global.snakeSize+1, Global.snakeSize + snakeSet, Global.snakeSize, Global.snakeSize))
        self.imageEast.blit(spriteSheet, (0, 0), (Global.snakeSize*3+2, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.imageWest.blit(spriteSheet, (0, 0), (Global.snakeSize*2+2, Global.snakeSize + 1 + snakeSet, Global.snakeSize, Global.snakeSize))
        
        self.headImageNorth.blit(spriteSheet, (0, 0), (Global.snakeSize * 4 + 3, Global.mouthHeight + snakeSet, Global.snakeSize, Global.snakeSize))
        self.headImageSouth.blit(spriteSheet, (0, 0), (Global.snakeSize * 5 + 4, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.headImageEast.blit(spriteSheet, (0, 0), (Global.snakeSize * 6 + 5, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.headImageWest.blit(spriteSheet, (0, 0), (Global.snakeSize * 7 + 6 + Global.mouthHeight * 2, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        
        self.bonusheadImageNorth.blit(spriteSheet, (0, 0), (Global.snakeSize * 4 + 3 + self.mouthOffset, 0 + snakeSet, Global.mouthWidth, Global.mouthHeight))
        self.bonusheadImageSouth.blit(spriteSheet, (0, 0), (Global.snakeSize * 5 + 4 + self.mouthOffset, Global.snakeSize + snakeSet, Global.mouthWidth, Global.mouthHeight))
        self.bonusheadImageEast.blit(spriteSheet, (0, 0), (Global.snakeSize * 6 + 5 + Global.snakeSize, self.mouthOffset + snakeSet, Global.mouthHeight, Global.mouthWidth))
        self.bonusheadImageWest.blit(spriteSheet, (0, 0), (Global.snakeSize * 7 + 6 + Global.mouthHeight, self.mouthOffset + snakeSet, Global.mouthHeight, Global.mouthWidth))
        
        self.tailImageNorth.blit(spriteSheet, (0, 0), (Global.snakeSize * 8 + 7 + Global.mouthHeight * 2, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.tailImageEast.blit(spriteSheet, (0, 0), (Global.snakeSize * 9 + 8 + Global.mouthHeight * 2, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.tailImageSouth.blit(spriteSheet, (0, 0), (Global.snakeSize * 10 + 9 + Global.mouthHeight * 2, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.tailImageWest.blit(spriteSheet, (0, 0), (Global.snakeSize * 11 + 10 + Global.mouthHeight * 2, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        
        self.shimNorth.blit(spriteSheet, (0, 0), (shimOffset, Global.snakeSize + snakeSet, shimWidth, shimHeight))
        self.shimSouth.blit(spriteSheet, (0, 0), (Global.snakeSize+1+shimOffset, Global.snakeSize-shimHeight + snakeSet, shimWidth, shimHeight))
        self.shimEast.blit(spriteSheet, (0, 0), (Global.snakeSize * 3 + 2 - shimHeight, shimOffset + snakeSet, shimHeight, shimWidth))
        self.shimWest.blit(spriteSheet, (0, 0), (Global.snakeSize * 3 + 2, Global.snakeSize + 1 + shimOffset + snakeSet, shimHeight, shimWidth))

        # For enemy snakes and segments in peril, get the green
        snakeSet = Global.snakeSize * 4 + 4

        self.GimageNorth.blit(spriteSheet, (0, 0), (0, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.GimageSouth.blit(spriteSheet, (0, 0), (Global.snakeSize+1, Global.snakeSize + snakeSet, Global.snakeSize, Global.snakeSize))
        self.GimageEast.blit(spriteSheet, (0, 0), (Global.snakeSize*3+2, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.GimageWest.blit(spriteSheet, (0, 0), (Global.snakeSize*2+2, Global.snakeSize + 1 + snakeSet, Global.snakeSize, Global.snakeSize))
        
        self.GheadImageNorth.blit(spriteSheet, (0, 0), (Global.snakeSize * 4 + 3, Global.mouthHeight + snakeSet, Global.snakeSize, Global.snakeSize))
        self.GheadImageSouth.blit(spriteSheet, (0, 0), (Global.snakeSize * 5 + 4, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.GheadImageEast.blit(spriteSheet, (0, 0), (Global.snakeSize * 6 + 5, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.GheadImageWest.blit(spriteSheet, (0, 0), (Global.snakeSize * 7 + 6 + Global.mouthHeight * 2, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        
        self.GbonusheadImageNorth.blit(spriteSheet, (0, 0), (Global.snakeSize * 4 + 3 + self.mouthOffset, 0 + snakeSet, Global.mouthWidth, Global.mouthHeight))
        self.GbonusheadImageSouth.blit(spriteSheet, (0, 0), (Global.snakeSize * 5 + 4 + self.mouthOffset, Global.snakeSize + snakeSet, Global.mouthWidth, Global.mouthHeight))
        self.GbonusheadImageEast.blit(spriteSheet, (0, 0), (Global.snakeSize * 6 + 5 + Global.snakeSize, self.mouthOffset + snakeSet, Global.mouthHeight, Global.mouthWidth))
        self.GbonusheadImageWest.blit(spriteSheet, (0, 0), (Global.snakeSize * 7 + 6 + Global.mouthHeight, self.mouthOffset + snakeSet, Global.mouthHeight, Global.mouthWidth))
                      
        self.GtailImageNorth.blit(spriteSheet, (0, 0), (Global.snakeSize * 8 + 7 + Global.mouthHeight * 2, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.GtailImageEast.blit(spriteSheet, (0, 0), (Global.snakeSize * 9 + 8 + Global.mouthHeight * 2, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.GtailImageSouth.blit(spriteSheet, (0, 0), (Global.snakeSize * 10 + 9 + Global.mouthHeight * 2, 0 + snakeSet, Global.snakeSize, Global.snakeSize))
        self.GtailImageWest.blit(spriteSheet, (0, 0), (Global.snakeSize * 11 + 10 + Global.mouthHeight * 2, 0 + snakeSet, Global.snakeSize, Global.snakeSize))                   
        
        self.GshimNorth.blit(spriteSheet, (0, 0), (shimOffset, Global.snakeSize + snakeSet, shimWidth, shimHeight))
        self.GshimSouth.blit(spriteSheet, (0, 0), (Global.snakeSize+1+shimOffset, Global.snakeSize-shimHeight + snakeSet, shimWidth, shimHeight))
        self.GshimEast.blit(spriteSheet, (0, 0), (Global.snakeSize * 3 + 2 - shimHeight, shimOffset + snakeSet, shimHeight, shimWidth))
        self.GshimWest.blit(spriteSheet, (0, 0), (Global.snakeSize * 3 + 2, Global.snakeSize + 1 + shimOffset + snakeSet, shimHeight, shimWidth))
            
        self.RimageWest = self.imageWest
        self.RimageEast = self.imageEast
        self.RimageNorth = self.imageNorth
        self.RimageSouth = self.imageSouth
        self.RshimWest = self.shimWest
        self.RshimEast = self.shimEast
        self.RshimNorth = self.shimNorth
        self.RshimSouth = self.shimSouth
        self.RheadImageWest = self.headImageWest
        self.RheadImageEast = self.headImageEast
        self.RheadImageNorth = self.headImageNorth
        self.RheadImageSouth = self.headImageSouth
        self.RbonusheadImageWest = self.bonusheadImageWest
        self.RbonusheadImageEast = self.bonusheadImageEast
        self.RbonusheadImageNorth = self.bonusheadImageNorth
        self.RbonusheadImageSouth = self.bonusheadImageSouth
        self.RtailImageNorth = self.tailImageNorth
        self.RtailImageSouth = self.tailImageSouth
        self.RtailImageEast = self.tailImageEast
        self.RtailImageWest = self.tailImageWest              
            
        self.plotter = pygame.image.load(os.path.join(artPathHack, "plotter" + str(Global.snakeSize) +".png")).convert_alpha()
                
        self.image = self.imageNorth
        self.headImage = self.headImageNorth
        self.bonusheadImage = self.bonusheadImageNorth
        self.tailImage = self.tailImageNorth
        self.shimImage = self.shimNorth
        self.color = color
        self.rect = self.image.get_rect()
        
        self.locationX = x; self.locationY = y
        self.nextX = float(x); self.nextY = float(y)
        
        # calculate the first erase box, then erasing is simple?
        paintX = int(self.locationX*Global.cellSize+Global.offsetX+Global.fudgeX)
        paintY = int(self.locationY*Global.cellSize+Global.offsetY+Global.fudgeY)
        self.rect.center = [paintX, paintY]
        self.eraseX = self.rect.left
        self.eraseY = self.rect.top
        self.erase2X = None
        self.erase2Y = None
        self.erase2Xwidth = 0
        self.erase2Xheight = 0
        self.erased = False
        
        self.goalX = x
        self.goalY = y
        self.stepsToGoal = 0
        self.velocity = velocity
        
        self.identity = 0
        self.body = []

        
        self.head = None
        self.tail = None
        self.graph = graph
        
        self.prevGoalX = self.prevGoal1X = x
        self.prevGoalY = self.prevGoal1Y = y
        
        # This stuff is more about segments, but here for consistency (in print if nothing else)
        self.prevDir = None
        self.allSegsOnPlayfield = False
        self.histLocation = 0
        self.histRipple = 0
        self.isTail = False
        self.moving = False
        
        self.debuggerName = "snake!"
        self.nibbled = False
        self.embarrassingDeath = False
        self.spawning = False
        
        if self.identity == 0:
            self.eggImage = pygame.image.load(os.path.join(artPathHack, eggGraphic + ".png")).convert_alpha()
        self.init_snake_per_level(countdown, True)
        self.history = []
        hist = (self.locationX, self.locationY, self.dir, False)
        self.history.insert(0, hist)        


    def init_snake_per_level(self, countdown, driveToStart):
        self.speedX = 0
        self.speedY = 0 
        self.dir = pygame.K_UP
        self.sprite = pygame.K_UP  
                
        self.homeGoal = [12, 8]
        self.coilGoal = 9
        self.coilCount = 0 
        self.countdown = countdown
        self.launching = True # Enemy snakes launch without concerns
        self.launchDoor = -99
        self.enterDoor = -99 # Enemies launch off-screen, of course
        self.outDoor = -99
        self.outDoorNumber = 1
        self.doorTimer = 3
        self.enterDoorTimer = 3
        self.fancyDoor = self.fancyEnterDoor = False # last seg moved past it
        self.fancyEnterDoor = False # smarty comment goes here
        self.anticipatingWinning = True # hope to win and find door out
        self.nodeBeforeOut = [12, 7]
        self.driveToStart = driveToStart
        self.offPlayfield = False
        self.eggSize = 0
        self.eggGrowTime = 100 # 100
        self.eggTimer = self.eggGrowTime
        self.eggOdds = .0008 # .0008? maybe even less for player
        self.origEggOdds = self.eggOdds


    def __str__(self):
        msg = "Segment ID: %s\n" % self.identity
        msg += " dir=%s x=%f y=%f Hist=%d Ripple=%d\n Nibbled:%s Tail:%s" % (
               pygame.key.name(self.dir), 
               self.locationX, 
               self.locationY, 
               self.histLocation, 
               self.histRipple, 
               self.nibbled,
               self.isTail)
        msg += " head ID: "
        if self.head == None:
            msg += "None"
        else:
            msg += "%s" % self.head.identity
        msg += " tail ID: "
        if self.tail == None:
            msg += "None"
        else:
            msg += "%s" % self.tail.identity            
        msg += "\n"
        if self.tail is not None:
            msg += "%s" % self.tail
        return msg        
        
    def add_segment(self, segment):
        #segment.rect.left = self.rect.left
        #segment.rect.top = self.rect.top
        self.body.append(segment)
        
    def set_sprite(self, dir, color=None):
        # color change here was for debugging only, and it was whack
        if self.sprite != dir:
            self.sprite = dir
            if color is None:
                if self.sprite == pygame.K_UP:
                    self.image = self.imageNorth
                    self.headImage = self.headImageNorth
                    self.bonusheadImage = self.bonusheadImageNorth
                    self.tailImage = self.tailImageNorth
                    self.shimImage = self.shimNorth
                elif self.sprite == pygame.K_DOWN:
                    self.image = self.imageSouth
                    self.headImage = self.headImageSouth
                    self.bonusheadImage = self.bonusheadImageSouth
                    self.tailImage = self.tailImageSouth
                    self.shimImage = self.shimSouth
                elif self.sprite == pygame.K_LEFT:
                    self.image = self.imageWest
                    self.headImage = self.headImageWest
                    self.bonusheadImage = self.bonusheadImageWest
                    self.tailImage = self.tailImageWest
                    self.shimImage = self.shimWest
                else:
                    self.image = self.imageEast
                    self.headImage = self.headImageEast
                    self.bonusheadImage = self.bonusheadImageEast
                    self.tailImage = self.tailImageEast
                    self.shimImage = self.shimEast
            elif color == "red":
                if self.sprite == pygame.K_UP:
                    self.image = self.GimageNorth
                elif self.sprite == pygame.K_DOWN:
                    self.image = self.GimageSouth
                elif self.sprite == pygame.K_LEFT:
                    self.image = self.GimageWest
                #elif self.sprite = "west":
                else:
                    self.image = self.GimageEast
                
                
    def check_frog(self, frog):
        if self.collision((self.locationX, self.locationY), (frog.atX, frog.atY)):
            return True
        return False
    
    def check_egg(self, egg, frogAt):
        return self.collision((egg[0], egg[1]), (frogAt[0], frogAt[1]), 2)
   
 
    def collision(self, a, b, teenyEggSize=None):
        if teenyEggSize is not None:
            # Without this teeny hack the frog could land on your head and
            # eat your egg before getting eaten, now he loses
            aRect = pygame.Rect(0, 0, teenyEggSize, teenyEggSize)
        else:
            aRect = pygame.Rect(0, 0, Global.snakeSize, Global.snakeSize)
        paintX = int(a[0]*Global.cellSize+Global.offsetX+Global.fudgeX)
        paintY = int(a[1]*Global.cellSize+Global.offsetY+Global.fudgeY)
        aRect.center = [paintX, paintY]        
        bRect = pygame.Rect(0, 0, Global.snakeSize, Global.snakeSize)
        paintX = int(b[0]*Global.cellSize+Global.offsetX+Global.fudgeX)
        paintY = int(b[1]*Global.cellSize+Global.offsetY+Global.fudgeY)
        bRect.center = [paintX, paintY]
        if aRect.colliderect(bRect):
            return True
        return False
    
    
    def mouth_collision(self, a, aDirection, aOffset, b, bDirection, bOffset, checkKiss):
        aRect = pygame.Rect(0, 0, Global.snakeSize, Global.snakeSize)
        paintX = int(a[0]*Global.cellSize+Global.offsetX+Global.fudgeX)
        paintY = int(a[1]*Global.cellSize+Global.offsetY+Global.fudgeY)
        aRect.center = [paintX, paintY]
        bRect = pygame.Rect(0, 0, Global.snakeSize, Global.snakeSize)
        paintX = int(b[0]*Global.cellSize+Global.offsetX+Global.fudgeX)
        paintY = int(b[1]*Global.cellSize+Global.offsetY+Global.fudgeY)
        bRect.center = [paintX, paintY]       
        if aDirection == pygame.K_UP:
            aMouth = pygame.Rect(0, 0, Global.mouthWidth, Global.mouthHeight)
            aMouth.left = aRect.left + aOffset
            aMouth.top = aRect.top - Global.mouthHeight
        elif aDirection == pygame.K_DOWN:
            aMouth = pygame.Rect(0, 0, Global.mouthWidth, Global.mouthHeight)
            aMouth.left = aRect.left + aOffset
            aMouth.top = aRect.top + Global.snakeSize
        elif aDirection == pygame.K_LEFT:
            aMouth = pygame.Rect(0, 0, Global.mouthHeight, Global.mouthWidth)
            aMouth.left = aRect.left - Global.mouthHeight
            aMouth.top = aRect.top + aOffset
        else:
            aMouth = pygame.Rect(0, 0, Global.mouthHeight, Global.mouthWidth)
            aMouth.left = aRect.left + Global.snakeSize
            aMouth.top = aRect.top + aOffset
        if bDirection == pygame.K_UP:
            bMouth = pygame.Rect(0, 0, Global.mouthWidth, Global.mouthHeight)
            bMouth.left = bRect.left + bOffset
            bMouth.top = bRect.top - Global.mouthHeight
        elif bDirection == pygame.K_DOWN:
            bMouth = pygame.Rect(0, 0, Global.mouthWidth, Global.mouthHeight)
            bMouth.left = bRect.left + bOffset
            bMouth.top = bRect.top + Global.snakeSize
        elif bDirection == pygame.K_LEFT:
            bMouth = pygame.Rect(0, 0, Global.mouthHeight, Global.mouthWidth)
            bMouth.left = bRect.left - Global.mouthHeight
            bMouth.top = bRect.top + bOffset
        else:
            bMouth = pygame.Rect(0, 0, Global.mouthHeight, Global.mouthWidth)
            bMouth.left = bRect.left + Global.snakeSize
            bMouth.top = bRect.top + bOffset
        if checkKiss:
            if aMouth.colliderect(bRect) and bMouth.colliderect(aRect):
                return True                
            else:
                return False
        else:
            if aMouth.colliderect(bRect):
                return True
        return False 


    def body_collision(self, a, aDirection, aOffset, b):
        aRect = pygame.Rect(0, 0, Global.snakeSize, Global.snakeSize)
        paintX = int(a[0]*Global.cellSize+Global.offsetX+Global.fudgeX)
        paintY = int(a[1]*Global.cellSize+Global.offsetY+Global.fudgeY)
        aRect.center = [paintX, paintY]        
        bRect = pygame.Rect(0, 0, Global.snakeSize, Global.snakeSize)
        paintX = int(b[0]*Global.cellSize+Global.offsetX+Global.fudgeX)
        paintY = int(b[1]*Global.cellSize+Global.offsetY+Global.fudgeY)
        bRect.center = [paintX, paintY]
        
        if aDirection == pygame.K_UP:
            aMouth = pygame.Rect(0, 0, Global.mouthWidth, Global.mouthHeight)
            aMouth.left = aRect.left + aOffset
            aMouth.top = aRect.top - Global.mouthHeight
        elif aDirection == pygame.K_DOWN:
            aMouth = pygame.Rect(0, 0, Global.mouthWidth, Global.mouthHeight)
            aMouth.left = aRect.left + aOffset
            aMouth.top = aRect.top + Global.snakeSize
        elif aDirection == pygame.K_LEFT:
            aMouth = pygame.Rect(0, 0, Global.mouthHeight, Global.mouthWidth)
            aMouth.left = aRect.left - Global.mouthHeight
            aMouth.top = aRect.top + aOffset
        else:
            aMouth = pygame.Rect(0, 0, Global.mouthHeight, Global.mouthWidth)
            aMouth.left = aRect.left + Global.snakeSize
            aMouth.top = aRect.top + aOffset        
        
        if aMouth.colliderect(bRect):
            #print " collision!"            
            return True
        return False           


    def move(self, nextMove, graph, screen=None, frog=None, enemiesList=None, autoPilot=False, player=None):
        arrivedAtStart = False
        if self.countdown > 0:
            self.countdown -= 1
            return
        assert self.stepsToGoal > -1, "negative steps=%d" % self.stepsToGoal
        if self.speedX <> 0 or self.speedY <> 0:            
            self.stepsToGoal -= 1
            assert (self.speedX == 0 or self.speedY == 0), "moving X and Y both %f %f" % (self.speedX, self.speedY)
            self.nextX += self.speedX; self.nextY += self.speedY
            if self.stepsToGoal < 1:
                if autoPilot:
                    #print "auto=true"
                    if self.take_me_home(graph):
                       return True
                elif self.driveToStart:
                    #print "drive to start!"
                    if self.take_me_home(graph):
                        arrivedAtStart = True
                else:
                    if nextMove is not None:
                        # I'm undecided about allowing "tangle" but if it
                        # is needed, here's the code:
##                        if ((self.dir == pygame.K_RIGHT and nextMove == pygame.K_LEFT) or
##                            (self.dir == pygame.K_LEFT and nextMove == pygame.K_RIGHT) or
##                            (self.dir == pygame.K_UP and nextMove == pygame.K_DOWN) or
##                            (self.dir == pygame.K_DOWN and nextMove == pygame.K_UP)):
##                            #print "move backtrack"
##                            if self.dir != self.tail.dir:
##                                #print "move tangle"
##                                self.set_goal_onward(graph, frog, enemiesList, player)
##                            else:
##                                #print "move backtrack"
##                                self.direct(nextMove, graph, False)
##                        else:
##                            #print "trying move", nextMove
##                            self.direct(nextMove, graph, False)
                        # For now tangle is allowed so simply:
                        self.direct(nextMove, graph, False)
                    else:
                        self.set_goal_onward(graph, frog, enemiesList, player)
            self.tail.move(self, False, screen) # may need "launching" switch  
        else:
            if autoPilot:
                return self.take_me_home(graph)
        return arrivedAtStart


    def take_me_home(self, graph):
        atX = int(self.goalX + .005)
        atY = int(self.goalY + .005)
        #print "Esnake onward! steps=", self.stepsToGoal, "at:", atX, atY
        assert self.stepsToGoal == 0, "negative steps! %d" % self.stepsToGoal
        neighborList = graph[(atX, atY)]
        path = self.shortest_path(graph, (atX, atY), (self.homeGoal[0], self.homeGoal[1]), True)
        assert path is not None, "no path found from %d %d to %d %d" % (atX, atY, self.homeGoal[0], self.homeGoal[1])
        if self.anticipatingWinning and not self.driveToStart:
            #print "win path", path
            #print self.nodeBeforeOut[0], self.nodeBeforeOut[1]
            if (self.nodeBeforeOut[0], self.nodeBeforeOut[1]) in path:
                indexAtDoor = path.index((self.nodeBeforeOut[0], self.nodeBeforeOut[1]))
                if indexAtDoor > 0: # could indeed be right on top of it
                    self.nodeBeforeOut = path[indexAtDoor-1]
            self.anticipatingWinning = False # we won!
        # Move toward next step on path. When the destination is met the baby
        # is home, else our snake needs to coil up a few times, then (set
        # location just to be safe for the next level) and we're home too.
        # NOTE: Later I realized it is way better to adjust the snake's
        # movement outside of "move" but this part was already done and seems
        # to work, so whatever.
        if len(path) < 2:
            if self.homeGoal == [12, 8]: # baby never coils 
                if self.coilCount < 1 and ((self.length > 1) or (not self.driveToStart)):
                    #print "coil!"
                    self.goalX = 12
                    self.goalY = self.coilGoal
                    self.coilCount += 1
                    self.speedX = self.speedY = 0
                else:
                    if self.length > 3 and self.coilCount < 2:
                        #print "recoil!"
                        self.goalY = self.coilGoal
                        self.coilCount += 1
                        self.speedX = self.speedY = 0
                    else:
                        #print "made it!", atX, atY
                        self.locationX = atX
                        self.locationY = atY
                        return True
            else:
                #print "baby made it too!"
                return True
        else:
            self.goalX = path[1][0]
            self.goalY = path[1][1]
        self.stepsToGoal = self.velocity
        calcSteps = float(self.stepsToGoal)
        self.speedX = (self.goalX - atX) / calcSteps
        self.speedY = (self.goalY - atY) / calcSteps
        if self.speedX > 0:
            dir = pygame.K_RIGHT
        elif self.speedX < 0:
            dir = pygame.K_LEFT
        elif self.speedY > 0:
            dir = pygame.K_DOWN
        else:
            dir = pygame.K_UP

        self.dir = dir
        self.set_sprite(self.dir) 
        return False


    def direct(self, dir, graph, allEnemiesDead, raceTrackDebug=None):
        # Seemed better than cluttering up mainline with a bunch of tests, which I ended up doing anyway mutter:
        if raceTrackDebug is not None:
            doingRaceTrack = raceTrackDebug
        else:
            doingRaceTrack = False
        if allEnemiesDead and not doingRaceTrack:
            return None
        if self.launching and dir == pygame.K_DOWN:
            #print "launch down"
            return None
        # There were per-pixel issues which seemed impossible if we allow
        # a "tangle" or rapid repeated backtracking. Now I'm solving those
        # pixel issues another way (or not!) so... nevermind, allow tangle.
##        if ((self.dir == pygame.K_RIGHT and dir == pygame.K_LEFT) or
##            (self.dir == pygame.K_LEFT and dir == pygame.K_RIGHT) or
##            (self.dir == pygame.K_UP and dir == pygame.K_DOWN) or
##            (self.dir == pygame.K_DOWN and dir == pygame.K_UP)):
##            if self.dir != self.tail.dir:
##                #print "tangle!"
##                #print self.stepsToGoal
##                return dir
        assert (self.speedX == 0 or self.speedY == 0), "moving X and Y both %d %d" % (self.speedX, self.speedY)
        # Cannot go any faster in the same direction
        if (dir == pygame.K_LEFT and self.speedX < 0 or
            dir == pygame.K_RIGHT and self.speedX > 0 or
            dir == pygame.K_DOWN and self.speedY > 0 or
            dir == pygame.K_UP and self.speedY < 0):
            return None
        # If we're not at a node, the only move we need to do anything with
        # now is a backtrack. If we have remaining steps and we're moving
        # right or left, buffer any suggestions to move up or down (and the
        # converse). Otherwise, we've already thrown out repeated
        # instructions to move forward, so the only thing left are the good
        # instructions to move backward... so reverse speed and we're golden.
        if self.stepsToGoal > 0:
            if self.speedX <> 0 and (dir == pygame.K_UP or dir == pygame.K_DOWN):
                return dir
            if self.speedY <> 0 and (dir == pygame.K_RIGHT or dir == pygame.K_LEFT):
                return dir
            
            #print "Backtrack: only had", self.stepsToGoal, "steps remaining to get to ", self.goalX, self.goalY, "and now you want to go", pygame.key.name(dir)
            if dir == pygame.K_DOWN:
                self.goalY += 1
            elif dir == pygame.K_UP:
                self.goalY -= 1
            elif dir == pygame.K_RIGHT:
                self.goalX += 1
            else:
                self.goalX -= 1
            self.speedX = self.speedX * -1
            self.speedY = self.speedY * -1
            if self.stepsToGoal <> self.velocity:            
                self.stepsToGoal = self.velocity - self.stepsToGoal
            else:
                #print "Backtrack: turn on a dime!"
                #print ">>>MUST TEST:", pygame.key.name(dir), "from", self.goalX, self.goalY
                fromX = self.goalX
                fromY = self.goalY
                if dir == pygame.K_DOWN:
                    self.goalY += 1
                elif dir == pygame.K_UP:
                    self.goalY -= 1
                elif dir == pygame.K_RIGHT:
                    self.goalX += 1
                else:
                    self.goalX -= 1
                # It's all good except in the case where maybe you went
                # down into a corner, turned right (since down and left are
                # invalid), then *immediately* backtrack from right to left.
                # Almost never happens but it can so fix it:
                #print ">>>May be invalid: from", fromX, fromY, "to",  self.goalX, self.goalY
                neighborList = graph[(fromX, fromY)]
                #print "      ", neighborList
                if not neighborList.has_key((self.goalX, self.goalY)):
                    #print "  bad turn on a dime!"
                    self.goalX = fromX
                    self.goalY = fromY
                    self.speedX = 0
                    self.speedY = 0
                    self.stepsToGoal = 0
                    return None
                else:
                    #print "  good turn on a dime"
                    pass
            self.dir = dir
            self.set_sprite(self.dir)                    
        else:
            # We're at a node so can try to go in the new direction. If that
            # does not work, and we were moving before, try to keep going
            # that same way... 
            if not self.set_goal_towards(dir, graph):
                #print "  Hit wall: speed=", self.speedX, self.speedY, "steps", self.stepsToGoal
                if self.speedX == 0 and self.speedY == 0:
                    return None
                else:
                    if not self.set_goal_onward(graph):
                        return None
                #print " Go same way"
                return dir
            else:
                #print " Now heading:", pygame.key.name(dir)
                self.dir = dir
                self.set_sprite(self.dir)
        return None


    def set_goal_onward(self, graph, frog=None, enemiesList=None, myself=None):
        if self.speedX > 0:
            dir = pygame.K_RIGHT
        elif self.speedX < 0:
            dir = pygame.K_LEFT
        elif self.speedY > 0:
            dir = pygame.K_DOWN
        else:
            dir = pygame.K_UP
            
        if not self.set_goal_towards(dir, graph):                        
            self.speedX = 0
            self.speedY = 0
            self.stepsToGoal = 0
            return False
        return True


    def set_goal_towards(self, dir, graph):
        """ Snake made it to the goal, set the next goal in the same
            direction.
        """
        assert self.stepsToGoal < 1, "self still had steps=%d" % self.stepsToGoal
        # We made it to the goal, set the next one in the same direction
        #if self.debuggerName == "snake!":
            #print " Now at", self.goalX, self.goalY
        prevX = self.goalX
        prevY = self.goalY
        
        atX = self.nextX = self.goalX; 
        atY = self.nextY = self.goalY
        neighborList = graph[(atX, atY)]
        #print neighborList

        self.goalX = atX; self.goalY = atY
        if dir == pygame.K_DOWN:
            self.goalY = atY + 1
        elif dir == pygame.K_UP:
            self.goalY = atY - 1
        elif dir == pygame.K_RIGHT:
            self.goalX = atX + 1
        else:
            self.goalX = atX - 1
        #if self.debuggerName == "snake!":
            #print "checking", neighborList, "for", self.goalX, self.goalY
        if neighborList.has_key((self.goalX, self.goalY)):
            self.stepsToGoal = self.velocity
            calcSteps = float(self.stepsToGoal)
            self.speedX = (self.goalX - atX) / calcSteps
            self.speedY = (self.goalY - atY) / calcSteps
            self.prevGoal1X = self.prevGoalX
            self.prevGoal1Y = self.prevGoalY
            self.prevGoalX = prevX
            self.prevGoalY = prevY            
            return True
        else:
            # The new goal was invalid and now it's our current goal.
            # Reset it to something valid... where we are at now.
            # (Why not set speed to zero too? Not needed for some reason).
            self.goalX = atX
            self.goalY = atY
            return False

        
    def fequal(self, a, b):
        return abs(a-b) < 1E-6
    

    def is_reverse(self, aDir, bDir):
        if (aDir == pygame.K_RIGHT and bDir == pygame.K_LEFT or
            aDir == pygame.K_LEFT and bDir == pygame.K_RIGHT or
            aDir == pygame.K_UP and bDir == pygame.K_DOWN or
            aDir == pygame.K_DOWN and bDir == pygame.K_UP):
            return True
        return False


    def locate(self, snake, snakeSpeed, alreadyDoingRipple):
        # Start with a vague approximation... we'll paint this cycle
        # wherever it calculated and then if possible, fix it up for the next
        paintX = int(self.nextX*Global.cellSize+Global.offsetX+Global.fudgeX)
        paintY = int(self.nextY*Global.cellSize+Global.offsetY+Global.fudgeY)
        self.rect.center = [paintX, paintY]
        headNext = pygame.Rect(0, 0, Global.snakeSize, Global.snakeSize)
        headNextX = int(self.head.nextX*Global.cellSize+Global.offsetX+Global.fudgeX)
        headNextY = int(self.head.nextY*Global.cellSize+Global.offsetY+Global.fudgeY)
        headNext.center = headNextX, headNextY
##        if self.debuggerName == "snake!" and self.identity == 1:
##            print "   locate seg 1:", self.nextX, self.nextY, self.head.nextX, self.head.nextY
        if self.moving and self.dir == self.head.dir and not alreadyDoingRipple:
            if self.dir == pygame.K_RIGHT:
                off = (self.rect.left - (headNext.left - Global.snakeSize) + 1) * -1
            elif self.dir == pygame.K_LEFT:
                off = ((headNext.left + Global.snakeSize) - self.rect.left) * -1 - 1
            elif self.dir == pygame.K_UP:
                off = self.rect.top - (headNext.top + Global.snakeSize) - 1
            else:
                off = (self.rect.top + Global.snakeSize - headNext.top) * -1 - 1
            # off indicates how many pixels away from perfect alignment;
            # 0 would place this segment exactly behind its head, -1 or worse
            # means overlap, and the best we can hope to do is not be off by
            # more than +1 or a few more, up to the number of pixels the head
            # moves per cycle. I believe without sub-pixel rendering this is
            # the best that can be done (spent weeks in the lab trying).
            
            #if self.debuggerName == "snake!" and self.identity == 1:
                #print "off", self.identity, pygame.key.name(self.dir), paintX, paintY, off
            if off < 0:
                return -1
            else:
                return off
        return 0
    
    
    def paint(self, screen):
        if self.identity == 0:
            paintX = int(self.locationX*Global.cellSize+Global.offsetX+Global.fudgeX)
            paintY = int(self.locationY*Global.cellSize+Global.offsetY+Global.fudgeY)
            self.rect.center = [paintX, paintY]
        # else we're expecting locate to have set the location of the segment
        self.eraseX = self.rect.left
        self.eraseY = self.rect.top        
        self.erase2X = None
        self.erase2Y = None        
        if self.identity == 0:
            pass # still tweaking the stupid shims
            screen.blit(self.headImage, [self.rect.left, self.rect.top])
        elif self.isTail:
            screen.blit(self.tailImage, [self.rect.left, self.rect.top])
        else:
            screen.blit(self.image, [self.rect.left, self.rect.top])
        if self.identity == 0:
            if self.dir == pygame.K_UP:
                screen.blit(self.bonusheadImage, (self.rect.left + self.mouthOffset, self.rect.top - Global.mouthHeight))
                self.erase2X = self.rect.left + self.mouthOffset
                self.erase2Y = self.rect.top - Global.mouthHeight
                self.erase2Xwidth = Global.mouthWidth
                self.erase2Xheight = Global.mouthHeight
            elif self.dir == pygame.K_DOWN:
                screen.blit(self.bonusheadImage, (self.rect.left + self.mouthOffset, self.rect.top + Global.snakeSize))
                self.erase2X = self.rect.left + self.mouthOffset
                self.erase2Y = self.rect.top + Global.snakeSize
                self.erase2Xwidth = Global.mouthWidth
                self.erase2Xheight = Global.mouthHeight
            elif self.dir == pygame.K_RIGHT:
                screen.blit(self.bonusheadImage, (self.rect.left + Global.snakeSize, self.rect.top + self.mouthOffset))
                self.erase2X = self.rect.left + Global.snakeSize
                self.erase2Y = self.rect.top + self.mouthOffset
                self.erase2Xwidth = Global.mouthHeight
                self.erase2Xheight = Global.mouthWidth
            else:
                screen.blit(self.bonusheadImage, (self.rect.left - Global.mouthHeight, self.rect.top + self.mouthOffset))
                self.erase2X = self.rect.left - Global.mouthHeight
                self.erase2Y = self.rect.top + self.mouthOffset
                self.erase2Xwidth = Global.mouthHeight
                self.erase2Xheight = Global.mouthWidth
        if self.identity > 0 and self.moving and self.dir == self.head.dir:
            # Hmm doing headNext twice per frame per segment... good thing
            # math is cheap on a 2011 computer, har har.
            headNext = pygame.Rect(0, 0, Global.snakeSize, Global.snakeSize)
            headNextX = int(self.head.nextX*Global.cellSize+Global.offsetX+Global.fudgeX)
            headNextY = int(self.head.nextY*Global.cellSize+Global.offsetY+Global.fudgeY)
            headNext.center = headNextX, headNextY
            if self.dir == pygame.K_UP:
                if abs(self.rect.top - headNext.top) >=  Global.snakeSize:
                    screen.blit(self.shimNorth, (self.rect.left + Global.shimOffset, self.rect.top - Global.shimHeight))
                    self.erase2X = self.rect.left + Global.shimOffset
                    self.erase2Y = self.rect.top - Global.shimHeight
                    self.erase2Xwidth = Global.shimWidth
                    self.erase2Xheight = Global.shimHeight                    
            elif self.dir == pygame.K_RIGHT:
                if abs(self.rect.left - headNext.left) >= Global.snakeSize:
                    screen.blit(self.shimEast, (self.rect.left + Global.snakeSize, self.rect.top + Global.shimOffset))
                    self.erase2X = self.rect.left + Global.snakeSize
                    self.erase2Y = self.rect.top + Global.shimOffset
                    self.erase2Xwidth = Global.shimHeight
                    self.erase2Xheight = Global.shimWidth
            elif self.dir == pygame.K_DOWN:
                if abs(self.rect.top - headNext.top) >=  Global.snakeSize:
                    screen.blit(self.shimSouth, (self.rect.left + Global.shimOffset, self.rect.top + Global.snakeSize))
                    self.erase2X = self.rect.left + Global.shimOffset
                    self.erase2Y = self.rect.top + Global.snakeSize
                    self.erase2Xwidth = Global.shimWidth
                    self.erase2Xheight = Global.shimHeight                    
            else:
                if abs(self.rect.left - headNext.left) >= Global.snakeSize:
                    screen.blit(self.shimWest, (self.rect.left - Global.shimHeight, self.rect.top + Global.shimOffset))    
                    self.erase2X = self.rect.left - Global.shimHeight
                    self.erase2Y = self.rect.top + Global.shimOffset
                    self.erase2Xwidth = Global.shimHeight
                    self.erase2Xheight = Global.shimWidth


    def erase(self, screen):
        pygame.draw.rect(screen, 
                         (0,0,0), 
                         (self.eraseX, self.eraseY, 
                              Global.snakeSize, Global.snakeSize), 
                         0)
        if self.erase2X is not None:
            pygame.draw.rect(screen, 
                             (0,0,0), 
                             (self.erase2X, self.erase2Y, 
                                  self.erase2Xwidth, self.erase2Xheight), 
                             0)


    def render(self, screen):
        if self.identity == 0:
            if self.locationX <> self.nextX or self.locationY <> self.nextY:
                # Need to know if we did a "backflip" since this kink will
                # throw off any attempt to reposition the segments
                isBackflip = False
                prevDir = self.history[0][2]
                if (self.dir == pygame.K_UP and prevDir == pygame.K_DOWN or
                    self.dir == pygame.K_DOWN and prevDir == pygame.K_UP or
                    self.dir == pygame.K_RIGHT and prevDir == pygame.K_LEFT or
                    self.dir == pygame.K_LEFT and prevDir == pygame.K_RIGHT):
                    isBackflip = True
                
                hist = (self.nextX, self.nextY, self.dir, isBackflip)
                self.history.insert(0, hist)
##            if self.debuggerName == "snake!":
##                print "snake hist length=", len(self.history)
##                for i in self.history:
##                    print "snake hist:", i[0], i[1], pygame.key.name(i[2])
##                print
        self.locationX = self.nextX
        self.locationY = self.nextY
        self.paint(screen)

        
    # Flip the enemy snake color based on length... must be in Snake and not
    # Esnake because segments are always Snake     
    def flip_color(self):
        #print "FLIP FROM:", self.color
        if self.color == (255,0,0):
            self.imageWest = self.GimageWest
            self.imageEast = self.GimageEast
            self.imageSouth = self.GimageSouth
            self.imageNorth = self.GimageNorth
            self.headImageNorth = self.GheadImageNorth
            self.headImageSouth = self.GheadImageSouth
            self.headImageEast = self.GheadImageEast
            self.headImageWest = self.GheadImageWest
            self.bonusheadImageNorth = self.GbonusheadImageNorth
            self.bonusheadImageSouth = self.GbonusheadImageSouth
            self.bonusheadImageEast = self.GbonusheadImageEast
            self.bonusheadImageWest = self.GbonusheadImageWest            
            self.tailImageNorth = self.GtailImageNorth
            self.tailImageSouth = self.GtailImageSouth
            self.tailImageEast = self.GtailImageEast
            self.tailImageWest = self.GtailImageWest            
            self.shimWest = self.GshimWest
            self.shimEast = self.GshimEast
            self.shimSouth = self.GshimSouth
            self.shimNorth = self.GshimNorth
            self.color = (0,255,0)
        else:
            self.imageWest = self.RimageWest
            self.imageEast = self.RimageEast
            self.imageSouth = self.RimageSouth
            self.imageNorth = self.RimageNorth
            self.headImageNorth = self.RheadImageNorth
            self.headImageSouth = self.RheadImageSouth
            self.headImageEast = self.RheadImageEast
            self.headImageWest = self.RheadImageWest
            self.bonusheadImageNorth = self.RbonusheadImageNorth
            self.bonusheadImageSouth = self.RbonusheadImageSouth
            self.bonusheadImageEast = self.RbonusheadImageEast
            self.bonusheadImageWest = self.RbonusheadImageWest            
            self.tailImageNorth = self.RtailImageNorth
            self.tailImageSouth = self.RtailImageSouth
            self.tailImageEast = self.RtailImageEast
            self.tailImageWest = self.RtailImageWest
            self.shimWest = self.RshimWest
            self.shimEast = self.RshimEast
            self.shimSouth = self.RshimSouth
            self.shimNorth = self.RshimNorth            
            self.color = (255,0,0)
        if self.tail is not None:
            self.tail.flip_color()
        self.sprite = "wrong"
        self.set_sprite(self.dir) 
        
    def check_enemy_kill(self, enemy, iWinOnTies):
        # Maybe our mouths didn't collide but instead our bodies overlapped,
        # if so it's a head kill and correctly determined below... but with
        # the new mouth-based logic... we could come here dead already, if
        # so get out, and we'll die in a minute (in check_death)
        if self.nibbled:
            return False
        # If it's a head-on crash the longer snake wins; if a tie the enemy
        # snake wins. If not a head-on collision, check for the interesting
        # case where our mouth hits their neck -- but it's too generous to
        # a shorter enemy snake, the player needs a break or will complain
        # "he's green how can he kill me"... so that's interesting too.
        #
        # In any event, if the heads don't come together in a kill, see if
        # our mouth bit any of the enemies body segs and if so those die.
        if self.mouth_collision((self.nextX, self.nextY), 
                                self.dir, 
                                self.mouthOffset, 
                                (enemy.nextX, enemy.nextY), 
                                enemy.dir, 
                                enemy.mouthOffset, 
                                True):
            # The heads collided, who's longer?
            if self.length > enemy.length:
                #print "We win!"
                return True
            else:
                if iWinOnTies and self.length == enemy.length:
                    #print "I win the tie!"
                    return True
                else:
                    pass
                    #print "They win!"
        elif (not self.color == (0,255,0) and 
              self.mouth_collision((self.nextX, 
                                    self.nextY), 
                                    self.dir, 
                                    self.mouthOffset, 
                                    (enemy.nextX, enemy.nextY), 
                                    enemy.dir, 
                                    enemy.mouthOffset, 
                                    False)): 
            #print "Nailed his neck!", self.color
            enemy.nibbled = True
            return True
        else:
            eHead = enemy
            while eHead.tail is not None:
                
                eHead = eHead.tail
                if self.body_collision((self.nextX, self.nextY), self.dir, self.mouthOffset, (eHead.nextX, eHead.nextY)):
                    #print "yum yum"
                    eHead.nibbled = True
                    # The green snake can't eat your head, but he can kill
                    # you if you let him eat all your segments. Especially
                    # funny if you just sit and he drives over you missing
                    # your mouth, the "bug" now turned into a feature.
                    # Wilhelm lives on :)
                    if self.color == (0,255,0):
                        eHead.embarrassingDeath = True
        return False


    def check_death(self):
        count = 0
        """ returns 1 if some segments died, negative on snake death,
            returns 0 if the enemy snake loses a segment while spawning"""
        if self.nibbled:
            return self.length * -1
        else:
            for i in range(self.length):
                if self.body[i].nibbled:
                    if self.body[i].embarrassingDeath:
                        self.embarrassingDeath = True
                    self.body[i].head.tail = None
                    if not self.body[i].spawning:
                        count += 1
                    oldLength = self.length
                    self.length = i
                    for j in range(oldLength - self.length):
                        self.body.pop()
                    if self.length == 0:
                        #print "I PERISH!"
                        return -1 # not quite accurate, but sufficient
                    break
        return count


    # http://code.activestate.com/recipes/119466/ Connelly Barnes    
    def shortest_path(self, G, start, end, goingHome=False):
        #print "   Pathfinder start=", start, "end=", end
        
        # Refuse to fail if the nodes are in the "safe zone"
        if ((end[0] == 0 and end[1] > 7) or
            ((end[0] == 12 and end[1] > 7) and not goingHome) or
            ((end[1] == 10 and end[0] > 9) and not goingHome) or
            (end[0] < 0 or end[1] < 0)):
            #print "REFUSE", end[0], end[1]
            return None       
        # More generally, refuse to fail if the nodes aren't on the graph
        if not G.has_key(start) or not G.has_key(end):
            #print "CANNOT", start, end
            return None
        def flatten(L):       # Flatten linked list of form [0,[1,[2,[]]]]
          while len(L) > 0:
             yield L[0]
             L = L[1]
        # Still we must catch exceptions, for the islands
        try:
            q = [(0, start, ())]  # Heap of (cost, path_head, path_rest).
            visited = set()       # Visited vertices.
            while True:
               (cost, v1, path) = heapq.heappop(q)
               if v1 not in visited:
                  visited.add(v1)
                  if v1 == end:
                     return list(flatten(path))[::-1] + [v1]
                  path = (v1, path)
                  for (v2, cost2) in G[v1].iteritems():
                     if v2 not in visited:
                        heapq.heappush(q, (cost + cost2, v2, path)) 
        except IndexError:
            return None


    def paint_egg(self, screen, smallFactor, x, y):
        smallEggImage = self.eggImage
        smaller = int(Global.snakeSize * smallFactor)
        smallEggImage = pygame.transform.scale(self.eggImage, 
                    (smaller, smaller))             
        paintX = int(x*Global.cellSize+Global.offsetX+Global.fudgeX) + \
                    (Global.snakeSize - smaller) / 2
        paintY = int(y*Global.cellSize+Global.offsetY+Global.fudgeY) + \
                    (Global.snakeSize - smaller) / 2
        self.rect.center = [paintX, paintY]
        screen.blit(smallEggImage, [self.rect.left, self.rect.top])


    def grow_egg(self):
        if self.eggSize == 0 and self.length > 1:
            if random.random() < self.eggOdds:
                self.eggSize = .1
            
        if self.eggSize > 0:
            if self.eggSize > 0 and self.eggSize < .9:
                if self.eggTimer < 1:
                    self.eggTimer = self.eggGrowTime
                    self.eggSize += .1
                else:
                    self.eggTimer -= 1
            else:
                self.eggSize = .1


if __name__ == '__main__':
    Global.init_resolution(0)
    pygame.init()
    screen = pygame.display.set_mode(Global.screenResolution) 
    
    # I forget why I wanted a maze here ever... this works... shrug
    pkl_file = open("../mazes/level1.maze", 'r')
    cruft = pickle.load(pkl_file)
    cruft = pickle.load(pkl_file)
    graph = pickle.load(pkl_file)        
    pkl_file.close()    
    
    #screen.fill((0,255,255))

    spriteSheet = pygame.image.load(os.path.join("../graphics", "snake" + str(Global.snakeSize) + "sheet.png")).convert_alpha()
    #screen.blit(spriteSheet, [0,0])
    
    snake = Snake(graph, 1, color=(0,0,255), x=0, y=0, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.paint(screen)
    screen.blit(snake.shimNorth, (snake.rect.left + Global.shimOffset, snake.rect.top + Global.snakeSize))
    
    snake = Snake(graph, 1, color=(0,0,255), x=3, y=0, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.dir = pygame.K_DOWN
    snake.set_sprite(snake.dir)
    snake.paint(screen)
    screen.blit(snake.shimSouth, (snake.rect.left + Global.shimOffset, snake.rect.top - Global.shimHeight))
    
    snake = Snake(graph, 1, color=(0,0,255), x=5, y=0, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.dir = pygame.K_RIGHT
    snake.set_sprite(snake.dir)    
    snake.paint(screen)    
    screen.blit(snake.shimEast, (snake.rect.left - Global.shimHeight, snake.rect.top + Global.shimOffset))
    
    snake = Snake(graph, 1, color=(0,0,255), x=7, y=0, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.dir = pygame.K_LEFT
    snake.set_sprite(snake.dir)    
    snake.paint(screen)
    screen.blit(snake.shimWest, (snake.rect.left + Global.snakeSize, snake.rect.top + Global.shimOffset))    

    snake = Snake(graph, 1, color=(0,0,255), x=1, y=0, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.isTail = True
    snake.identity = 1
    snake.paint(screen)
    
    snake = Snake(graph, 1, color=(0,0,255), x=4, y=0, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.isTail = True
    snake.identity = 1
    snake.dir = pygame.K_DOWN
    snake.set_sprite(snake.dir)
    snake.paint(screen)
    
    snake = Snake(graph, 1, color=(0,0,255), x=6, y=0, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.isTail = True
    snake.identity = 1
    snake.dir = pygame.K_RIGHT
    snake.set_sprite(snake.dir)    
    snake.paint(screen)    
    
    snake = Snake(graph, 1, color=(0,0,255), x=8, y=0, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.isTail = True
    snake.identity = 1
    snake.dir = pygame.K_LEFT
    snake.set_sprite(snake.dir)    
    snake.paint(screen)


    snake = Snake(graph, 1, color=(255,0,0), x=0, y=2, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.paint(screen)
    screen.blit(snake.shimNorth, (snake.rect.left + Global.shimOffset, snake.rect.top + Global.snakeSize))
    
    snake = Snake(graph, 1, color=(255,0,0), x=3, y=2, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.dir = pygame.K_DOWN
    snake.set_sprite(snake.dir)
    snake.paint(screen)
    screen.blit(snake.shimSouth, (snake.rect.left + Global.shimOffset, snake.rect.top - Global.shimHeight))
    
    snake = Snake(graph, 1, color=(255,0,0), x=5, y=2, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.dir = pygame.K_RIGHT
    snake.set_sprite(snake.dir)    
    snake.paint(screen)    
    screen.blit(snake.shimEast, (snake.rect.left - Global.shimHeight, snake.rect.top + Global.shimOffset))
    
    snake = Snake(graph, 1, color=(255,0,0), x=7, y=2, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.dir = pygame.K_LEFT
    snake.set_sprite(snake.dir)    
    snake.paint(screen)
    screen.blit(snake.shimWest, (snake.rect.left + Global.snakeSize, snake.rect.top + Global.shimOffset)) 
    
    snake = Snake(graph, 1, color=(255,0,0), x=1, y=2, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.isTail = True
    snake.identity = 1
    snake.paint(screen)
    
    snake = Snake(graph, 1, color=(255,0,0), x=4, y=2, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.isTail = True
    snake.identity = 1
    snake.dir = pygame.K_DOWN
    snake.set_sprite(snake.dir)
    snake.paint(screen)
    
    snake = Snake(graph, 1, color=(255,0,0), x=6, y=2, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.isTail = True
    snake.identity = 1
    snake.dir = pygame.K_RIGHT
    snake.set_sprite(snake.dir)    
    snake.paint(screen)    
    
    snake = Snake(graph, 1, color=(255,0,0), x=8, y=2, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.isTail = True
    snake.identity = 1
    snake.dir = pygame.K_LEFT
    snake.set_sprite(snake.dir)    
    snake.paint(screen)


    snake = Snake(graph, 1, color=(255,0,0), x=0, y=4, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.dir = pygame.K_UP
    snake.flip_color()
    snake.paint(screen)
    screen.blit(snake.shimNorth, (snake.rect.left + Global.shimOffset, snake.rect.top + Global.snakeSize))
    
    snake = Snake(graph, 1, color=(255,0,0), x=3, y=4, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.dir = pygame.K_DOWN
    snake.set_sprite(snake.dir)
    snake.flip_color()    
    snake.paint(screen)
    screen.blit(snake.shimSouth, (snake.rect.left + Global.shimOffset, snake.rect.top - Global.shimHeight))
    
    snake = Snake(graph, 1, color=(255,0,0), x=5, y=4, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.dir = pygame.K_RIGHT
    snake.set_sprite(snake.dir)    
    snake.flip_color()    
    snake.paint(screen)    
    screen.blit(snake.shimEast, (snake.rect.left - Global.shimHeight, snake.rect.top + Global.shimOffset))
    
    snake = Snake(graph, 1, color=(255,0,0), x=7, y=4, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.dir = pygame.K_LEFT
    snake.set_sprite(snake.dir)    
    snake.flip_color()    
    snake.paint(screen)
    screen.blit(snake.shimWest, (snake.rect.left + Global.snakeSize, snake.rect.top + Global.shimOffset))  

    snake = Snake(graph, 1, color=(255,0,0), x=1, y=4, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.isTail = True
    snake.identity = 1
    snake.flip_color()        
    snake.paint(screen)
    
    snake = Snake(graph, 1, color=(255,0,0), x=4, y=4, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.isTail = True
    snake.identity = 1
    snake.dir = pygame.K_DOWN
    snake.set_sprite(snake.dir)
    snake.flip_color()    
    snake.paint(screen)
    
    snake = Snake(graph, 1, color=(255,0,0), x=6, y=4, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.isTail = True
    snake.identity = 1
    snake.dir = pygame.K_RIGHT
    snake.set_sprite(snake.dir)    
    snake.flip_color()    
    snake.paint(screen)    
    
    snake = Snake(graph, 1, color=(255,0,0), x=8, y=4, artPathHack = "../graphics", spriteSheet = spriteSheet)
    snake.isTail = True
    snake.identity = 1
    snake.dir = pygame.K_LEFT
    snake.set_sprite(snake.dir)    
    snake.flip_color()    
    snake.paint(screen)
    while True:    
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()
            if event.type == pygame.KEYDOWN:
                if (pygame.key.get_mods() == 1024): # command key
                    if (event.key == 113):   # q
                        sys.exit()
                if (event.key == pygame.K_ESCAPE):                        
                    sys.exit()
                elif (event.key == pygame.K_g):
                    if snake.eggSize < .9:
                        snake.eggSize += .1
                        print "egg size=", snake.eggSize

        if snake.eggSize > 0:
            snake.paint_egg(screen, snake.eggSize, snake.locationX, snake.locationY)                        
                        
        pygame.display.flip()                     
        