# coding=UTF8
# Copyright © 2010 Rob Leachman
# Please see the file COPYING in this distribution for license information
import sys
import pickle
import pygame
import pprint
from pygame.color import THECOLORS
from pygame.locals import *

from lib import Global
from lib import MapSprite
from lib import SpotSprite
from lib import MazeWallSprite
from lib import Maze
from lib import NodeSprite

Global.init_resolution(1)
intersections=[[0]*12 for x in range(14)]

Global.offsetX = 0
Global.offsetY = 0
WALLSIZE = Global.snakeSize + 6  
OFFSETX = Global.snakeSize+Global.wallWidth/2-1
OFFSETY = 50


def render_editor():
    # Draw the editing grid
    for x in range (1,13):
        for y in range (0,11):
            wall = MazeWallSprite([x,y], "vertical", (64,64,64), Global.wallWidth)
            wall.paint(screen, Global.wallWidth)
    for y in range (1,11):
        for x in range (0,13):
            wall = MazeWallSprite([x,y], "horizontal", (64,64,64), Global.wallWidth)
            wall.paint(screen, Global.wallWidth)

def render_maze(maze):
    for x in range (13):
        for y in range (12):
            if maze.getHorizontal(x,y) > 0:
                wall = MazeWallSprite([x,y], "horizontal", (255,255,255), Global.wallWidth)
                wall.paint(screen, Global.wallWidth)
    for y in range (11):                
        for x in range (14):
            if maze.getVertical(x,y) > 0:
                wall = MazeWallSprite([x,y], "vertical", (255,255,255), Global.wallWidth)
                wall.paint(screen, Global.wallWidth)
                
    for x in range (14):
        for y in range (12):
            if intersections[x][y] > 0:
                spot = SpotSprite([x, y], Global.wallWidth, (255,255,255))
                spot.paint(screen) 
                
def render_graph(screen, graph):
# "You drew all this?" Vera said.
#
# "Oh no." Montalban laughed. "No human being could ever construct a map
# this sophisticated. Investor-analysis correlation engines use distributive
# intelligence."
#
# There, I feel better.

    for x in range (13):
        for y in range(11):
            nodeX = x*Global.cellSize+OFFSETX 
            nodeY = y*Global.cellSize+OFFSETY
            #pygame.draw.rect(screen, [128,128,128], [nodeX-2, nodeY-2, 5, 5], 0)  
            node = graph[x,y]
            #print "Node:", x, y, node
            if not node.has_key((x-1, y)):
                #print "need west"
                pygame.draw.lines(screen, (255,255,255), False, [[nodeX-Global.snakeSize,nodeY-WALLSIZE],[nodeX-Global.snakeSize,nodeY+WALLSIZE]],Global.wallWidth)
            if not node.has_key((x+1, y)):
                #print "need east"
                pygame.draw.lines(screen, (255,255,255), False, [[nodeX+Global.snakeSize,nodeY-WALLSIZE],[nodeX+Global.snakeSize,nodeY+WALLSIZE]],Global.wallWidth)
            if not node.has_key((x, y-1)):
                #print "need north"
                pygame.draw.lines(screen, (255,255,255), False, [[nodeX-WALLSIZE,nodeY-Global.snakeSize],[nodeX+WALLSIZE,nodeY-Global.snakeSize]],Global.wallWidth)
            if not node.has_key((x, y+1)):
                #print "need south"
                pygame.draw.lines(screen, (255,255,255), False, [[nodeX-WALLSIZE,nodeY+Global.snakeSize],[nodeX+WALLSIZE,nodeY+Global.snakeSize]],Global.wallWidth)


def produce_graph_from_funky_maze(screen, maze):
    for x in range (13):
        for y in range (12):
            if maze.getHorizontal(x,y) > 0:
                wall = MazeWallSprite([x,y], "horizontal", (255,255,255), Global.wallWidth)
                wall.paint(screen, Global.wallWidth)
    for y in range (11):                
        for x in range (14):
            if maze.getVertical(x,y) > 0:
                wall = MazeWallSprite([x,y], "vertical", (255,255,255), Global.wallWidth)
                wall.paint(screen, Global.wallWidth)
    for x in range (14):
        for y in range (12):
            if intersections[x][y] > 0:
                spot = SpotSprite([x, y], Global.wallWidth, (255,255,255))
                spot.paint(screen)
    graph = {}
    for x in range (13):
        for y in range (11):                
            if not maze.getVertical(x,y) > 0:
                nodeKey = (x, y)
                newArc = (x-1, y)
                if nodeKey in graph:
                    neighborList = graph[nodeKey]
                else:
                    neighborList = {}
                neighborList[newArc] = 1
                graph[nodeKey] = neighborList                                       
            if not maze.getVertical(x+1,y) > 0:
                nodeKey = (x, y)
                newArc = (x+1, y)
                if nodeKey in graph:
                    neighborList = graph[nodeKey]
                else:
                    neighborList = {}
                neighborList[newArc] = 1
                graph[nodeKey] = neighborList                                       
            if not maze.getHorizontal(x,y) > 0:
                nodeKey = (x, y)
                newArc = (x, y-1)
                if nodeKey in graph:
                    neighborList = graph[nodeKey]
                else:
                    neighborList = {}
                neighborList[newArc] = 1
                graph[nodeKey] = neighborList                                       
            if not maze.getHorizontal(x,y+1) > 0:
                nodeKey = (x, y)
                newArc = (x, y+1)
                if nodeKey in graph:
                    neighborList = graph[nodeKey]
                else:
                    neighborList = {}
                neighborList[newArc] = 1
                graph[nodeKey] = neighborList                                       
            n = NodeSprite([x, y])
            n.paint(screen, graph)
    return graph                   


class MouseCursor(pygame.sprite.Sprite):
    def __init__(self):
        pygame.sprite.Sprite.__init__(self)
        image_surface = pygame.surface.Surface([20,20])
        image_surface.fill([0,255,0])
        self.image = image_surface.convert()
        
    def update_location(self, location):
        self.rect = self.image.get_rect()
        self.rect.center = location

    def paint(self, screen):
        screen.blit(self.image, [self.rect.left,self.rect.top])
        
    def find_wall(self, wallList):
        cursorGroup = pygame.sprite.Group()
        cursorGroup.add(self)
            
        foundLocation = None
        foundOrientation = None
        walls = pygame.sprite.groupcollide(cursorGroup, wallList, 0, 0)
        for wall in walls.keys():
            if len(walls[wall]) > 1:
                print "AMBIGUIOUS"
            else:
                #print walls[wall][0].location, walls[wall][0].orientation
                return walls[wall][0].location[0], walls[wall][0].location[1], walls[wall][0].orientation
        return None,None,None


if __name__ == '__main__':

    
    pygame.init()
    screen = pygame.display.set_mode(Global.screenResolution)
    # Given a choice of figuring out why I need to set these to zero
    # to accomodate the mouse coords which are zero-based... or working
    # on something else, I'd rather type this dumb comment and:
    Global.mapOffsetX = 0
    Global.mapOffsetY = 0
    # I don't think forcing these in Global is required anymore, shrug
        
    pygame.display.set_caption('Pytentene Maze Editor')
    myClock = pygame.time.Clock()
    
    delay=100
    interval=50
    pygame.key.set_repeat(delay,interval)
    
    #pygame.draw.rect(screen, THECOLORS["white"], [0,0,800,600], 2)
    
    maze=Maze()
    maze["intersection"] = intersections
    
    # Init the editing grid intersections
    for x in range (14):
        for y in range (12):
            intersections[x][y] = 0
            
    # Set the outside walls
    for x in range (13):
        maze["horizontal"] = (x,0)
        maze["horizontal"] = (x,11)
        
    for y in range (11):
        maze["vertical"] = (0,y)
        maze["vertical"] = (13,y) 
        
    pygame.draw.rect(screen, THECOLORS["black"], [0,0,Global.screenResolution[0], Global.screenResolution[1]], 0)
    render_editor()
    render_maze(maze)
                
    wallList = pygame.sprite.Group()                 
                
    for x in range (1,13):
        for y in range (0,11):
            wall = MazeWallSprite([x,y], "vertical", (64,64,64), Global.wallWidth)
            wall.paint(screen, Global.wallWidth)
            wallList.add(wall)
    for y in range (1,11):
        for x in range (0,13):
            wall = MazeWallSprite([x,y], "horizontal", (64,64,64), Global.wallWidth)
            wall.paint(screen, Global.wallWidth)
            wallList.add(wall)        
    
    cursor = MouseCursor()
    graph = None

    while True:    
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()
            elif event.type == pygame.MOUSEMOTION:
                cursor.update_location(event.pos)
                #cursor.paint(screen)
            elif event.type == pygame.MOUSEBUTTONDOWN:
                [x, y, orientation] = cursor.find_wall(wallList)
                print x,y,orientation
                if orientation != None:
                    if orientation == "vertical":
                        if maze.getVertical(x,y) == 2:
                            print "REQUIRED"
                        elif maze.getVertical(x,y) > 0:
                            maze["verticalOFF"] = (x,y)
                            wall = MazeWallSprite([x,y], "vertical", (64,64,64), Global.wallWidth)
                            wall.paint(screen, Global.wallWidth)                             
                        else:
                            maze["vertical"] = (x,y)
                            wall = MazeWallSprite([x,y], "vertical", (255,255,255), Global.wallWidth)
                            wall.paint(screen, Global.wallWidth)
                    else:
                        if maze.getHorizontal(x,y) == 2:
                            print "REQUIRED"
                        elif maze.getHorizontal(x,y) > 0:
                            maze["horizontalOFF"] = (x,y)
                            wall = MazeWallSprite([x,y], "horizontal", (64,64,64), Global.wallWidth)
                            wall.paint(screen, Global.wallWidth)                             
                        else:
                            maze["horizontal"] = (x,y)
                            wall = MazeWallSprite([x,y], "horizontal", (255,255,255), Global.wallWidth)
                            wall.paint(screen, Global.wallWidth)                        
                
            elif event.type == pygame.KEYDOWN:
                if (pygame.key.get_mods() == 1024): # command key
                    if (event.key == 113):   # q
                        sys.exit()
                elif event.key == pygame.K_r:
                    print "RENDER"
                    pygame.draw.rect(screen, THECOLORS["black"], [0,0,Global.screenResolution[0], Global.screenResolution[1]], 0)
                    render_maze(maze)
                elif event.key == pygame.K_c:
                    print "CLEAR"
                    maze.clear()
                    graph = produce_graph_from_funky_maze(screen, maze)
                    pygame.draw.rect(screen, THECOLORS["black"], [0,0,Global.screenResolution[0], Global.screenResolution[1]], 0)
                    render_maze(maze)                    
                elif event.key == pygame.K_e:
                    print "EDIT"
                    pygame.draw.rect(screen, THECOLORS["black"], [0,0,Global.screenResolution[0], Global.screenResolution[1]], 0)
                    render_editor()
                    render_maze(maze)
                elif event.key == pygame.K_f:
                    print "FUNKY MAZE-> graph!"
                    pygame.draw.rect(screen, THECOLORS["black"], [0,0,Global.screenResolution[0], Global.screenResolution[1]], 0)
                    graph = produce_graph_from_funky_maze(screen, maze)
                    #pp = pprint.PrettyPrinter(indent=4)
                    #pp.pprint(graph)
                elif event.key == pygame.K_m:
                    if graph is None:
                        print "Press F to calc the graph, then M to render it"
                    else:
                        print "MAP FROM GRAPH"
                        pygame.draw.rect(screen, THECOLORS["black"], [0,0,Global.screenResolution[0], Global.screenResolution[1]], 0)
                        render_graph(screen, graph)
                    
                elif event.key == pygame.K_p:
                    print "PUT"
                    graph = produce_graph_from_funky_maze(screen, maze)
                    maze.save(graph)
                    
                    
                elif event.key == pygame.K_g:
                    print "GET"
                    graph = maze.load()
                    pygame.draw.rect(screen, THECOLORS["black"], [0,0,Global.screenResolution[0], Global.screenResolution[1]], 0)
                    render_maze(maze)
                    

        #cursor.paint(screen)
        
        myClock.tick(30)
        pygame.display.flip()    