# coding=UTF8
# Copyright © 2010 Rob Leachman
# Please see the file COPYING in this distribution for license information
import sys
import random
import pickle
import pygame
import pprint
from pygame.color import THECOLORS
from pygame.locals import *

from lib import Global

MY_SPEED = 26 # how many steps to traverse a node, more is slower
BUG_SPEED = 30

#intersections=[[0]*12 for x in range(14)]


                
def render_graph(screen, graph):

    for x in range (13):
        for y in range(11):
            nodeX = x*Global.cellSize+Global.offsetX 
            nodeY = y*Global.cellSize+Global.offsetY
            #pygame.draw.rect(screen, [128,128,128], [nodeX-2, nodeY-2, 5, 5], 0)  
            node = graph[x,y]
            #print "Node:", x, y, node
            if not node.has_key((x-1, y)):
                #print "need west"
                pygame.draw.lines(screen, (255,255,255), False, [[nodeX-Global.snakeSize,nodeY-Global.wallSize],[nodeX-Global.snakeSize,nodeY+Global.wallSize]],Global.wallWidth)
            if not node.has_key((x+1, y)):
                #print "need east"
                pygame.draw.lines(screen, (255,255,255), False, [[nodeX+Global.snakeSize,nodeY-Global.wallSize],[nodeX+Global.snakeSize,nodeY+Global.wallSize]],Global.wallWidth)
            if not node.has_key((x, y-1)):
                #print "need north"
                pygame.draw.lines(screen, (255,255,255), False, [[nodeX-Global.wallSize,nodeY-Global.snakeSize],[nodeX+Global.wallSize,nodeY-Global.snakeSize]],Global.wallWidth)
            if not node.has_key((x, y+1)):
                #print "need south"
                pygame.draw.lines(screen, (255,255,255), False, [[nodeX-Global.wallSize,nodeY+Global.snakeSize],[nodeX+Global.wallSize,nodeY+Global.snakeSize]],Global.wallWidth)
    
                   

class Bug(pygame.sprite.Sprite):
    count = 0
    def __init__(self, graph, x=None, y=None, color=[0,0,255]):
        self.identity = self.__class__.count
        self.__class__.count += 1
        
        pygame.sprite.Sprite.__init__(self)
        image_surface = pygame.surface.Surface([Global.snakeSize, Global.snakeSize])
        image_surface.fill(color)
        self.image = image_surface.convert()
        self.rect = self.image.get_rect()
        
        #print neighborList
        #print len(neighborList)
        # Pick a random neighbor
        if x is None:
            randomNodeNumber = random.randrange(0, len(graph)) 
            randomNode = graph.keys()[randomNodeNumber]
            x = randomNode[0]
            y = randomNode[1]
        
        self.locationX = x
        self.locationY = y
        self.nextX = float(x); self.nextY = float(y)
        
        self.eraserOffset = Global.snakeSize/2
        
        self.goalX = 0
        self.goalY = 0
        self.stepsToGoal = 0
        
        self.speedX = 0
        self.speedY = 0
        
    def set_goal(self, graph, identity=None):
        atX = int(self.nextX +.05); atY = int(self.nextY + .05)
        self.nextX = atX; self.nextY = atY
        neighborList = graph[(atX, atY)]
        
        
        # Let's not let the bugs go backward
        backX = -9
        if (self.speedX <> 0 or self.speedY <> 0) and len(neighborList) > 1:
            backX = atX; backY = atY
            if self.speedX > 0:
                backX -= 1
            elif self.speedX < 0:
                backX += 1
            elif self.speedY > 0:
                backY -= 1
            else:
                backY += 1
            #print "Bug at", atX, atY, "speed:", self.speedX, self.speedY, "back:", backX, backY, "list:", neighborList, "len=", len(neighborList)
            
            del neighborList[(backX, backY)] # actually updates the graph!
            #print "without backtrack:", neighborList

        randomNeighborNumber = random.randrange(0, len(neighborList)) 
        if identity is not None:
            if identity == 0:
                randomNeighborNumber = 1
            elif identity == 1:
                randomNeighborNumber = 1
        
        randomNeighbor = neighborList.keys()[randomNeighborNumber]
        self.goalX = randomNeighbor[0]
        self.goalY = randomNeighbor[1]

        if backX <> -9:
            # Now put it back!
            neighborList[(backX, backY)] = 1
            graph[(atX, atY)] = neighborList
            
            neighborList = graph[(atX, atY)]
            #print "PROOF", neighborList
        
        #print "new goal FROM x", self.nextX, "y", self.nextY, "GOAL x", self.goalX, "y", self.goalY
        
        self.stepsToGoal = BUG_SPEED
        calcSteps = float(self.stepsToGoal)
        self.speedX = (self.goalX - atX) / calcSteps
        self.speedY = (self.goalY - atY) / calcSteps
        
        #print "SPEED: x", self.speedX, "y", self.speedY, "for", self.stepsToGoal, "steps"
        
        
    def move(self, graph):
        # steps remaining might be zero, in the case where somebody
        # collided with this bug before it had a chance to move...
        # but will never be negative or that is indeed a bug
        assert self.stepsToGoal > -1, "steps=%d" % self.stepsToGoal
        self.stepsToGoal -= 1
        #print "Bug", self.identity, "moves", self.speedX
        self.nextX += self.speedX; self.nextY += self.speedY
        if self.stepsToGoal < 1:
            #print "set next goal"
            self.set_goal(graph)

    def paint(self, screen):
        paintX = int(self.locationX*Global.cellSize+Global.offsetX+1)
        paintY = int(self.locationY*Global.cellSize+Global.offsetY+1)
        
        self.rect.center = [paintX, paintY]         
        screen.blit(self.image, [self.rect.left, self.rect.top])

        
    def erase(self, screen):
        paintX = self.locationX*Global.cellSize+Global.offsetX+1 
        paintY = self.locationY*Global.cellSize+Global.offsetY+1
                
        pygame.draw.rect(screen, [0,0,0], [paintX-self.eraserOffset, paintY-self.eraserOffset, Global.snakeSize, Global.snakeSize], 0)
        
    def render(self, screen):
        self.locationX = self.nextX
        self.locationY = self.nextY
        
        self.paint(screen)
        
class Me(Bug):
    
    def __init__(self, graph, x=None, y=None, color=[0,0,255]):
        Bug.__init__(self, graph, x, y, color)
        self.velocity = MY_SPEED
        print "My velocity:", self.velocity
        
    def move(self, nextMove, graph):
        assert self.stepsToGoal > -1, "self steps=%d" % self.stepsToGoal
        if self.speedX <> 0 or self.speedY <> 0:
            
            #print "Move: steps=", self.stepsToGoal, "speed:", self.speedX, self.speedY
            self.stepsToGoal -= 1
            self.nextX += self.speedX; self.nextY += self.speedY
            if self.stepsToGoal < 1:
                #print " move: set next goal"
                if nextMove is not None:
                    #print "  move: try", nextMove
                    self.direct(nextMove, graph)
                else:
                    #print " keep going"
                    self.set_goal_onward(graph)                    
            else:
                #print " move: not quite at goal", self.goalX, self.goalY
                pass
                
    def direct(self, dir, graph):
        assert (self.speedX == 0 or self.speedY == 0), "moving X and Y both %d %d" % (self.speedX, self.speedY)
        # My coding effort can't go any faster unless we get out and push
        if (dir == pygame.K_LEFT and self.speedX < 0 or
            dir == pygame.K_RIGHT and self.speedX > 0 or
            dir == pygame.K_DOWN and self.speedY > 0 or
            dir == pygame.K_UP and self.speedY < 0):
            return None
        # If we're not at a node, the only move we need to do anything with
        # now is a backtrack. If we have remaining steps and we're moving
        # right or left, buffer any suggestions to move up or down (and the
        # converse). Otherwise, we've already thrown out repeated
        # instructions to move forward, so the only thing left are the good
        # instructions to move backward... so reverse speed and we're golden.
        if self.stepsToGoal > 0:
            if self.speedX <> 0 and (dir == pygame.K_UP or dir == pygame.K_DOWN):
                return dir
            if self.speedY <> 0 and (dir == pygame.K_RIGHT or dir == pygame.K_LEFT):
                return dir
            
            print "Backtrack: only had", self.stepsToGoal, "steps remaining to get to ", self.goalX, self.goalY, "and now you want to go", pygame.key.name(dir)
            if dir == pygame.K_DOWN:
                self.goalY += 1
            elif dir == pygame.K_UP:
                self.goalY -= 1
            elif dir == pygame.K_RIGHT:
                self.goalX += 1
            else:
                self.goalX -= 1
            self.speedX = self.speedX * -1
            self.speedY = self.speedY * -1
            if self.stepsToGoal <> self.velocity:            
                self.stepsToGoal = self.velocity - self.stepsToGoal
            else:
                print "TURN ON A DIME"
                if dir == pygame.K_DOWN:
                    self.goalY += 1
                elif dir == pygame.K_UP:
                    self.goalY -= 1
                elif dir == pygame.K_RIGHT:
                    self.goalX += 1
                else:
                    self.goalX -= 1
                    
            print "new goal=", self.goalX, self.goalY
                
            # We're moving in the right direction... just need to reset
            # our goal...
            
        else:
            # We're at a node so can try to go in the new direction. If that
            # does not work, and we were moving before, try to keep going
            # that same way... 
            if not self.set_goal_towards(dir, graph):
                #print "HIT WALL, speedX", self.speedX, "steps", self.stepsToGoal
                if self.speedX == 0 and self.speedY == 0:
                    return None
                else:
                    if not self.set_goal_onward(graph):
                        return None
                return dir
        return None
    
    def set_goal_onward(self, graph):
        if self.speedX > 0:
            dir = pygame.K_RIGHT
        elif self.speedX < 0:
            dir = pygame.K_LEFT
        elif self.speedY > 0:
            dir = pygame.K_DOWN
        else:
            dir = pygame.K_UP
            
        if not self.set_goal_towards(dir, graph):                        
            self.speedX = 0
            self.speedY = 0
            self.stepsToGoal = 0
            return False
        return True
    
    def set_goal_towards(self, dir, graph):
        assert self.stepsToGoal < 1, "self still had steps=%d" % self.stepsToGoal
        atX = self.nextX = self.goalX; atY = self.nextY = self.goalY
        
        #self.nextX = atX; self.nextY = atY
        neighborList = graph[(atX, atY)]
        #print "my neighbors:", neighborList 

        self.goalX = atX; self.goalY = atY
        if dir == pygame.K_DOWN:
            self.goalY = atY + 1
        elif dir == pygame.K_UP:
            self.goalY = atY - 1
        elif dir == pygame.K_RIGHT:
            self.goalX = atX + 1
        else:
            self.goalX = atX - 1
        if neighborList.has_key((self.goalX, self.goalY)):
            #print "GOOD NEW DIR GOAL"
            self.stepsToGoal = self.velocity
            calcSteps = float(self.stepsToGoal)
            self.speedX = (self.goalX - atX) / calcSteps
            self.speedY = (self.goalY - atY) / calcSteps
            #print "SET SPEED:", self.speedX, self.speedY 
            return True
        else:
            # Hmm the new goal was invalid and now it's our current goal.
            # Reset it to something valid... where we are at now
            self.goalX = atX; self.goalY = atY
            return False   
        
        
    
if __name__ == '__main__':
    
    pygame.init()
    screen = pygame.display.set_mode(Global.screenResolution)
    pygame.draw.rect(screen, THECOLORS["black"], [0, 0, Global.screenResolution[0], Global.screenResolution[1]], 0)
    pygame.display.set_caption('Pytentene Bugs!')
    myClock = pygame.time.Clock()
    
    delay=100
    interval=50
    pygame.key.set_repeat(delay,interval)
    
    
    pkl_file = open("level2.maze", 'r')
    cruft = pickle.load(pkl_file)
    junk = pickle.load(pkl_file)
    graph = pickle.load(pkl_file)        
    pkl_file.close()    
    
    render_graph(screen, graph)
    
    bugs = []
    firstBug = Bug(graph, x=0, y=0)
    firstBug.set_goal(graph, 0)
    bugs.append(firstBug)
    
    secondBug = Bug(graph, x=1, y=0)
    secondBug.set_goal(graph,1)
    bugs.append(secondBug)
    
    # Time to get myself moving
    me = Me(graph, color=[0,255,0], x=0, y=0)
    nextMove = None

    while True:
        # Decode the stupid keys... but otherwise we're faster than working
        # strings all the time...
        if nextMove is not None:
            if nextMove == pygame.K_UP: printKey = "up"
            elif nextMove == pygame.K_DOWN: printKey = "down"
            elif nextMove == pygame.K_RIGHT: printKey = "right"
            elif nextMove == pygame.K_LEFT: printKey = "left"
            else:
                printKey = "keycode " + str(nextMove)
        else:
            printKey = None
        if printKey is not None:
            #print "next move=", printKey
            pass
        pygame.event.pump()
        key=pygame.key.get_pressed() # we need keys held down, not key events
        if key[pygame.K_DOWN] or key[pygame.K_s]:
            nextMove = me.direct(pygame.K_DOWN, graph)
        elif key[pygame.K_UP] or key[pygame.K_w]:
            nextMove = me.direct(pygame.K_UP, graph)
        elif key[pygame.K_RIGHT] or key[pygame.K_d]:
            nextMove = me.direct(pygame.K_RIGHT, graph)            
        elif key[pygame.K_LEFT] or key[pygame.K_a]:
            nextMove = me.direct(pygame.K_LEFT, graph)            
        else:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                    
                elif event.type == pygame.KEYDOWN:
                    if (pygame.key.get_mods() == 1024): # command key
                        if (event.key == 113):   # q
                            sys.exit()
                    elif event.key == pygame.K_b:
                        print "NEW BUG"
                        b = Bug(graph)
                        b.set_goal(graph)
                        b.paint(screen)
                        bugs.append(b)

        for b in bugs:
            newX = b.nextX + b.speedX; newY = b.nextY + b.speedY
            for t in bugs:
                if b.identity <> t.identity:
                    newTX = t.nextX + t.speedX; newTY = t.nextY + t.speedY
                    collision = False
                    # Works, except when diagonal:
                    #dist = abs(newX - newTX) + abs(newY - newTY)
                    #if dist < .5:
                    #    collision = True
                    
                    paintX = int(newX*Global.cellSize+Global.offsetX)
                    paintY = int(newY*Global.cellSize+Global.offsetY)
                    b.rect.center = [paintX, paintY]                    
                    paintX = int(newTX*Global.cellSize+Global.offsetX)
                    paintY = int(newTY*Global.cellSize+Global.offsetY)
                    t.rect.center = [paintX, paintY]                    
                    
                    if  b.rect.colliderect(t.rect):
                        collision = True
                    if collision:
                        #print "COLLISION"
                        assert b.stepsToGoal < BUG_SPEED + 1, "steps=%d" % b.stepsToGoal
                        assert t.stepsToGoal < BUG_SPEED + 1, "steps=%d" % t.stepsToGoal
                        b.speedX = b.speedX * -1; b.speedY = b.speedY * -1
                        b.stepsToGoal = BUG_SPEED - b.stepsToGoal
                        assert b.stepsToGoal > -1, "steps=%d" % b.stepsToGoal                        
                        t.speedX = t.speedX * -1; t.speedY = t.speedY * -1
                        t.stepsToGoal = BUG_SPEED - t.stepsToGoal
                        assert t.stepsToGoal > -1, "steps=%d" % t.stepsToGoal

        for b in bugs:
            b.move(graph)
        for b in bugs:
            b.erase(screen)            
        for b in bugs:
            b.render(screen)
            
        if nextMove is not None:
            nextMove = me.direct(nextMove, graph)
            
        me.move(nextMove, graph)
        me.erase(screen)
        me.render(screen)
        

        #cursor.paint(screen)
        
        #print myClock.tick(60)
        time = min(float(myClock.tick(60))/60*1.2, 1.5)
        fps = myClock.get_fps()
        #print fps
        pygame.display.flip()    